#!/usr/bin/env python

#import pygtk
#pygtk.require("2.0")

# for debugging within Wing IDE
try:
    import wingdbstub
except:
    pass

import gtk
import gobject
import sys
import redis
import tempfile
import shutil
import linuxcnc
import hal
import gremlin
import os
import pango
import subprocess
import math
import errno

# Tormach modules
from constants import *
from errors import *
import lathe_conversational
import numpad
import tormach_file_util
from ui_common import *
import regression_tests

class lathe(TormachUIBase):

    def __init__(self):
        TormachUIBase.__init__(self)
        self.program_exit_code = EXITCODE_SHUTDOWN

        # glade/gtk.builder setup
        gladefile = os.path.join(GLADE_DIR, 'tormach_lathe_ui.glade')
        self.builder = gtk.Builder()
        self.builder.add_from_file(gladefile)
        self.builder.connect_signals(self)

        # retrieve main window, notebooks, fixed containers
        self.window = self.builder.get_object("main_window")
        self.notebook = self.builder.get_object("notebook")
        self.notebook_main_fixed = self.builder.get_object("notebook_main_fixed")
        self.conv_notebook = self.builder.get_object("conversational_notebook")
        self.offsets_notebook = self.builder.get_object("offsets_notebook")
        self.fixed = self.builder.get_object("fixed")
        notebook_main_fixed = self.builder.get_object("notebook_main_fixed")
        self.notebook_file_util_fixed = self.builder.get_object("notebook_file_util_fixed")
        self.notebook_tool_fixed = self.builder.get_object("tool_offsets_fixed")
        work_offsets_fixed = self.builder.get_object("work_offsets_fixed")

        # throw out mousewheel events to prevent scrolling through notebooks on wheel
        self.notebook.connect("scroll-event", self.on_mouse_wheel_event)
        self.conv_notebook.connect("scroll-event", self.on_mouse_wheel_event)
        self.offsets_notebook.connect("scroll-event", self.on_mouse_wheel_event)

        # redis db for persistent storage.  File location set in redis.conf, currently in config dir (rdb.dump)
        self.redis = redis.Redis()

        # --------------------------------------------------------
        # linuxcnc command and status objects
        # --------------------------------------------------------
        self.command = linuxcnc.command()
        self.status = linuxcnc.stat()
        self.error = linuxcnc.error_channel()


        # --------------------------------------------------------
        # gremlin tool path display setup
        # --------------------------------------------------------
        ini_file_name = sys.argv[2]
        self.inifile = linuxcnc.ini(ini_file_name)
        self.gremlin = Tormach_Lathe_Gremlin(self,680,410)
        notebook_main_fixed.put(self.gremlin, 322, 0)
        

        # message line
        self.notebook_main_fixed.put(self.message_line, 322, 372)
        self.clear_message_line_text()

        # --------------------------------------------------------
        # lathe GUI is used for lathe and duality.
        # Not all HAL pins can be connected if we're running a duality config
        # --------------------------------------------------------

        self.config_label = self.builder.get_object('config_label')
        self.config_label.modify_font(pango.FontDescription('Bebas ultra-condensed 8'))
        
        # elapsed time lable on top of gremlin
        self.notebook_main_fixed.put(self.elapsed_time_label, 935, 388)

        self.is_rapidturn = False
        self.is_sim_config = False

        if "lathe" in ini_file_name:
            config_name = "ZMM_SP503"
            self.is_rapidturn = False
        elif "770" in ini_file_name:
            self.mill_model = '770'
        elif "1100_3_dig" in ini_file_name:
            self.mill_model = '1100-1'
        elif "1100-1" in ini_file_name:
            self.mill_model = '1100-1'
        elif "1100-2" in ini_file_name:
            self.mill_model = '1100-2'
        elif "1100-3" in ini_file_name:
            self.mill_model = '1100-3'
        else:
            self.mill_model = 'Frank'

        if "duality" in ini_file_name:
            config_name = "RapidTurn"
            self.is_rapidturn = True
            config_name += self.mill_model

        if "_sim" in ini_file_name:
            config_name += '-sim'
            self.is_sim_config = True

        self.config_label.set_text(config_name)

        self.setup_hal_pins()

        # errors
        self.error_handler = error_handler(self.builder, self.moving, self.hal)

        # --------------------------------------------------------
        # conversational
        # --------------------------------------------------------
        self.conversational = lathe_conversational.conversational(self,
                                                                  self.status,
                                                                  self.error_handler,
                                                                  self.redis,
                                                                  self.hal)

        # g code file history combo box
        self.file_history_liststore = gtk.ListStore(str, str)
        self.loaded_gcode_filename_combobox = self.builder.get_object("loaded_gcode_combobox")
        self.loaded_gcode_filename_combobox.set_model(self.file_history_liststore)
        cell = gtk.CellRendererText()
        self.loaded_gcode_filename_combobox.pack_start(cell, True)
        self.loaded_gcode_filename_combobox.add_attribute(cell, 'text', 0)
        try:
            for i in range(0, self.redis.llen('recent_file_history')):
                # file history liststore is in form of ['filename', 'path'], but redis only stores the path
                path = self.redis.lpop('recent_file_history')
                self.file_history_liststore.append([os.path.basename(path), path])
        except:
            self.error_handler.write("Retrieval of recent file history failed.", ALARM_LEVEL_DEBUG)
            pass

        self.file_history_liststore.append([CLEAR_CURRENT_PROGRAM, CLEAR_CURRENT_PROGRAM])

        # -------------------------------------------------------
        # PostGUI HAL
        # -------------------------------------------------------
        postgui_halfile = self.inifile.find("HAL", "POSTGUI_HALFILE")
        if postgui_halfile:
            if subprocess.call(["halcmd", "-i", sys.argv[2], "-f", postgui_halfile]):
                self.error_handler.write("Error: something failed running halcmd on '" + postgui_halfile + "'", ALARM_LEVEL_DEBUG)
                sys.exit(1)
        else:
            # complain about missing POSTGUI_HALFILE
            self.error_handler.write("Error: missing POSTGUI_HALFILE in .INI file.", ALARM_LEVEL_DEBUG)
            sys.exit(1)

        # configure the ShuttleXpress (if preset)
        postgui_shuttlexpress_halfile = self.inifile.find("HAL", "POSTGUI_SHUTTLEXPRESS_HALFILE")
        self.jog_shuttle_load_failure = False
        if postgui_shuttlexpress_halfile:
            if subprocess.call(["halcmd", "-i", sys.argv[2], "-f", postgui_shuttlexpress_halfile]):
                #self.error_handler("Warning: something failed running halcmd on " + postgui_shuttlexpress_halfile, ALARM_LEVEL_DEBUG)
                # warning of this is controlled by 'machine_prefs', 'check_for_jog_shuttle'
                self.jog_shuttle_load_failure = True
        else:
            # complain about missing POSTGUI_SHUTTLEXPRESS_HALFILE
            self.error_handler.write("Warning: missing POSTGUI_SHUTTLEXPRESS_HALFILE in .INI file.", ALARM_LEVEL_DEBUG)

        # initialize mesa watchdog status
        self.mesa_watchdog_has_bit_seen = False

        # initialize jog shuttle ring speeds
        # make speed 7 max_velocity, scale such that speed 1 is min
        # INI file [AXIS_0|2] MAX_VELOCITY = 1.5 inches/second or 90 inches/minute
        # 90 IPM / 0.5 min speed => 180, 6th root of 180 is about 2.390
        '''
        min_speed = 0.5, g_multiplier = 2.39
        0.5
        1.195
        2.85605
        6.8259595
        16.314043205
        38.99056326
        93.1874461913
        '''
        min_speed = 0.5
        g_multiplier = 2.390
        c_multiplier = 1.0
        self.hal['jog-ring-speed-1'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-2'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-3'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-4'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-5'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-6'] = c_multiplier * min_speed
        c_multiplier = c_multiplier * g_multiplier
        self.hal['jog-ring-speed-7'] = c_multiplier * min_speed

        # trajmaxvel from ini for maxvel slider
        self.maxvel = self.ini_float('TRAJ', 'MAX_LINEAR_VELOCITY', 1.5)
        
        # jogging direction, conversational tool orientation validation
        self.is_front_toolpost_lathe = False
        display_geometry = self.inifile.find("DISPLAY", "GEOMETRY")
        if display_geometry == "XZ":
            self.is_front_toolpost_lathe = True
        
        # ---------------------------------------------
        # member variable init
        # ---------------------------------------------

        # Set initial toggle button states - previous states used to conditionally update bitmaps

        self.status.poll()
        self.g21 = self.prev_g21 = False
        self.jog_metric_scalar = 1
        self.first_run = True
        self.estop_alarm = True
        self.interp_alarm = False
        self.single_block_active = False
        self.feedhold_active = self.prev_feedhold_active = False
        self.m01_break_active = True
        self.x_referenced = self.prev_x_referenced = 0
        self.z_referenced = self.prev_z_referenced = 0
        self.y_referenced = self.prev_y_referenced = 0
        self.spindle_direction = self.prev_spindle_direction = self.status.spindle_direction
        self.css_active = self.prev_css_active = self.status.gcodes[13] == 960
        self.f_per_rev_active = self.prev_f_per_rev_active = self.status.gcodes[7] == 950
        self.door_switch = self.hal['door-switch']
        self.prev_door_switch = not self.door_switch
        self.home_switch = self.hal['home-switch']
        self.prev_home_switch = not self.home_switch
        self.collet_closer_status = self.prev_collet_closer_status = self.hal['collet-closer-status']
        self.Inputled6 = self.prev_Inputled0 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled1 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled2 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled3 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled4 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled5 = self.hal['Inputled6']
        self.Inputled6 = self.prev_Inputled6 = self.hal['Inputled6']
        self.Inputled7 = self.prev_Inputled7 = self.hal['Inputled7']
        self.Inputled8 = self.prev_Inputled8 = self.hal['Inputled8']
        self.Inputled9 = self.prev_Inputled9 = self.hal['Inputled9']
        self.Inputled10 = self.prev_Inputled10 = self.hal['Inputled10']
        self.Inputled11 = self.prev_Inputled11 = self.hal['Inputled11']
        self.Inputled12 = self.prev_Inputled12 = self.hal['Inputled12']
        self.Inputled13 = self.prev_Inputled13 = self.hal['Inputled13']
        self.Inputled14 = self.prev_Inputled14 = self.hal['Inputled14']
        self.Inputled15 = self.prev_Inputled15 = self.hal['Inputled15']
        self.Outputled0 = self.prev_Outputled0 = self.hal['Outputled0']
        self.Outputled1 = self.prev_Outputled1 = self.hal['Outputled1']
        self.Outputled2 = self.prev_Outputled2 = self.hal['Outputled2']
        self.Outputled3 = self.prev_Outputled3 = self.hal['Outputled3']
        self.Outputled4 = self.prev_Outputled4 = self.hal['Outputled4']
        self.Outputled5 = self.prev_Outputled5 = self.hal['Outputled5']
        self.Outputled6 = self.prev_Outputled6 = self.hal['Outputled6']
        self.Outputled7 = self.prev_Outputled7 = self.hal['Outputled7']
        self.Outputled8 = self.prev_Outputled8 = self.hal['Outputled8']
        self.Outputled9 = self.prev_Outputled9 = self.hal['Outputled9']
        self.Outputled10 = self.prev_Outputled10 = self.hal['Outputled10']
        self.Outputled11 = self.prev_Outputled11 = self.hal['Outputled11']
        self.Outputled12 = self.prev_Outputled12 = self.hal['Outputled12']
        self.Outputled13 = self.prev_Outputled13 = self.hal['Outputled13']
        self.Outputled14 = self.prev_Outputled14 = self.hal['Outputled14']
        self.Outputled15 = self.prev_Outputled15 = self.hal['Outputled15']
        self.notebook_locked = False
        self.cycle_start_led_color = 'dark'
        self.conv_face_basic_ext = 'basic'
        self.conv_od_basic_ext = 'basic'
        self.conv_id_basic_ext = 'basic'
        self.conv_drill_tap = 'drill'
        self.conv_thread_ext_int = 'external'
        self.conv_thread_note = ''
        self.conv_groove_part = 'groove'
        self.conv_chamfer_radius = 'chamfer'
        self.conv_chamfer_od_id = 'od'
        self.combobox_masked = False
        self.dros_locked = False
        self.f_word = self.prev_f_word = 0
        self.s_word = 0
        self.jog_axis = self.prev_jog_axis = -1
        self.jog_ring_axis = self.prev_jog_ring_axis = -1
        self.prev_jog_ring_speed = 0.0
        self.spindle_range_alarm = self.prev_spindle_range_alarm = 0
        self.reset_image_color = 'red'
        self.mm_inch_scalar = 1
        self.is_gcode_program_loaded = False
        self.program_paused_for_door_sw_open = False
        self.current_page = self.prev_page = MAIN_PAGE
        self.current_tool = 0
        self.cpu_usage = 0

        # editor is called from a wrapper script
        # this script changes the directory to ~/gcode
        # then runs the editor with an LD_PRELOAD
        # for a custom libgtk with locked down file chooser dialog
        # and execs the editor (gedit)
        self.gcode_edit_program_to_run = "editscript"


        # ----------------------------------------------
        # Buttons (gtk.eventbox with image on top)
        # ----------------------------------------------

        # create a list of image object names
        self.pixbuf_dict = {}
        self.image_list = ('cycle_start_image', 'single_block_image', 'm01_break_image', 'feedhold_image',
                           'coolant_image', 'rev_image', 'fwd_image', 'spindle_override_100_image',
                           'feed_override_100_image', 'maxvel_override_100_image', 'reset_image',
                           'css_image', 'feed_per_rev_image', 'jog_0001_image', 'jog_0010_image',
                           'jog_0100_image', 'jog_1000_image', 'jog_x_active_led', 'jog_z_active_led', 'jog_y_active_led',
                           'turret_fwd_image', 'tool_touch_chuck', 'touch_x_image', 'touch_z_image', 'reset_tool_image',
                           'set_g30_image', 'conv_id_turn_background', 'id_basic_extended_btn_image',
                           'id_turn_main_image', 'id_turn_detail_image',
                           'conv_drill_tap_image', 'drill_tap_btn_image',
                           'thread_ext_int_image', 'conv_thread_image', 'thread_pitch_image',
                           'conv_groove_part_image', 'groove_part_btn_image', 'finish-editing-button',
                           'corner_chamfer_radius_btn_image', 'corner_id_od_btn_image', 'conv_corner_background',
                           'door_sw_led','encoder_a_led', 'encoder_b_led', 'encoder_z_led', 'tool_orientation_image',
                           'tool_table_orientation_image', 'tool_gcode_orientation_image', 'machine_ok_led',
                           'usbio_input_0_led', 'usbio_input_1_led', 'usbio_input_2_led', 'usbio_input_3_led',
                           'usbio_output_0_led', 'usbio_output_1_led', 'usbio_output_2_led', 'usbio_output_3_led',
                           'error_image_1', 'error_image_2', 'home_sw_led', 'ref_x_image', 'ref_y_image', 'ref_z_image',
                           'collet_closer_image', 'Inputled0', 'Inputled1', 'Inputled2', 'Inputled3', 'Inputled4',
                           'Inputled5',  'Inputled6', 'Inputled7', 'Inputled8', 'Inputled9', 'Inputled10',
                           'Inputled11', 'Inputled12', 'Inputled13', 'Inputled14', 'Inputled15', 'Outputled0', 'Outputled1',
                           'Outputled1', 'Outputled2', 'Outputled3', 'Outputled4', 'Outputled5', 'Outputled6', 'Outputled7', 
                           'Outputled8', 'Outputled9', 'Outputled10', 'Outputled11', 'Outputled12', 'Outputled13',
                           'Outputled14', 'Outputled15')

        self.tool_image_list = ('tt_rear_tp_profile_image', 'tt_rear_tp_lh_turn_image', 'tt_front_tp_profile_image',
                                'tt_front_tp_lh_turn_image', 'tt_rear_tp_turn_image', 'tt_front_tp_turn_image',
                                'tt_rear_tp_thread_image', 'tt_front_tp_thread_image', 'tt_rear_tp_part_image',
                                'tt_front_tp_part_image', 'tt_rear_id_thread_image', 'tt_center_drill_image',
                                'tt_rear_boring_bar_image', 'tt_front_boring_bar_image', 'tt_front_id_thread_image')

        # create dictionary of key value pairs of image names, image objects
        self.image_list = dict(((i, self.builder.get_object(i))) for i in self.image_list)
        self.tool_image_list = dict(((i, self.builder.get_object(i))) for i in self.tool_image_list)

        # gtk.eventboxes
        self.button_list = ('cycle_start', 'single_block', 'm01_break', 'feedhold', 'stop', 'coolant',
                            'reset', 'feed_100', 'rpm_100', 'rapid_100', 'ccw', 'spindle_stop', 'cw', 'zero_z',
                            'feed_per_rev', 'css', 'jog_0001', 'jog_0010', 'jog_0100', 'jog_1000', 'jog_100',
                            'switch_to_mill',
                            'turret_fwd', 'set_g30', 'goto_g30', 'touch_x','touch_z', 'reset_tool',
                            'exit', 'id_basic_extended',
                            'drill_tap', 'thread_ext_int', 'groove_part', 'update',
                            'corner_chamfer_radius', 'corner_id_od',
                            'post_to_file', 'append_to_file', 'clear', 'ref_x', 'ref_y', 'ref_z',
                            'collet_closer')

        self.tool_touch_eventbox_list = ('tt_rear_tp_profile_eventbox', 'tt_rear_id_thread_eventbox',
                                         'tt_front_tp_profile_eventbox','tt_rear_tp_lh_turn_eventbox',
                                         'tt_front_tp_lh_turn_eventbox', 'tt_rear_tp_turn_eventbox',
                                         'tt_front_tp_turn_eventbox', 'tt_rear_tp_thread_eventbox',
                                         'tt_front_tp_thread_eventbox', 'tt_rear_tp_part_eventbox',
                                         'tt_front_tp_part_eventbox', 'tt_center_drill_eventbox',
                                         'tt_rear_boring_bar_eventbox', 'tt_front_boring_bar_eventbox',
                                         'tt_front_id_thread_eventbox')

        # create dictionary of glade names, eventbox objects
        self.button_list = dict(((i, self.builder.get_object(i))) for i in self.button_list)
        self.tool_touch_eventbox_list = dict(((i, self.builder.get_object(i))) for i in self.tool_touch_eventbox_list)

        # Create additional buttons manually
        self.setup_gcode_buttons()
        self.setup_copy_buttons()
        # get initial x/y locations for eventboxes
        for name, eventbox in self.button_list.iteritems():
            eventbox.x = get_x_pos(eventbox)
            eventbox.y = get_y_pos(eventbox)



        # ----------------------------------------------
        # DROs (gtk.entry)
        # ----------------------------------------------

        self.dro_long_format = "%2.4f"
        self.dro_medium_format = "%.1f"
        self.dro_short_format = "%.0F"
        self.dro_dwell_format = "%.2f"
        if self.g21:
            self.dro_tpu_format = "%3.2f"
        else:
            self.dro_tpu_format = "%2.2f"

        # DRO object names on main screen and tool tab of notebook
        self.dro_list = ('dia_dro', 'z_dro', 'y_dro', 'feed_per_min_dro', 'touch_dia_dro', 'touch_z_dro',
                         'feed_per_rev_dro', 'spindle_rpm_dro', 'spindle_css_dro', 'tool_dro',
                         'touch_dia_dro', 'touch_z_dro', 'encoder_counts_dro', 'encoder_counts_dro1', 'encoder_counts_dro2')

        # create  dictionary of DRO names, gtk.entry objects
        self.dro_list = dict(((i, self.builder.get_object(i))) for i in self.dro_list)
        font_description = pango.FontDescription('helvetica ultra-condensed 22')
        for name, dro in self.dro_list.iteritems():
            dro.modify_font(font_description)
            dro.masked = False

        self.dro_list['y_dro'].modify_font(pango.FontDescription('helvetica ultra-condensed 12'))



        # DROs common to all conversational routines
        self.conv_dro_list = (
            'conv_title_dro',
            'conv_work_offset_dro',
            'conv_rough_sfm_dro',
            'conv_rough_fpr_dro',
            'conv_finish_sfm_dro',
            'conv_finish_fpr_dro',
            'conv_max_spindle_rpm_dro')
        
        self.conv_dro_list = dict(((i, self.builder.get_object(i))) for i in self.conv_dro_list)
        for name, dro in self.conv_dro_list.iteritems():
            dro.modify_font(font_description)
            
        self.conv_dro_list['conv_title_dro'].modify_font(pango.FontDescription('helvetica ultra-condensed 18'))
            
        self.conv_rough_sfm_label = self.builder.get_object("conv_rough_sfm_text")
        self.conv_finish_sfm_label = self.builder.get_object("conv_finish_sfm_text")
        self.conv_rough_fpr_label = self.builder.get_object("conv_rough_fpr_text")
        self.conv_finish_fpr_label = self.builder.get_object("conv_finish_fpr_text")


        # od turn DROs, list order sets focus passing order, wraps at end
        self.od_turn_dro_list = (
            'od_turn_tool_num_dro',
            'od_turn_z_end_dro',
            'od_turn_stock_dia_dro',
            'od_turn_final_dia_dro',
            'od_turn_fillet_dro',
            'od_turn_z_start_dro',
            'od_turn_tc_dro',
            'od_turn_finish_doc_dro',
            'od_turn_rough_doc_dro')

        self.od_focus_list = self.od_turn_dro_list

        self.od_turn_dro_list = dict(((i, self.builder.get_object(i))) for i in self.od_turn_dro_list)
        for name, dro in self.od_turn_dro_list.iteritems():
            dro.modify_font(font_description)


        # id turn DROs, list order sets focus passing order, wraps at end
        self.id_basic_dro_list = (
            'id_basic_tool_num_dro',
            'id_basic_z_start_dro',
            'id_basic_z_end_dro',
            'id_basic_final_dia_dro',
            'id_basic_pilot_dia_dro',
            'id_basic_rough_doc_dro',
            'id_basic_finish_doc_dro',
            'id_basic_tc_dro')

        self.id_basic_focus_list = self.id_basic_dro_list

        self.id_basic_dro_list = dict(((i, self.builder.get_object(i))) for i in self.id_basic_dro_list)
        for name, dro in self.id_basic_dro_list.iteritems():
            dro.modify_font(font_description)


        # id turn DROs, list order sets focus passing order, wraps at end
        self.id_turn_dro_list = (
            'id_turn_tool_num_dro',
            'id_turn_z_start_dro',
            'id_turn_z_end_dro',
            'id_turn_final_dia_dro',
            'id_turn_pilot_dia_dro',
            'id_turn_fillet_dro',
            'id_turn_pilot_dro',
            'id_turn_rough_doc_dro',
            'id_turn_finish_doc_dro',
            'id_turn_tc_dro',
            'id_turn_face_doc_dro')

        self.id_focus_list = self.id_turn_dro_list

        self.id_turn_dro_list = dict(((i, self.builder.get_object(i))) for i in self.id_turn_dro_list)
        for name, dro in self.id_turn_dro_list.iteritems():
            dro.modify_font(font_description)

        self.id_turn_fillet_label     = self.builder.get_object("id_turn_fillet_text")
        self.id_turn_pilot_end_label  = self.builder.get_object("id_turn_pilot_end_text")
        self.id_turn_face_doc_label   = self.builder.get_object("id_turn_face_doc_text")
        self.id_turn_tool_text1_label = self.builder.get_object("id_turn_tool_text1")
        self.id_turn_roughing_label   = self.builder.get_object("id_turn_roughing_text")


        # facing DROs, list order sets focus passing order, wraps at end
        self.face_dro_list = (
            'face_tool_num_dro',
            'face_z_end_dro',
            'face_stock_dia_dro',
            'face_x_end_dro',
            'face_z_start_dro',
            'face_tc_dro',
            'face_rough_doc_dro',
            'face_finish_doc_dro',)

        self.face_focus_list = self.face_dro_list

        self.face_dro_list = dict(((i, self.builder.get_object(i))) for i in self.face_dro_list)
        for name, dro in self.face_dro_list.iteritems():
            dro.modify_font(font_description)

        # chamfering DROs, list order sets focus passing order, wraps at end
        self.corner_chamfer_od_focus_list = [
            'corner_chamfer_od_tool_num_dro',
            'corner_chamfer_od_z_start_dro',
            'corner_chamfer_od_z_end_dro',
            'corner_chamfer_od_od_dro',
            'corner_chamfer_od_angle_dro',
            'corner_chamfer_od_rough_doc_dro',
            'corner_chamfer_od_finish_doc_dro',
            'corner_chamfer_od_tc_dro']

        self.corner_chamfer_od_dro_list = self.corner_chamfer_od_focus_list

        self.corner_chamfer_od_dro_list = dict(((i, self.builder.get_object(i))) for i in self.corner_chamfer_od_dro_list)
        for name, dro in self.corner_chamfer_od_dro_list.iteritems():
            dro.modify_font(font_description)

        # Lables
        self.corner_od_x_label = self.builder.get_object('corner_od_x_text')
        self.corner_id_x_label = self.builder.get_object('corner_id_x_text')

        self.corner_chamfer_od_angle_label = self.builder.get_object('corner_chamfer_od_angle_text')
        self.corner_chamfer_od_rough_doc_label = self.builder.get_object('corner_chamfer_od_rough_doc_text')
        self.corner_chamfer_od_finish_doc_label = self.builder.get_object('corner_chamfer_od_finish_doc_text')
        self.corner_chamfer_od_tc_label = self.builder.get_object('corner_chamfer_od_tc_text')

        self.corner_chamfer_id_focus_list = [
            'corner_chamfer_id_tool_num_dro',
            'corner_chamfer_id_z_start_dro',
            'corner_chamfer_id_z_end_dro',
            'corner_chamfer_id_id_dro',
            'corner_chamfer_id_angle_dro',
            'corner_chamfer_id_rough_doc_dro',
            'corner_chamfer_id_finish_doc_dro',
            'corner_chamfer_id_tc_dro']

        self.corner_chamfer_id_dro_list = self.corner_chamfer_id_focus_list

        self.corner_chamfer_id_dro_list = dict(((i, self.builder.get_object(i))) for i in self.corner_chamfer_id_dro_list)

        for name, dro in self.corner_chamfer_id_dro_list.iteritems():
            dro.modify_font(font_description)

        self.corner_chamfer_id_angle_label = self.builder.get_object('corner_chamfer_id_angle_text')
        self.corner_chamfer_id_rough_doc_label = self.builder.get_object('corner_chamfer_id_rough_doc_text')
        self.corner_chamfer_id_finish_doc_label = self.builder.get_object('corner_chamfer_id_finish_doc_text')
        self.corner_chamfer_id_tc_label = self.builder.get_object('corner_chamfer_id_tc_text')

        # Corner Radius DROs, list order sets focus passing order, wraps at end
        self.corner_radius_od_focus_list = [
            'corner_radius_od_tool_num_dro',
            'corner_radius_od_z_start_dro',
            'corner_radius_od_z_end_dro',
            'corner_radius_od_od_dro',
            'corner_radius_od_rough_doc_dro',
            'corner_radius_od_finish_doc_dro',
            'corner_radius_od_tc_dro']

        self.corner_radius_od_dro_list = self.corner_radius_od_focus_list

        self.corner_radius_od_dro_list = dict(((i, self.builder.get_object(i))) for i in self.corner_radius_od_dro_list)
        for name, dro in self.corner_radius_od_dro_list.iteritems():
            dro.modify_font(font_description)

        self.corner_radius_od_rough_doc_label = self.builder.get_object('corner_radius_od_rough_doc_text')
        self.corner_radius_od_finish_doc_label = self.builder.get_object('corner_radius_od_finish_doc_text')
        self.corner_radius_od_tc_label = self.builder.get_object('corner_radius_od_tc_text')

        self.corner_radius_id_focus_list = [
            'corner_radius_id_tool_num_dro',
            'corner_radius_id_z_start_dro',
            'corner_radius_id_z_end_dro',
            'corner_radius_id_id_dro',
            'corner_radius_id_rough_doc_dro',
            'corner_radius_id_finish_doc_dro',
            'corner_radius_id_tc_dro']

        self.corner_radius_id_dro_list = self.corner_radius_id_focus_list

        self.corner_radius_id_dro_list = dict(((i, self.builder.get_object(i))) for i in self.corner_radius_id_dro_list)

        for name, dro in self.corner_radius_id_dro_list.iteritems():
            dro.modify_font(font_description)

        self.corner_radius_id_rough_doc_label = self.builder.get_object('corner_radius_id_rough_doc_text')
        self.corner_radius_id_finish_doc_label = self.builder.get_object('corner_radius_id_finish_doc_text')
        self.corner_radius_id_tc_label = self.builder.get_object('corner_radius_id_tc_text')

        # Groove DROs and labels, list order sets focus passing order, wraps at end
        self.groove_dro_list = (
            'groove_tool_num_dro',
            'groove_tw_dro',
            'groove_stock_dia_dro',
            'groove_final_dia_dro',
            'groove_z_start_dro',
            'groove_z_end_dro',
            'groove_rough_doc_dro',
            'groove_finish_doc_dro',
            'groove_tc_dro' )

        self.groove_focus_list = self.groove_dro_list

        self.groove_dro_list = dict(((i, self.builder.get_object(i))) for i in self.groove_dro_list)
        for name, dro in self.groove_dro_list.iteritems():
            dro.modify_font(font_description)

        self.groove_z_end_label = self.builder.get_object('groove_z_end_text')
        self.groove_r_doc_label = self.builder.get_object('groove_rough_doc_text')
        self.groove_f_doc_label = self.builder.get_object('groove_finish_doc_text')
        self.groove_tw_label    = self.builder.get_object('groove_tool_width_text')
        

        # Part DROs and labels, list order sets focus passing order, wraps at end
        self.part_dro_list = (
            'part_tool_num_dro',
            'part_tw_dro',
            'part_stock_dia_dro',
            'part_final_dia_dro',
            'part_z_start_dro',
            'part_peck_dro',
            'part_retract_dro',
            'part_tc_dro',
            'part_ebw_dro')

        self.part_focus_list = self.part_dro_list

        self.part_dro_list = dict(((i, self.builder.get_object(i))) for i in self.part_dro_list)
        for name, dro in self.part_dro_list.iteritems():
            dro.modify_font(font_description)

        self.part_peck_label       = self.builder.get_object('part_peck_text')
        self.part_retract_label    = self.builder.get_object('part_retract_text')
        self.part_ebw_label       = self.builder.get_object('part_ebw_text')
        self.part_tw_label    = self.builder.get_object('part_tool_width_text')

        # Drill DRO's and labels, list order sets focus passing order, wraps at end
        self.drill_dro_list = (
            'drill_tool_num_dro',
            'drill_z_start_dro',
            'drill_tc_dro',
            'drill_peck_dro',
            'drill_z_end_dro',
            'drill_spindle_rpm_dro',
            'drill_dwell_dro')

        self.drill_focus_list = self.drill_dro_list

        self.drill_dro_list = dict(((i, self.builder.get_object(i))) for i in self.drill_dro_list)
        for name, dro in self.drill_dro_list.iteritems():
            dro.modify_font(font_description)
            
        self.drill_through_hole_hint_label = self.builder.get_object('drill_through_hole_hint_text')

        self.drill_hole_depth_label = self.builder.get_object('drill_hole_depth_text')
        self.drill_peck_depth_label = self.builder.get_object('drill_peck_depth_text')
        self.drill_dwell_label      = self.builder.get_object('drill_dwell_text')
        self.drill_z_end_label      = self.builder.get_object('drill_z_end_text')

        # Tap DRO's and labels, list order sets focus passing order, wraps at end
        self.tap_dro_list = (
            'tap_tool_num_dro',
            'tap_z_start_dro',
            'tap_peck_dro',
            'tap_tc_dro',
            'tap_z_end_dro',
            'tap_spindle_rpm_dro',
            'tap_tpu_dro',
            'tap_pitch_dro')

        self.tap_focus_list = self.tap_dro_list

        self.tap_dro_list = dict(((i, self.builder.get_object(i))) for i in self.tap_dro_list)
        for name, dro in self.tap_dro_list.iteritems():
            dro.modify_font(font_description)

        self.tap_thread_depth_label = self.builder.get_object("tap_thread_depth_text")
        self.tap_tpu_label          = self.builder.get_object("tap_tpu_text")
        self.tap_pitch_label        = self.builder.get_object("tap_pitch_text")
        self.tap_z_end_label        = self.builder.get_object("tap_z_end_text")

        # Thread DRO's and labels
        self.thread_dro_list = (
            'thread_tool_num_dro',
            'thread_z_start_dro',
            'thread_z_end_dro',
            'thread_tc_dro',
            'thread_major_dia_dro',
            'thread_minor_dia_dro',
            'thread_lead_dro',
            'thread_spindle_rpm_dro',
            'thread_tpu_dro',
            'thread_doc_dro',
            'thread_pitch_dro',
            'thread_pass_dro',
            'thread_taper_dro')

        # self.thread_focus_list = self.thread_dro_list

        # The following list order sets focus passing order, wraps at end.
        # Normally the DRO list would set order but not all DROs are passed to.
        self.thread_focus_list = (
            'thread_tool_num_dro',
            'thread_z_start_dro',
            'thread_z_end_dro',
            'thread_taper_dro',
            'thread_tc_dro',
            'thread_major_dia_dro',
            'thread_minor_dia_dro',
            'thread_lead_dro',
            'thread_spindle_rpm_dro',
            'thread_tpu_dro',
            'thread_doc_dro')

        self.thread_dro_list = dict(((i, self.builder.get_object(i))) for i in self.thread_dro_list)
        for name, dro in self.thread_dro_list.iteritems():
            dro.modify_font(font_description)

        self.thread_major_dia_label = self.builder.get_object('thread_major_dia_text')
        self.thread_minor_dia_label = self.builder.get_object('thread_minor_dia_text')
        self.conv_thread_calc_label = self.builder.get_object('thread_calc_text')
        self.conv_thread_pitch_label = self.builder.get_object('thread_pitch_text')
        self.thread_chart_notes_label = self.builder.get_object('thread_chart_notes_text')

        self.thread_chart_combobox = self.builder.get_object('thread_chart')
        self.thread_chart_g20_liststore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.thread_chart_combobox.set_model(self.thread_chart_g20_liststore)

        self.pipe_label1 = self.builder.get_object('pipe_label1')
        self.pipe_label2 = self.builder.get_object('pipe_label2')
        self.pipe_label3 = self.builder.get_object('pipe_label3')
        self.pipe_label4 = self.builder.get_object('pipe_label4')
        self.pipe_label5 = self.builder.get_object('pipe_label5')
        self.pipe_label6 = self.builder.get_object('pipe_label6')

        cell = gtk.CellRendererText()
        self.thread_chart_combobox.pack_start(cell, True)
        self.thread_chart_combobox.add_attribute(cell, 'text', 0)
        cellview = self.thread_chart_combobox.get_child()
        cellview.set_displayed_row(0)

        # provide a 'no selection' entry
        self.thread_chart_g20_liststore.append([' ', ' '])

        thread_filename = THREAD_DATA_SAE
        try:
            with open(thread_filename, 'r') as thread_file:
                for line in thread_file:
                    # remove leading and trailing whitespace from line
                    line = line.strip()
                    if line != '' and line[0] != '#':
                        thread_size, thread_params = line.split(',', 1)
                        self.thread_chart_g20_liststore.append([thread_size, thread_params])
        except:
            self.error_handler.write("Failed to import thread data: %s" % (thread_filename), ALARM_LEVEL_DEBUG)

        # metric threads
        self.thread_chart_g21_liststore = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.thread_chart_g21_liststore.append([' ', ' '])

        thread_filename = THREAD_DATA_METRIC
        try:
            with open(thread_filename, 'r') as thread_file:
                for line in thread_file:
                    # remove leading and trailing whitespace from line
                    line = line.strip()
                    if line != '' and line[0] != '#':
                        thread_size, thread_params = line.split(',', 1)
                        self.thread_chart_g21_liststore.append([thread_size, thread_params])
        except:
            self.error_handler.write("Failed to import thread data: %s" % (thread_filename), ALARM_LEVEL_DEBUG)

        if self.g21:
            self.thread_chart_combobox.set_model(self.thread_chart_g21_liststore)
        else:
            self.thread_chart_combobox.set_model(self.thread_chart_g20_liststore)

        # checkbuttons on settings screen
        self.checkbutton_list = ('use_atc_checkbutton', 'use_gang_tooling_checkbutton', 'use_manual_toolchange_checkbutton',
                                 'use_5c_pulley_ratio_checkbutton', 'use_d1_4_chuck_pulley_checkbutton',
                                 'use_touchscreen_checkbutton', 'disable_keyboard_jogging_checkbutton',
                                 'disable_home_switches_checkbutton', 'enable_usbio_checkbutton')
        self.checkbutton_list = dict(((i, self.builder.get_object(i))) for i in self.checkbutton_list)

        #------------------------------
        # Labels (gtk.label)
        #------------------------------

        # version label shows major version + hash associated with most recent commit past version tag.
        version_label = self.builder.get_object('version_label')
        version_label.modify_font(pango.FontDescription('Bebas ultra-condensed 8'))
        try:
            vf = open(os.path.join(LINUXCNC_HOME_DIR, 'version.txt'))
            for line in vf:
                line = line.strip()
                if line[0] != "":
                    version = line
            version_label.set_text(version)
            self.error_handler.write('UI version: ' + version, ALARM_LEVEL_DEBUG)
        except:
            self.error_handler.write('Failed to retrieve version information!', ALARM_LEVEL_MEDIUM)
            pass




        # dtg labels
        self.x_dtg_label = self.builder.get_object('x_dtg_label')
        self.z_dtg_label = self.builder.get_object('z_dtg_label')
        self.x_dtg_label.modify_font(font_description)
        self.z_dtg_label.modify_font(font_description)

        # /min, /rev, RPM, and FPM labels
        self.f_per_min_label = self.builder.get_object('f_per_min_text')
        self.f_per_rev_label = self.builder.get_object('f_per_rev_text')
        self.rpm_label = self.builder.get_object('rpm_text')
        self.css_label = self.builder.get_object('css_text')

        # tool label for active tool, manual tool change requests
        self.tool_label = self.builder.get_object('tool_label')
        self.tool_label.modify_font(pango.FontDescription('Bebas ultra-condensed 18'))
        self.tool_label.set_text('T:    00')

        # tool touch labels
        self.ftp_text = self.builder.get_object('ftp_text')
        self.rtp_text = self.builder.get_object('rtp_text')

        # ---------------------------------------------
        # overrides sliders (gtk.scale and gtk.adjustment)
        # and associated labels
        # ---------------------------------------------

        # the 'adjustment' object(value, lower bound, upper bound, step inc, page inc, page size)
        # is pre-set in the glade file.
        # Page inc and size should be zero to prevent accidental motion
        # mouse wheel events thrown out to prevent default gtk behavior
        self.feedrate_override_adjustment = self.builder.get_object("feedrate_override_adjustment")
        self.feedrate_override_scale = self.builder.get_object("feedrate_override_scale")
        self.feedrate_override_scale.connect("scroll-event", self.on_mouse_wheel_event)
        self.spindle_override_adjustment = self.builder.get_object("spindle_override_adjustment")
        self.spindle_override_scale = self.builder.get_object("spindle_override_scale")
        self.spindle_override_scale.connect("scroll-event", self.on_mouse_wheel_event)
        self.maxvel_override_adjustment = self.builder.get_object("maxvel_override_adjustment")
        self.maxvel_override_scale = self.builder.get_object("maxvel_override_scale")
        self.maxvel_override_scale.connect("scroll-event", self.on_mouse_wheel_event)
        self.jog_speed_adjustment = self.builder.get_object("jog_speed_adjustment")
        self.jog_speed_scale = self.builder.get_object("jog_speed_scale")
        self.jog_speed_scale.connect("scroll-event", self.on_mouse_wheel_event)
        self.feedrate_override_label = self.builder.get_object("feedrate_override_label")
        self.spindle_override_label = self.builder.get_object("spindle_override_label")
        self.maxvel_override_label = self.builder.get_object("maxvel_override_label")
        self.jog_speed_label = self.builder.get_object('jog_speed_label')
        self.active_gcodes_label = self.builder.get_object("active_gcodes_label")
        self.feedrate_override_adjustment.set_value(100)
        self.spindle_override_adjustment.set_value(100)
        self.maxvel_override_adjustment.set_value(100)


        self.setup_gcode_marks()

        # -----------------------------------------
        # jogging - keyboard and shuttlexpress
        # -----------------------------------------

        # set jog increment
        self.jog_increment = 0.0001
        # set jog speed
        ini_jog_speed = (
            self.inifile.find("DISPLAY", "DEFAULT_LINEAR_VELOCITY")
            or self.inifile.find("TRAJ", "DEFAULT_LINEAR_VELOCITY")
            or self.inifile.find("TRAJ", "DEFAULT_VELOCITY")
            or 1.0)
        self.jog_speed = (float(ini_jog_speed))
        self.jog_speeds = dict((ind,float(self.inifile.find("AXIS_%d" % ind,"MAX_VELOCITY"))) for ind in [0, 2])
        if self.is_rapidturn:
            # rapid turn has Y axis
            self.jog_speeds[1] = float(self.inifile.find("AXIS_1", "MAX_VELOCITY"))

        # set jog speed percentage
        self.jog_override_pct = .4
        # default to continuous jog mode
        self.jog_mode_continuous = True
        # initial jog percent to 40
        self.jog_speed_adjustment.set_value(self.jog_override_pct * 100)

        self.clear_jog_LEDs()

        # keyboard jogging
        self.window.connect("key_press_event", self.on_key_press_or_release)
        self.window.connect("key_release_event", self.on_key_press_or_release)


        # mode/state tracking (debug only)
        self.prev_lcnc_task_mode = -1
        self.prev_lcnc_interp_state = -1
        self.prev_task_state = -1


        # ---------------------------------------------------
        # MDI line (gtk.entry)
        # ---------------------------------------------------

        # global MDI history list of typed in commands
        # arrow up/down inside edit displays
        self.mdi_history = []
        self.mdi_history_index = -1
        try:
            hist_length = self.redis.llen('mdi_history')
            for i in range(0, hist_length):
                self.mdi_history.append(self.redis.lpop('mdi_history'))
            self.mdi_history_index = int(self.redis.hget('machine_prefs', 'mdi_history_index'))
        except:
            self.error_handler.write("Retrieval of MDI history failed.", ALARM_LEVEL_DEBUG)
            pass
        self.mdi_history_max_entry_count = 100
        self.mdi_line = self.builder.get_object("mdi_line")
        self.mdi_line_masked = 0
        font_description = pango.FontDescription('helvetica ultra-condensed 18')
        self.mdi_line.modify_font(font_description)
        self.mdi_line.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('#C3C9CA'))




        # ------------------------------
        # periodic UI updates
        # ------------------------------

        # connect timer to status poll periodic function, 50 ms
        gobject.timeout_add(50, self.periodic_status)
        # this counter is used to call the slower periodic function
        self.periodic_loopcount = 0

        # --------------------------------------------------
        # restore saved machine settings/prefs (Redis)
        # --------------------------------------------------

        # tool change type
        try:
            toolchange_type = self.redis.hget('machine_prefs', 'toolchange_type')
        except:
            pass

        if toolchange_type == None: toolchange_type = 'manual'
        if 'atc' in toolchange_type:
            self.checkbutton_list['use_atc_checkbutton'].set_active(True)
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_ATC_8POSITION
            self.button_list['turret_fwd'].show()
            self.button_list['turret_fwd'].set_sensitive(True)
        elif 'gang' in toolchange_type:
            self.checkbutton_list['use_gang_tooling_checkbutton'].set_active(True)
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_GANG
            self.button_list['turret_fwd'].hide()
            self.button_list['turret_fwd'].set_sensitive(False)
        else:
            self.checkbutton_list['use_manual_toolchange_checkbutton'].set_active(True)
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_MANUAL
            self.button_list['turret_fwd'].hide()
            self.button_list['turret_fwd'].set_sensitive(False)


        # spindle range - high is pulleys with 5C, low is pulleys for D1-4 chuck
        # default to low range if redis error, that's the HAL default
        # as well as slower actual spindle speed if belt is not really in the low
        # position
        self.checkbutton_list['use_5c_pulley_ratio_checkbutton'].set_active(False)
        self.checkbutton_list['use_d1_4_chuck_pulley_checkbutton'].set_active(True)
        self.hal["spindle-range"] = 0
        self.max_rpm = LOW_SPEED_PULLEY_MAX_RPM

        try:
            spindle_range = self.redis.hget('machine_prefs', 'spindle_range')
            # this throws an exception if 'spindle_range' is not found
            if 'low' in spindle_range:
                self.checkbutton_list['use_d1_4_chuck_pulley_checkbutton'].set_active(True)
                self.hal["spindle-range"] = 0
                self.max_rpm = LOW_SPEED_PULLEY_MAX_RPM
            else:
                self.checkbutton_list['use_5c_pulley_ratio_checkbutton'].set_active(True)
                self.hal["spindle-range"] = 1
                self.max_rpm = HIGH_SPEED_PULLEY_MAX_RPM
        except:
            # write to redis to avoid future messages
            if self.hal["spindle-range"] == 0:
                self.redis.hset('machine_prefs', 'spindle_range', 'low')
            else:
                self.redis.hset('machine_prefs', 'spindle_range', 'high')

        # touchscreen
        # remember! Redis stores strings, not bools
        self.touchscreen_enabled = False
        try:
            self.touchscreen_enabled = self.redis.hget('machine_prefs', 'touchscreen') == "True"
        except:
            self.error_handler.write("exception looking for 'machine_prefs', 'touchscreen' in redis, defaulting to False", ALARM_LEVEL_LOW)
            # write to redis to avoid future messages
            self.redis.hset('machine_prefs', 'touchscreen', 'False')
        self.checkbutton_list['use_touchscreen_checkbutton'].set_active(self.touchscreen_enabled)

        # warn if jog shuttle not present
        self.keyboard_jogging_disabled = False
        try:
            self.keyboard_jogging_disabled = self.redis.hget('machine_prefs', 'keyboard_jogging_disabled') == "True"
        except:
            self.error_handler.write("exception looking for 'machine_prefs', 'keyboard_jogging_disabled' in redis, defaulting to False", ALARM_LEVEL_LOW)
            # write to redis to avoid future messages
            self.redis.hset('machine_prefs', 'keyboard_jogging_disabled', 'False')
        self.checkbutton_list['disable_keyboard_jogging_checkbutton'].set_active(self.keyboard_jogging_disabled)

        # limit switch enable
        self.home_switches_enabled = True
        try:
            redis_response = self.redis.hget('machine_prefs', 'home_switches_enabled')
            if redis_response == 'True' or redis_response == None:
                self.home_switches_enabled = True
            else:
                self.home_switches_enabled = False
        except:
            self.error_handler.write("exception looking for 'machine_prefs', 'home_switches_enabled' in redis, defaulting to True", ALARM_LEVEL_LOW)
            # write to redis to avoid future messages
            self.redis.hset('machine_prefs', 'home_switches_enabled', 'True')
        self.checkbutton_list['disable_home_switches_checkbutton'].set_active(not self.home_switches_enabled)
        self.hal['home-switch-enable'] = self.home_switches_enabled


        # usbio board
        try:
            self.usbio_enabled = self.redis.hget('machine_prefs', 'usbio_enabled') == 'True'
            # calls to show and hide usbio_leds are unreliable in init.  These are duplicated with 'first_run' flag behind reset button
            if not self.usbio_enabled:
                self.hide_usbio_leds()
        except:
            self.error_handler.write("exception looking for 'machine_prefs', 'usbio enabled' in redis, defaulting to False", ALARM_LEVEL_LOW)
            # write to redis to avoid future messages
            self.redis.hset('machine_prefs', 'usbio_enabled', 'False')
        self.checkbutton_list['enable_usbio_checkbutton'].set_active(self.usbio_enabled)

        # restore last used values on conversational screens
        self.restore_conv_parameters()

        self.gcodes_display = tormach_file_util.active_codes_display()
        self.notebook_settings_fixed = self.builder.get_object('notebook_settings_fixed')
        self.notebook_settings_fixed.put(self.gcodes_display.sw, 10, 10)

        self.setup_filechooser()

        if self.is_rapidturn:
            self.checkbutton_list['use_manual_toolchange_checkbutton'].set_active(True)

        # dump system info to stdout: motherboard, BIOS, CPU(s), memory
        '''bus = dbus.Bus (dbus.Bus.TYPE_SYSTEM)
        obj = bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/devices/computer')
        dbus_properties = obj.GetAllProperties(dbus_interface='org.freedesktop.Hal.Device')
        self.error_handler('Motherboard, BIOS, and kernel information:', ALARM_LEVEL_DEBUG)
        dbus_key_list = ['system.board.vendor', 'system.board.product', 'system.board.version',
                         'system.board.serial', 'system.firmware.vendor', 'system.firmware.release_date',
                         'system.firmware.version', 'system.kernel.machine', 'system.kernel.name',
                         'system.kernel.version']
        for name in dbus_key_list:
            self.error_handler(name + ': ' + dbus_properties[name], ALARM_LEVEL_DEBUG)

        # /usr/sbin/lsusb
        p = subprocess.Popen('/usr/sbin/lsusb', stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.error_handler(p.stdout.read(), ALARM_LEVEL_DEBUG)

        # /proc/cpuinfo
        self.error_handler('CPU information:', ALARM_LEVEL_DEBUG)
        fp = open('/proc/cpuinfo', 'r')
        self.error_handler(fp.read(), ALARM_LEVEL_DEBUG)
        fp.close()

        # /proc/meminfo
        self.error_handler('memory information:', ALARM_LEVEL_DEBUG)
        fp = open('/proc/meminfo', 'r')
        self.error_handler(fp.read(), ALARM_LEVEL_DEBUG)
        fp.close()
        '''
        # -----------------------------------------
        # tool table init (gtk.treeview)
        # -----------------------------------------

        # tool types stored in front angle value of geo offset register
        self.tool_type_dic = {'rtp_profile': 1, 'rtp_lh_turn': 2, 'rtp_turn': 3, 'rtp_thread': 4,
                              'rtp_part': 5, 'rtp_boring_bar': 6, 'rtp_id_thread': 7, 'center_drill_chuck':8, 'ftp_profile': -1,
                              'ftp_lh_turn': -2, 'ftp_turn': -3, 'ftp_thread': -4, 'ftp_part': -5,
                              'ftp_boring_bar': -6, 'ftp_id_thread': -7,'none': 0}

        # geometry defs for tool_types
        self.front_angle_dic = {'rtp_profile': -107.5, 'rtp_lh_turn': -95, 'rtp_turn': -85, 'rtp_thread': -120,
                                'rtp_part': -90, 'rtp_boring_bar': 5, 'rtp_id_thread': 60, 'center_drill_chuck':-15,
                                'ftp_profile': 72.5, 'ftp_lh_turn': 175, 'ftp_turn': 5, 'ftp_thread': 60, 'ftp_part': 90,
                                'ftp_boring_bar': -85, 'ftp_id_thread': -120, 'none': 0}

        self.back_angle_dic = {'rtp_profile': -72.5, 'rtp_lh_turn': -175, 'rtp_turn': -5, 'rtp_thread': -60,
                               'rtp_part': -90, 'rtp_boring_bar': 85, 'rtp_id_thread': 120, 'center_drill_chuck':15,
                               'ftp_profile': 107.5, 'ftp_lh_turn': 95, 'ftp_turn': 85, 'ftp_thread': 120, 'ftp_part': 90,
                               'ftp_boring_bar': -5, 'ftp_id_thread': -60, 'none': 0}

        self.orientation_dic = {'rtp_profile': 6, 'rtp_lh_turn': 1, 'rtp_turn': 2, 'rtp_thread': 6,
                                'rtp_part': 1, 'rtp_boring_bar': 3, 'rtp_id_thread': 8, 'center_drill_chuck':7,
                                'ftp_profile': 8, 'ftp_lh_turn': 4, 'ftp_turn': 3, 'ftp_thread': 8, 'ftp_part': 4,
                                'ftp_boring_bar': 2, 'ftp_id_thread': 6, 'none': 9}

        if self.is_front_toolpost_lathe:
            self.orientation_dic = {'rtp_profile': 8, 'rtp_lh_turn': 4, 'rtp_turn': 3, 'rtp_thread': 8,
                                    'rtp_part': 4, 'rtp_boring_bar': 2, 'rtp_id_thread': 8, 'center_drill_chuck':7,
                                    'ftp_profile': 6, 'ftp_lh_turn': 1, 'ftp_turn': 2, 'ftp_thread': 6, 'ftp_part': 1,
                                    'ftp_boring_bar': 3, 'ftp_id_thread': 8, 'none': 9}

        # using a treeview/liststore for the tool table, on the tools page of the notebook
        self.tool_liststore = gtk.ListStore(int, str, str, str, str, str, str, int, str, str)

        tool_table_filename = self.inifile.find("EMCIO", "TOOL_TABLE") or ""
        if tool_table_filename == "":
            tool_table_filename = "tool.tbl"
        self.refresh_tool_liststore()

        # Create a TreeView and let it know about the model we created above
        tool_table_treeview = gtk.TreeView(self.tool_liststore)

        # create columns
        tool_num_column =        gtk.TreeViewColumn('Tool')
        tool_description_column= gtk.TreeViewColumn('Description')
        x_geo_offset_column =    gtk.TreeViewColumn('X')
        z_geo_offset_column =    gtk.TreeViewColumn('Z')
        x_wear_offset_column =   gtk.TreeViewColumn('X Wear')
        z_wear_offset_column =   gtk.TreeViewColumn('Z Wear')
        tool_nose_rad_column =   gtk.TreeViewColumn('Nose R')
        tool_tip_orient_column = gtk.TreeViewColumn('Tip')

        # add columns to treeview
        tool_table_treeview.append_column(tool_num_column)
        tool_table_treeview.append_column(tool_description_column)
        tool_table_treeview.append_column(x_geo_offset_column)
        tool_table_treeview.append_column(z_geo_offset_column)
        tool_table_treeview.append_column(x_wear_offset_column)
        tool_table_treeview.append_column(z_wear_offset_column)
        tool_table_treeview.append_column(tool_nose_rad_column)
        tool_table_treeview.append_column(tool_tip_orient_column)

        tool_col_renderer = gtk.CellRendererText()
        tool_col_renderer.set_property('editable', True)
        tool_col_renderer.set_property('cell-background', '#E1B3B7')
        tool_num_column.pack_start(tool_col_renderer, True)
        tool_num_column.set_attributes(tool_col_renderer, text=0)

        tool_description_renderer = gtk.CellRendererText()
        tool_description_renderer.set_property('editable', True)
        tool_description_renderer.set_property('cell-background', '#B3E1D7')
        tool_description_column.pack_start(tool_description_renderer, True)
        tool_description_column.set_attributes(tool_description_renderer, text=1)
        tool_description_column.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        tool_description_column.set_fixed_width(200)
        tool_description_renderer.connect( 'edited', self.on_tool_description_column_edited, self.tool_liststore )

        x_geo_offset_renderer = gtk.CellRendererText()
        x_geo_offset_renderer.set_property('editable', True)
        x_geo_offset_renderer.set_property('cell-background', '#D5E1B3')
        x_geo_offset_column.pack_start(x_geo_offset_renderer, True)
        x_geo_offset_column.set_attributes(x_geo_offset_renderer, text=2)
        x_geo_offset_renderer.connect( 'edited', self.on_x_geo_offset_col_edited, self.tool_liststore )

        z_geo_offset_renderer = gtk.CellRendererText()
        z_geo_offset_renderer.set_property('editable', True)
        z_geo_offset_renderer.set_property('cell-background', '#D6D76C')
        z_geo_offset_column.pack_start(z_geo_offset_renderer, True)
        z_geo_offset_column.set_attributes(z_geo_offset_renderer, text=3)
        z_geo_offset_renderer.connect( 'edited', self.on_z_geo_offset_col_edited, self.tool_liststore )


        x_wear_offset_renderer = gtk.CellRendererText()
        x_wear_offset_renderer.set_property('editable', True)
        x_wear_offset_renderer.set_property('cell-background', '#D5E1B3')
        x_wear_offset_column.pack_start(x_wear_offset_renderer, True)
        x_wear_offset_column.set_attributes(x_wear_offset_renderer, text=4, foreground = 8)
        x_wear_offset_renderer.connect( 'edited', self.on_x_offset_col_edited, self.tool_liststore )

        z_wear_offset_renderer = gtk.CellRendererText()
        z_wear_offset_renderer.set_property('editable', True)
        z_wear_offset_renderer.set_property('cell-background', '#D6D76C')
        z_wear_offset_column.pack_start(z_wear_offset_renderer, True)
        z_wear_offset_column.set_attributes(z_wear_offset_renderer, text=5, foreground = 9)
        z_wear_offset_renderer.connect( 'edited', self.on_z_offset_col_edited, self.tool_liststore )

        nose_radius_col_renderer = gtk.CellRendererText()
        nose_radius_col_renderer.set_property('editable', True)
        nose_radius_col_renderer.set_property('cell-background', '#B3E1D7')
        tool_nose_rad_column.pack_start(nose_radius_col_renderer, True)
        tool_nose_rad_column.set_attributes(nose_radius_col_renderer, text=6)
        nose_radius_col_renderer.connect('edited', self.on_nose_radius_col_edited, self.tool_liststore )

        tip_orient_renderer = gtk.CellRendererText()
        tip_orient_renderer.set_property('editable', True)
        tip_orient_renderer.set_property('cell-background', '#E1D5B3')
        tool_tip_orient_column.pack_start(tip_orient_renderer, True)
        tool_tip_orient_column.set_attributes(tip_orient_renderer, text=7)
        tip_orient_renderer.connect( 'edited', self.on_tip_orient_col_edited, self.tool_liststore )


        # show in notebook

        # create a scrolled window to hold the treeview
        scrolled_window_tool_table = gtk.ScrolledWindow()
        scrolled_window_tool_table.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

        #self.notebook_tool_fixed.put(self.scrolled_window_tool_table, 505, 5)
        work_offsets_fixed.put(scrolled_window_tool_table, 10, 7)
        scrolled_window_tool_table.add_with_viewport(tool_table_treeview)
        scrolled_window_tool_table.set_size_request(585, 357)

        # -----------------------------------------
        # Work Offset table init (gtk.treeview)
        # -----------------------------------------
        work_font = pango.FontDescription('helvetica ultra-condensed 14')

        self.work_liststore = gtk.ListStore(str, float, float, float, float, str, str)


        # Create a TreeView and let it know about the model we created above
        work_treeview = gtk.TreeView(self.work_liststore)

        work_id_column = gtk.TreeViewColumn('')
        work_id_label = gtk.Label('Work')
        #work_id_label.modify_font(work_font)
        work_id_column.set_widget(work_id_label)
        work_id_label.show()

        work_x_column  = gtk.TreeViewColumn('')
        work_x_label = gtk.Label('X')
        #work_x_label.modify_font(work_font)
        work_x_column.set_widget(work_x_label)
        work_x_label.show()

        work_z_column  = gtk.TreeViewColumn('')
        work_z_label = gtk.Label('Z')
        #work_z_label.modify_font(work_font)
        work_z_column.set_widget(work_z_label)
        work_z_label.show()

        work_treeview.append_column(work_id_column)
        work_treeview.append_column(work_x_column)
        work_treeview.append_column(work_z_column)

        work_treeview.set_rules_hint(True)

        work_id_renderer = gtk.CellRendererText()
        work_id_renderer.set_property('editable', False)
        #work_id_renderer.set_property('font-desc', work_font)
        work_id_renderer.set_property('cell-background', '#E1B3B7')
        work_id_renderer.set_property('width',80)
        work_id_column.pack_start(work_id_renderer, True)
        work_id_column.set_attributes(work_id_renderer, text=0)

        work_x_renderer = gtk.CellRendererText()
        work_x_renderer.set_property('editable', False)
        #work_x_renderer.set_property('font-desc', work_font)
        work_x_renderer.set_property('xalign',0.8)
        work_x_renderer.set_property('width',100)
        work_x_column.pack_start(work_x_renderer, True)
        work_x_column.set_attributes(work_x_renderer, text=1, foreground=5, background=6)

        work_z_renderer = gtk.CellRendererText()
        work_z_renderer.set_property('editable', False)
        #work_z_renderer.set_property('font-desc', work_font)
        work_z_renderer.set_property('xalign',0.8)
        work_z_renderer.set_property('width',100)
        work_z_column.pack_start(work_z_renderer, True)
        work_z_column.set_attributes(work_z_renderer, text=3, foreground=5, background=6)


        self.refresh_work_offset_liststore()

        # show in notebook

        # create a scrolled window to hold the treeview
        scrolled_window_work_table = gtk.ScrolledWindow()
        scrolled_window_work_table.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)

        work_offsets_fixed.put(scrolled_window_work_table, 680, 7)
        scrolled_window_work_table.add_with_viewport(work_treeview)
        scrolled_window_work_table.set_size_request(300, 233)

        # get user's home directory
        self.home_dir = os.getenv('HOME')

        # holds path to currently loaded gcode file
        # slow periodic polls for changes and reloads if appropriate
        self.set_current_gcode_path('')

        # display jog shuttle not detected if configured to do so
        if self.jog_shuttle_load_failure:
            self.error_handler.write("Warning - jog shuttle not detected!", ALARM_LEVEL_DEBUG)

        # numlock status
        self.numlock_on = True
        try:
            redis_response = self.redis.hget('machine_prefs', 'numlock_on')
            if redis_response == 'True' or redis_response == None:
                self.numlock_on = True
            else:
                self.numlock_on = False
        except:
            #self.error_handler.write("exception looking for 'machine_prefs', 'numlock_on' in redis, defaulting to True", ALARM_LEVEL_LOW)
            # write to redis to avoid future messages
            self.redis.hset('machine_prefs', 'numlock_on', 'True')
        #self.checkbutton_list['numlock_on'].set_active(self.numlock_on)
        #self.set_numlock(self.numlock_on)

        # NetBIOS name
        self.netbios_name = get_netbios_name(netbios_name_conf_file)
        self.netbios_name_widget = self.builder.get_object('netbios_name')
        self.netbios_name_widget.set_text(self.netbios_name)

        self.setup_key_sets()

        if self.is_rapidturn: self.rapidturn_init()

        self.alt_keyboard_shortcuts = (
            (gtk.keysyms.r, self.button_list['cycle_start']),
            (gtk.keysyms.s, self.button_list['stop']),
            (gtk.keysyms.f, self.button_list['coolant'])
        )

        self.ctrl_keyboard_shortcuts = (
            #(gtk.keysyms.a, self.button_list['foo']),
            #(gtk.keysyms.b, self.button_list['bar']),
            #(gtk.keysyms.c, self.button_list['baz'])
        )

    def setup_hal_pins(self):
        # --------------------------------------------------------
        # HAL setup.  Pins/signals must be connected in POSTGUI halfile
        # --------------------------------------------------------
        self.hal = hal.component("tormach")
        self.hal.newpin("coolant", hal.HAL_BIT, hal.HAL_OUT)
        self.hal.newpin("coolant-iocontrol", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("jog-axis-x-enabled", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("jog-axis-y-enabled", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("jog-axis-a-enabled", hal.HAL_BIT, hal.HAL_IN)

        self.hal.newpin("jog-axis-z-enabled", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("jog-step-button", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("jog-counts", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("jog-ring-speed-signed", hal.HAL_FLOAT, hal.HAL_IN)
        self.hal.newpin("jog-ring-selected-axis", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("jog-gui-step-index", hal.HAL_U32, hal.HAL_OUT)
        self.hal.newpin("jog-gui-ustep-size", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-is-metric", hal.HAL_BIT, hal.HAL_OUT)
        self.hal.newpin("tool-changer-type", hal.HAL_S32, hal.HAL_OUT)
        self.hal.newpin("manual-tool-changed", hal.HAL_BIT, hal.HAL_IO)
        self.hal.newpin("tool-change", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("tool-prep-number", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("spindle-speed-out", hal.HAL_FLOAT, hal.HAL_IN)
        self.hal.newpin("spindle-range", hal.HAL_BIT, hal.HAL_OUT)
        self.hal.newpin("spindle-range-alarm", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("spindle", hal.HAL_BIT, hal.HAL_OUT)
        self.hal.newpin("spindle-iocontrol", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("spindle-disable", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("door-switch", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("collet-closer-status", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled0", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled1", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled2", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled3", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled4", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled5", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled6", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled7", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled8", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled9", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled10", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled11", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled12", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled13", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled14", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Inputled15", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled0", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled1", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled2", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled3", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled4", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled5", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled6", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled7", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled8", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled9", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled10", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled11", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled12", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled13", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled14", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("Outputled15", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("encoder-count", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("encoder-count1", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("encoder-count2", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("jog-ring-speed-1", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-2", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-3", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-4", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-5", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-6", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("jog-ring-speed-7", hal.HAL_FLOAT, hal.HAL_OUT)
        self.hal.newpin("mesa-watchdog-has-bit", hal.HAL_BIT, hal.HAL_IO)
        self.hal.newpin("cycle-time-hours", hal.HAL_U32, hal.HAL_IN)
        self.hal.newpin("cycle-time-minutes", hal.HAL_U32, hal.HAL_IN)
        self.hal.newpin("cycle-time-seconds", hal.HAL_U32, hal.HAL_IN)

        self.hal.newpin("motion-program-line", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("motion-next-program-line", hal.HAL_S32, hal.HAL_IN)
        self.hal.newpin("motion-completed-program-line", hal.HAL_S32, hal.HAL_IN)

        self.hal.newpin("encoder-z", hal.HAL_BIT, hal.HAL_IN)

        if not self.is_rapidturn:
            self.hal.newpin("encoder-a", hal.HAL_BIT, hal.HAL_IN)
            self.hal.newpin("encoder-b", hal.HAL_BIT, hal.HAL_IN)

        self.hal.newpin("machine-ok", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("home-switch", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("home-switch-enable", hal.HAL_BIT, hal.HAL_OUT)

        # usbio board
        self.hal.newpin("usbio-enabled", hal.HAL_BIT, hal.HAL_OUT)
        self.hal.newpin("usbio-input-0", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-input-1", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-input-2", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-input-3", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-output-0", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-output-1", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-output-2", hal.HAL_BIT, hal.HAL_IN)
        self.hal.newpin("usbio-output-3", hal.HAL_BIT, hal.HAL_IN)

        # set for nothing, let redis set it later
        self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_GANG
        self.hal['manual-tool-changed'] = False

        # initial hal states:
        self.prev_coolant_iocontrol = self.hal['coolant'] = 0
        self.coolant_status = self.prev_coolant_status = 0
        self.coolant_lockout = COOLANT_LOCK_OUT_PERIOD

        self.prev_spindle_iocontrol = self.hal['spindle'] = 0

        self.x_jog_enabled = self.hal['jog-axis-x-enabled']
        self.prev_x_jog_enabled = not self.x_jog_enabled
        self.z_jog_enabled = self.hal['jog-axis-z-enabled']
        self.prev_z_jog_enabled = not self.z_jog_enabled
        if self.is_rapidturn:
            self.y_jog_enabled = self.prev_y_jog_enabled = False

        # shuttlexpress step button flag
        self.jog_step_button_was_pressed = 0
        self.prev_hal_jog_counts = self.hal['jog-counts']


        self.hal.ready()

    # ----------------------------------------------------------
    # callbacks
    # ----------------------------------------------------------
    def get_linear_scale(self):
        """Return the scale factor for all linear axes based on current G20/G21 mode"""
        return 25.4 if self.g21 else 1.0

    def get_axis_scale(self, axis_ind):
        return self.get_linear_scale() if axis_ind < 3 else 1.0

    # button press event for all widgets simply moves the eventbox to simulate a 'regular' gtk button motion
    def on_button_press_event(self, widget, data=None):
        # gtk.buildable.get_name returns glade object names
        # wigdet.get_name gets imagebutton names        
        name = gtk.Buildable.get_name(widget) or widget.get_name()
        msg = name + " button was pressed at " + datetime.datetime.now().strftime("-%Y-%m-%d-%H-%M-%S")
        self.error_handler.write(msg, ALARM_LEVEL_DEBUG)	
        do_button_press_motion(widget)

    def on_collet_closer_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        if self.collet_closer_status == 1:
            if not self.spindle_running():
                self.issue_mdi('M65 P2')
            else:
                self.error_handler.write("Cannot open collet while spindle is on", ALARM_LEVEL_MEDIUM)
        else:
            self.issue_mdi('M64 P2')


    def on_ref_x_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        # x = 0, y = 1, z = 2, etc)
        if self.moving(): return
        if self.home_switches_enabled and self.hal['home-switch'] == True:
            # shared switch is aleady active
            self.error_handler.write("Cannot reference axis while home switch is active.", ALARM_LEVEL_MEDIUM)
            return
        if self.status.homed[0]:
            dialog = tormach_file_util.ok_cancel_popup('X    axis    already    homed.    Re-home?');
            dialog.run()
            response = dialog.response;
            dialog.destroy()
            if response != gtk.RESPONSE_OK :
                return           
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.command.home(0)

    def on_ref_y_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        # x = 0, y = 1, z = 2, etc)
        if self.moving(): return
        if self.home_switches_enabled and self.hal['home-switch'] == True:
            # shared switch is aleady active
            self.error_handler.write("Cannot reference axis while home switch is active.", ALARM_LEVEL_MEDIUM)
            return
        if self.status.homed[1]:
            dialog = tormach_file_util.ok_cancel_popup('Y    axis    already    homed.    Re-home?');
            dialog.run()
            response = dialog.response;
            dialog.destroy()
            if response != gtk.RESPONSE_OK :
                return            
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.command.home(1)

    def on_ref_z_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        # x = 0, y = 1, z = 2, etc)
        if self.moving(): return
        if self.home_switches_enabled and self.hal['home-switch'] == True:
            # shared switch is aleady active
            self.error_handler.write("Cannot reference axis while home switch is active.", ALARM_LEVEL_MEDIUM)
            return
        if self.status.homed[2]:
            dialog = tormach_file_util.ok_cancel_popup('Z    axis    already    homed.    Re-home?');
            dialog.run()
            response = dialog.response;
            dialog.destroy()
            if response != gtk.RESPONSE_OK :
                return            
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.command.home(2)

    # right click popup on g code preview
    def on_gcode_sourceview_populate_popup(self, textview, menu):
        set_start_line_menu = gtk.MenuItem("Set start line")
        set_start_line_menu.connect("activate", self.set_start_line_callback)
        # get rid of all default gtk.sourceview menu options
        for child in menu.get_children():
            menu.remove(child)
        menu.append(set_start_line_menu)
        set_start_line_menu.show()

    def set_start_line_callback(self, widget):
        self.set_start_line()

    def on_gcode_scrollbar_button_press(self, scrollbar, event, data=None):
        # mask the motion line update so user can scroll through code without periodic update stepping on his scrolling
        self.sourceview.masked = True

    def on_gcode_scrollbar_button_release(self, scroll_window, event, data=None):
        self.sourceview.masked = False

    # MDI line

    def on_mdi_line_changed(self, widget, data=None):
        # catch all typing, convert to uppercase
        widget.set_text(widget.get_text().upper())

    def on_mdi_line_gets_focus(self, widget, event):
        # prevent access to MDI line when running a program.
        if self.moving():
            self.window.set_focus(None)
            return True
        if not widget.has_focus():
            self.mdi_line.set_text('')
        self.mdi_line.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('black'))
        self.mdi_line.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color(HIGHLIGHT))
        self.mdi_line_masked = 1
        if self.touchscreen_enabled:
            keypad = numpad.numpad_popup(widget, True)
            self.mdi_line_masked = 0
            self.window.set_focus(None)
            return True

    def on_mdi_line_loses_focus(self, widget, event):
        self.mdi_line.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color('white'))
        self.mdi_line.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('#C3C9CA'))
        self.mdi_line.set_text('MDI')
        self.mdi_line.select_region(0, 0)
        if not self.touchscreen_enabled:
            self.mdi_line_masked = 0

    def on_mdi_line_key_press_event(self, widget, event):
        if event.keyval == gtk.keysyms.Escape:
            self.window.set_focus(None)
            if self.touchscreen_enabled:
                self.on_mdi_line_loses_focus(self, widget)
            return True

        if event.keyval == gtk.keysyms.Return:
            if self.gcode_pattern_search.find_last(event):
                return True

        if event.keyval == gtk.keysyms.Down:
#           if self.gcode_pattern_search.mdi_key_command(event):
#               return True
            self.mdi_history_index -= 1
            # range check history index
            if self.mdi_history_index < -1:
                self.mdi_history_index = -1
            if self.mdi_history_index >= 0:
                history_len = len(self.mdi_history)
                if history_len > 0:
                    # place historic text in the display
                    self.mdi_line.set_text(self.mdi_history[self.mdi_history_index])
                    # place cursor at end of line
                    self.mdi_line.set_position(-1)
            else:
                self.mdi_line.set_text("")
            # indicate key has been processed
            return True
        elif event.keyval == gtk.keysyms.Up:
#           if self.gcode_pattern_search.mdi_key_command(event):
#               return True
            self.mdi_history_index += 1
            # range check history index
            history_len = len(self.mdi_history)
            if self.mdi_history_index >= history_len:
                self.mdi_history_index = history_len - 1
            if history_len > 0:
                # place the historic text in the display
                self.mdi_line.set_text(self.mdi_history[self.mdi_history_index])
                # place cursor at end of line
                self.mdi_line.set_position(-1)
            # indicate key has been processed
            return True

        # indicate key not processed
        return False

    def on_mdi_line_activate(self, widget):
        self.mdi_history_index = -1
        command_text = self.mdi_line.get_text()
        # remove leading white space
        command_text = command_text.lstrip()
        # remove trailing white space
        command_text = command_text.rstrip()
        # ignore empty command text
        if len(command_text) == 0:
            # ignore empty command text
            return

        # insert into history
        self.mdi_history.insert(0, command_text)
        history_len = len(self.mdi_history)
        # limit number of history entries
        if history_len > self.mdi_history_max_entry_count:
            # remove oldest entry
            self.mdi_history.pop()
        # delete second occurance of this command if present
        try:
            second_occurance = self.mdi_history.index(command_text, 1)
            if second_occurance > 0:
                self.mdi_history.pop(second_occurance)
        except ValueError:
            # not a problem
            pass

        if command_text == 'REGRESSION_TEST':
            self.start_regression_test_vehicle()
            return

        if (mdi_find_command(self, command_text)):
            return
        
        if (mdi_admin_commands(self, command_text)):
            return

        if not (self.x_referenced and self.z_referenced):
            self.error_handler.write("Must reference X and Z axes before issuing command: " + command_text, ALARM_LEVEL_LOW)
            return
        
        # next lines check for tool change command and don't allow it if turret config and door open
        # feels like a kludge, but is better to nip the command in the bud before it's issued than to catch it
        # in the hal component because IO control v1 has no tool changer fault mechanism and io does funny stuff
        # after a timeout on tool change.
        if 'T' in command_text:
            if (self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_ATC_8POSITION) and (self.hal['door-switch'] == 1):
                self.error_handler.write("Must close enclosure door before indexing turret")
                return
        
        # change to MODE_MDI and issue command
        self.issue_mdi(command_text)
        # clear the text on the input line
        self.mdi_line.set_text("")
        if self.touchscreen_enabled:
            # we don't want the blue highlight to remain
            self.window.set_focus(None)
            self.on_mdi_line_loses_focus(self, widget)
        # look for an M30 in the string
        # it could be inside a comment but who would type a comment into an MDI line?
        if 'M30' in command_text:
            # reset starting line
            self.gcodelisting_mark_start_line(1)

    # Program Control Group

    def on_cycle_start_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        if not self.is_gcode_program_loaded:
            self.error_handler.write("Must load a valid g code program before pressing cycle start", ALARM_LEVEL_LOW)
            return
        if not (self.x_referenced and self.z_referenced):
            self.error_handler.write("Must reference X and Z axes before executing a gcode program", ALARM_LEVEL_LOW)
            return
        if self.notebook.get_current_page() != MAIN_PAGE:
            self.error_handler.write("Cannot start program while not on Main screen", ALARM_LEVEL_LOW)
            return
        if self.hal['door-switch'] == 1:
            self.error_handler.write("Must close door or chuck guard before starting program", ALARM_LEVEL_LOW)
            return
        self.gcode_pattern_search.clear()
        self.window.set_focus(None)
        self.clear_message_line_text()
        if self.program_paused_for_door_sw_open:
            self.resume_program_from_door_sw_open()

        # tool change in progress and manual tool changer
        if self.hal['tool-change'] and self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_MANUAL:
            # tell tool changer comp that operator has indicated tool change complete
            self.hal['manual-tool-changed'] = 1

        # pressing CS should always clear feedhold.
        self.feedhold_active = False
        self.set_image('feedhold_image', 'Feedhold-Black.jpg')
        self.status.poll()
        # if status.paused, we're already in MODE_AUTO, so resume the program
        if self.status.paused:
            if self.single_block_active:
                self.command.auto(linuxcnc.AUTO_STEP)
            else:
                self.command.auto(linuxcnc.AUTO_RESUME)
                return
        # Otherwise, switch to MODE_AUTO and run the code
        if self.status.interp_state == linuxcnc.INTERP_IDLE:
            self.ensure_mode(linuxcnc.MODE_AUTO)
            if self.single_block_active:
                if self.gcode_start_line != 1:
                    self.command.auto(linuxcnc.AUTO_RUN, self.gcode_start_line -1)
                    self.command.auto(linuxcnc.AUTO_PAUSE)
                    self.command.auto(linuxcnc.AUTO_STEP)
                    self.set_start_line(1)
                else:
                    self.command.auto(linuxcnc.AUTO_STEP)
            else:
                self.command.auto(linuxcnc.AUTO_RUN, self.gcode_start_line)
                self.gcode_start_line = 1;

    def on_single_block_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        if not self.program_running():
            if self.single_block_active:
                self.single_block_active = False
                self.set_image('single_block_image', 'Single-Block-Black.jpg')
            else:
                self.single_block_active = True
                self.set_image('single_block_image', 'Single-Block-Green.jpg')
        # not using program running here to include queued MDI
        elif self.status.queue > 0 or self.status.paused:
            if self.single_block_active:
                # only do the auto_resume if we're already in the middle of a move!
                if self.status.current_vel != 0:
                    self.command.auto(linuxcnc.AUTO_RESUME)
                self.single_block_active = False
                self.set_image('single_block_image', 'Single-Block-Black.jpg')
            else:
                self.command.auto(linuxcnc.AUTO_PAUSE)
                self.single_block_active = True
                self.command.auto(linuxcnc.AUTO_STEP)
                self.set_image('single_block_image', 'Single-Block-Green.jpg')

    def on_m01_break_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        if  self.m01_break_active:
            self.command.set_optional_stop(False)
            self.m01_break_active = False
            self.set_image('m01_break_image', 'M01-Break-Black.jpg')
        else:
            self.command.set_optional_stop(True)
            self.m01_break_active = True
            self.set_image('m01_break_image', 'M01-Break-Green.jpg')

    def on_feedhold_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        if self.moving():
            if not self.feedhold_active:
                self.command.auto(linuxcnc.AUTO_PAUSE)
                self.feedhold_active = True
                self.set_image('feedhold_image', 'Feedhold-Green.jpg')

    def on_stop_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.command.abort()
        self.clear_message_line_text()
        if self.hal['coolant']:
            self.hal['coolant'] = False
        if self.feedhold_active:
            self.feedhold_active = False
            self.set_image('feedhold_image', 'Feedhold-Black.jpg')
        if self.is_gcode_program_loaded: self.gcodelisting_mark_motion_line(0)

    def on_coolant_button_release_event(self, widget, data=None):
        # POSTGUI_HALFILE contains:
        # net coolant-flood tormach.coolant => parport.0.pin-14-out
        # net coolant-flood-io iocontrol.0.coolant-flood => tormach.coolant-iocontrol
        # The UI code here watches tormach.coolant-iocontrol for changes from LinuxCNC.
        # See the periodic handler for details
        do_button_release_motion(widget)
        self.window.set_focus(None)
	# disable coolant lockout : if user commands coolant, do it immediately
        self.coolant_lockout = 0
        # use our tormach.coolant HAL pin to track actual coolant state
        coolant_on = self.hal['coolant']
        if not coolant_on:
            self.hal['coolant'] = 1
        else:
            self.hal['coolant'] = 0

    def on_reset_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.clear_message_line_text()

	# temporary until lathe behaves like mill with regard to RESET with a limit active
	if (self.status.limit[0] != 0):
	    error_msg = 'X limit switch active. Cannot Reset'
	    self.error_handler.write(error_msg, ALARM_LEVEL_MEDIUM)
	if (self.status.limit[1] != 0):
	    error_msg = 'Y limit switch active. Cannot Reset'
	    self.error_handler.write(error_msg, ALARM_LEVEL_MEDIUM)
	if (self.status.limit[2] != 0):
	    error_msg = 'Z limit switch active. Cannot Reset'
	    self.error_handler.write(error_msg, ALARM_LEVEL_MEDIUM)
	if (self.status.limit[0] != 0) or (self.status.limit[1] != 0) or (self.status.limit[2] != 0):
	    error_msg = 'Disable limit switches in Settings to Reset then jog off limit.'
	    self.error_handler.write(error_msg, ALARM_LEVEL_MEDIUM)
	    return

        # clear feedhold
        if self.feedhold_active:
            self.feedhold_active = False
            self.set_image('feedhold_image', 'Feedhold-Black.jpg')
        # clear SB status
        self.single_block_active = False
        self.set_image('single_block_image', 'Single-Block-Black.jpg')
        # clear Mesa card watchdog
        self.hal['mesa-watchdog-has-bit'] = 0
        self.mesa_watchdog_has_bit_seen = False
        # clear alarm
        self.estop_alarm = False
        self.error_handler.clear()

        # reset e-stop
        if self.status.task_state != linuxcnc.STATE_ESTOP_RESET:
            self.command.state(linuxcnc.STATE_ESTOP_RESET)
            # leave this wait_complete() in place, see issue #264
            self.command.wait_complete()
        # apparently must be turned on again after being reset from estop
        if self.status.task_state != linuxcnc.STATE_ON:
            self.command.state(linuxcnc.STATE_ON)
            self.command.wait_complete()
            
        # stop motion
        self.command.abort()

        # Turn off coolant & reset associated states
        self.prev_coolant_iocontrol = self.hal['coolant'] = 0
        self.coolant_status = self.prev_coolant_status = 0
	self.coolant_lockout = COOLANT_LOCK_OUT_PERIOD

        # reset/rewind program
	# as long as a limit switch is not active
	if self.is_rapidturn:
	    if (self.status.limit[0] == 0) and (self.status.limit[1] == 0) and (self.status.limit[2] == 0):
		self.issue_mdi('M30')
	else:
	    if (self.status.limit[0] == 0) and (self.status.limit[2] == 0):
		self.issue_mdi('M30')
        self.command.wait_complete()

        # clear live plotter
        self.gremlin.clear_live_plotter()
        # rewind program listing
        # set starting line
        if self.is_gcode_program_loaded:
            self.gcodelisting_mark_start_line(1)
            self.sourceview.scroll_to_mark(self.gcodelisting_start_mark, 0, True, 0, 0.5)

        self.status.poll()
        self.ensure_mode(linuxcnc.MODE_MDI)
        if self.first_run:
            try:
                self.g21 = self.redis.hget('machine_prefs', 'g21') == "True"
                if self.g21 == True:
                    self.issue_mdi("G21")
                tool_num = self.redis.hget('machine_prefs', 'active_tool')
                if (tool_num != '0') and (int(tool_num) <= MAX_NUM_TOOLS):
                    self.issue_mdi('M61 Q' + tool_num)
                    self.command.wait_complete()
                    self.issue_mdi('M61 Q' + tool_num)
                    self.command.wait_complete()
                    offset_num = str(int(tool_num) + 10000)
                    self.issue_mdi('G43 H' + offset_num)
            except:
                self.error_handler.write("Redis failed to retrieve tool information!", ALARM_LEVEL_DEBUG)
                pass
            if not self.usbio_enabled:
                self.hide_usbio_leds()
            self.first_run = False

    def on_feedrate_override_adjustment_value_changed(self, adjustment):
        self.feedrate_override_label.set_text(str(int(adjustment.value))+"%")
        self.command.feedrate(adjustment.value / 100.0)
        self.window.set_focus(None)

    def on_feed_100_button_release_event(self, widget, event, data=None):
        do_button_release_motion(widget)
        self.feedrate_override_adjustment.set_value(100)

    def on_spindle_override_adjustment_value_changed(self, adjustment):
        self.spindle_override_label.set_text(str(int(adjustment.value))+"%")
        self.command.spindleoverride(adjustment.value / 100.0)
        self.window.set_focus(None)

    def on_rpm_100_button_release_event(self, widget, event, data=None):
        do_button_release_motion(widget)
        self.spindle_override_adjustment.set_value(100)

    def on_maxvel_override_adjustment_value_changed(self, adjustment):
        self.maxvel_override_label.set_text(str(int(adjustment.value))+"%")
        # scale this to a percentage of the traj setting
        scaled_val = adjustment.value * self.maxvel / 100
        self.command.maxvel(scaled_val)
        self.window.set_focus(None)

    def on_rapid_100_button_release_event(self, widget, event, data=None):
        do_button_release_motion(widget)
        self.maxvel_override_adjustment.set_value(100)



    # Position/Status Readout Group

    # common dro callbacks

    def on_dro_gets_focus(self, widget, event):
        if self.moving() or not widget.has_focus():
            return
        widget.prev_val = widget.get_text()
        widget.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color(HIGHLIGHT))
        widget.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('black'))
        widget.masked = True
        # only highlight the whole field if the user hasn't selected a portion of it
        if not widget.get_selection_bounds():
            widget.select_region(0, -1)
        if self.touchscreen_enabled:
            numpad.numpad_popup(widget)
            widget.masked = 0
            widget.select_region(0, 0)
            self.window.set_focus(None)


    def on_dro_loses_focus(self, widget, data=None):
        if widget_in_alarm_state(widget): return
        widget.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color('white'))
        widget.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('black'))
        if not self.touchscreen_enabled:
            widget.masked = False
            widget.select_region(0, 0)

    def on_dro_key_press_event(self, widget, event, data=None):
        kv = event.keyval
        if kv == gtk.keysyms.Escape:
            widget.modify_base(gtk.STATE_NORMAL, gtk.gdk.Color('white'))
            widget.modify_text(gtk.STATE_NORMAL, gtk.gdk.Color('black'))
            widget.masked = False
            self.window.set_focus(None)
            return True

    def on_css_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        if self.css_active:
            self.on_spindle_rpm_dro_activate(self.dro_list['spindle_rpm_dro'])
        else:
            self.on_spindle_css_dro_activate(self.dro_list['spindle_css_dro'])

    def on_feed_per_rev_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        if self.f_per_rev_active:
            self.on_feed_per_min_dro_activate(self.dro_list['feed_per_min_dro'])
        else:
            self.on_feed_per_rev_dro_activate(self.dro_list['feed_per_rev_dro'])

    def on_zero_z_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        if self.moving(): return
        self.set_work_offset("Z", 0)

    def on_spindle_css_dro_activate(self, widget, data=None):
        # user entry validation
        valid, dro_val, error_msg = self.conversational.validate_surface_speed(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        css_val = abs(dro_val)
        g96_command = "G96 S%.0f" % (css_val)
        self.issue_mdi(g96_command)

        # unmask DROs
        widget.masked = 0
        self.window.set_focus(None)

    def on_spindle_rpm_dro_activate(self, widget, data=None):
        # issue G97, LED off, FPM computed from X DRO: RPM * (PI * diameter)
        # get DRO value
        # TODO: this should not use the conversational rpm validation
        # it should have its own that uses INI file settings and selected chuck/5c config setting
        valid, dro_val, error_msg = self.conversational.validate_max_spindle_rpm(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        s_rpm = abs(dro_val)
        if s_rpm >= self.max_rpm:
            s_rpm = self.max_rpm
        g97_command = "G97 S%.0f" % (s_rpm)
        self.issue_mdi(g97_command)
                # unmask DROs
        widget.masked = 0
        self.window.set_focus(None)

    def on_feed_per_rev_dro_activate(self, widget, data=None):
        valid, dro_val, error_msg = self.conversational.validate_feedrate(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        feed_per_rev = abs(dro_val)
        # compute feed /min from RPM, set DRO text
        g95_command = "G95 F%.4f" % (feed_per_rev)
        self.issue_mdi(g95_command)
        widget.masked = 0
        self.window.set_focus(None)


    def on_feed_per_min_dro_activate(self, widget, data=None):
        # get DRO value
        valid, dro_val, error_msg = self.conversational.validate_feedrate(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        feed_per_min = abs(dro_val)
        g94_command = "G94 F%.4f" % (feed_per_min)
        self.issue_mdi(g94_command)
        # unmask DROs
        widget.masked = 0
        self.window.set_focus(None)


    def on_z_dro_activate(self, widget):
        valid, dro_val, error_msg = self.conversational.validate_z_point(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        self.set_work_offset("Z", dro_val)
        # allow updates
        widget.masked = 0
        self.window.set_focus(None)

    # duality only
    def on_y_dro_activate(self, widget):
        valid, dro_val, error_msg = self.conversational.validate_z_point(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        self.set_work_offset("Y", dro_val)
        # allow updates
        widget.masked = 0
        self.window.set_focus(None)


    def on_dia_dro_activate(self, widget):
        valid, dro_val, error_msg = self.conversational.validate_dia_val(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        x = dro_val * 2
        self.set_work_offset("X", x)
        # allow updates
        widget.masked = 0
        self.window.set_focus(None)


    def on_key_press_or_release(self, widget, event, data=None):
        kv = event.keyval

        # Utilities
        if kv == gtk.keysyms.Print and event.type == gtk.gdk.KEY_PRESS:
            self.screen_grab()

        # open new terminal window
        # MOD1_MASK indicates the left Alt key pressed
        # CONTROL_MASK indicates either Ctrl key is pressed
        if event.state & gtk.gdk.MOD1_MASK and event.state & gtk.gdk.CONTROL_MASK:
            if kv == gtk.keysyms.x and event.type == gtk.gdk.KEY_PRESS:
                # start a terminal window in $HOME directory
                subprocess.Popen(args=["gnome-terminal", "--working-directory=" + os.getenv('HOME')]).pid
                return True


        # Keyboard functions
        # Return True on TAB to prevent tabbing focus changes
        if kv == gtk.keysyms.Tab:
            return True

        # Disable jogging and pass through depending on focus
        if kv in self.jogging_keys:

            #First, handle specific cases that don't behave by type rules

            # no jogging while mdi line has focus
            if self.mdi_line_masked:
                # Make sure to pass through key presses to navigate MDI
                return False if kv in self.mdi_mask_keys else True

            # Next, check the type of the current focused item and pass through
            # keys if needed
            focused_item = type(self.window.get_focus())
            #print focused_item

            if focused_item in self.key_mask:
                return False if kv in self.key_mask[focused_item] else True


            # No jogging on File or Alarms tab
            active_page = self.notebook.get_current_page()

            # Look up the key mask we need to use for the current notebook
            # This is a clean way to check if we need to disable jogging on a
            # per-page basis. If there isn't an entry for the notebook, then we
            # assume that it's ok if that notebook is active
            if active_page in self.disable_jog_pages and not self.moving():
                #Reject keypress since it's not handled by the focused element
                return True


        # Preconditions checked -  Jogging handled below
        # check to see if we're releasing the key

        # Force jogging to stop whenever shift keys are pressed or released
        # (Mach3 Style)
        if kv in [gtk.keysyms.Shift_L, gtk.keysyms.Shift_R] and not self.program_running():
            for jog_axis in [0,2]:
                self.stop_jog(jog_axis)
            return True

        if event.type == gtk.gdk.KEY_RELEASE and kv in self.jogging_keys:
            self.jogging_key_pressed[kv] = False
            # in or out - x axis
            if kv ==  gtk.keysyms.Left or kv == gtk.keysyms.Right:
                jog_axis = 2
            elif self.is_rapidturn and (kv == gtk.keysyms.Prior or kv == gtk.keysyms.Next):
                # duality y axis
                jog_axis = 1
            # along spindle axis - z axis
            elif kv == gtk.keysyms.Up or kv == gtk.keysyms.Down:
                jog_axis = 0
            else:
                return False

            if self.jog_mode_continuous:
                if not self.program_running():
                    self.stop_jog(jog_axis)
            return True

        elif event.type == gtk.gdk.KEY_PRESS and kv in self.jogging_keys:
            if kv == gtk.keysyms.Right:
                # right arrow - X positive
                jog_axis = 2
                jog_direction = 1
            elif kv == gtk.keysyms.Left:
                jog_axis = 2
                jog_direction = -1
            elif self.is_rapidturn and kv == gtk.keysyms.Prior:
                jog_axis = 1
                jog_direction = 1
            elif self.is_rapidturn and kv == gtk.keysyms.Next:
                jog_axis = 1
                jog_direction = -1
            elif kv == gtk.keysyms.Up:
                jog_axis = 0
                jog_direction = 1
            elif kv == gtk.keysyms.Down:
                jog_axis = 0
                jog_direction = -1
            else:
                return False
            # After determining the axis and direction, run the jog iff the key
            # is not already depressed

            jogging_rapid = event.state & gtk.gdk.SHIFT_MASK

            if not self.jogging_key_pressed[kv]:
                actual_scale = 1.0 if jogging_rapid else self.jog_override_pct
                self.jog(jog_axis, jog_direction, self.jog_speeds[jog_axis] * actual_scale ,False)
            # Update the state of the pressed key
            self.jogging_key_pressed[kv]=True
            self.jogging_rapid = jogging_rapid
            return True

        # Handle feed hold
        if kv == gtk.keysyms.space and self.moving():
            self.enqueue_button_press_release(self.button_list['feedhold'])
            return True

        # Escape key for stop - removed due to issues with turret tool change being aborted
        # if you press ESC rigth after an MDI tool change, task thinks the change didn't complete
        # but the turret successfully changes.  The result is that the operator thinks the new tool is
        # active, but the system has T0 active with no offsets applied
        # better to remove this.  Having ESC defocus and stop is a bad design choice.  Will leave in mill for now
        '''if kv == gtk.keysyms.Escape:
            self.enqueue_button_press_release(self.button_list['stop'])
            return True'''
            
        # alt key shortcuts
        # MOD1_MASK indicates the left alt key pressed
        # MOD5_MASK indicates the right alt key pressed
        if event.state & (gtk.gdk.MOD1_MASK | gtk.gdk.MOD5_MASK) and event.type == gtk.gdk.KEY_RELEASE:
            # alt-Enter, give focus to MDI line
            # only when Main notebook tab is active
            if self.notebook.get_current_page() == MAIN_PAGE and kv == gtk.keysyms.Return and not self.program_running():
                self.on_mdi_line_gets_focus(self.mdi_line, None)
                self.mdi_line.grab_focus()            
            # alt-e, edit current gcode program
            if kv == gtk.keysyms.e:
                # cannot enqueue edit_gcode button press - it only works after File tab has been opened
                path = self.current_gcode_file_path
                if not self.moving():
                    if path != '':
                        file_name_for_display = path.replace(GCODE_BASE_PATH + '/', '')
                        self.error_handler.write('Editing loaded G code file: %s' % file_name_for_display, ALARM_LEVEL_DEBUG)
                        try:
                            # run the editor, store the pid somewhere in case of crash
                            self.editor_pid = subprocess.Popen([self.gcode_edit_program_to_run, path]).pid
                        except OSError:
                            self.error_handler.write("OSError exception raised. Could not run edit program: %s" % self.gcode_edit_program_to_run, ALARM_LEVEL_LOW)
                        except IOError:
                            print "path %s is not a valid G code file" % path
                    else:
                        try:
                            # run the editor, store the pid somewhere in case of crash
                            self.editor_pid = subprocess.Popen([self.gcode_edit_program_to_run]).pid
                        except OSError:
                            self.error_handler.write("OSError exception raised. Could not run edit program: %s" % self.gcode_edit_program_to_run, ALARM_LEVEL_LOW)                            

            for (k_val, k_widget) in self.alt_keyboard_shortcuts:
                if kv == k_val:
                    self.enqueue_button_press_release(k_widget)

        # ctrl key shortcuts
        # CONTROL_MASK indicates the left ctrl key pressed
        if event.state & gtk.gdk.CONTROL_MASK and event.type == gtk.gdk.KEY_RELEASE:
            for (k_val, k_widget) in self.ctrl_keyboard_shortcuts:
                if kv == k_val:
                    self.enqueue_button_press_release(k_widget)

        return False

    def on_mouse_wheel_event(self, widget, event):
        return True

    def on_jog_speed_scale_button_press_event(self, widget, data=None):
        if not self.jog_mode_continuous:
            self.clear_jog_LEDs()
            self.hal['jog-gui-ustep-size'] = 0
            self.jog_mode_continuous = True

    def clear_jog_LEDs(self):
        # turn all the JOG button LEDs "off"
        if self.g21:
            self.set_image('jog_0001_image', 'Metric-001-Black.jpg')
            self.set_image('jog_0010_image', 'Metric-100-Black.jpg')
            self.set_image('jog_0100_image', 'Metric-1.00-Black.jpg')
            self.set_image('jog_1000_image', 'Metric-10.0-Black.jpg')
        else:
            self.set_image('jog_0001_image', '0001-Pressed-Black.jpg')
            self.set_image('jog_0010_image', '0010-Pressed-Black.jpg')
            self.set_image('jog_0100_image', '0100-Pressed-Black.jpg')
            self.set_image('jog_1000_image', '1000-Pressed-Black.jpg')

    def on_jog_speed_adjustment_value_changed(self, adjustment):
        self.jog_speed_label.set_text(str(int(adjustment.value))+"%")
        if not self.jog_mode_continuous:
            self.clear_jog_LEDs()
            self.hal['jog-gui-ustep-size'] = 0
            self.jog_mode_continuous = True
        self.jog_override_pct = (adjustment.value)/100

    def on_jog_100_button_release_event(self, widget, event, data=None):
        do_button_release_motion(widget)
        self.jog_speed_adjustment.set_value(100)

    def on_jog_0001_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.jog_increment = 0.0001
        self.hal['jog-gui-step-index'] = 0
        self.hal['jog-gui-ustep-size'] = self.jog_increment/1000.0
        self.jog_mode_continuous = False
        self.clear_jog_LEDs()
        if self.g21:
            self.set_image('jog_0001_image', 'Metric-001-Green.jpg')
        else:
            self.set_image('jog_0001_image', '0001-Pressed-Green.jpg')

    def on_jog_0010_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.jog_increment = 0.001
        self.hal['jog-gui-step-index'] = 1
        self.hal['jog-gui-ustep-size'] = self.jog_increment/1000.0
        self.jog_mode_continuous = False
        self.clear_jog_LEDs()
        if self.g21:
            self.set_image('jog_0010_image', 'Metric-100-Green.jpg')
        else:
            self.set_image('jog_0010_image', '0010-Pressed-Green.jpg')

    def on_jog_0100_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.jog_increment = 0.01
        self.hal['jog-gui-step-index'] = 2
        self.hal['jog-gui-ustep-size'] = self.jog_increment/1000.0
        self.jog_mode_continuous = False
        self.clear_jog_LEDs()
        if self.g21:
            self.set_image('jog_0100_image', 'Metric-1.00-Green.jpg')
        else:
            self.set_image('jog_0100_image', '0100-Pressed-Green.jpg')

    def on_jog_1000_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.jog_increment = 0.1
        self.hal['jog-gui-step-index'] = 3
        self.hal['jog-gui-ustep-size'] = self.jog_increment/1000.0
        self.jog_mode_continuous = False
        self.clear_jog_LEDs()
        if self.g21:
            self.set_image('jog_1000_image', 'Metric-10-Green.jpg')
        else:
            self.set_image('jog_1000_image', '1000-Pressed-Green.jpg')

    def on_ccw_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        # Per conversation with John, better to do this with command.spindle_fwd
        # quick look at touchy/axis makes me think that there's no way to set
        # spindle speed in MODE_MANUAL.
        if self.door_switch == 1:
            self.error_handler.write("Cannot start spindle with door open", ALARM_LEVEL_LOW)
        else:
            # TODO: this should not use the conversational rpm validation
            # it should have its own that uses INI file settings and selected chuck/5c config setting
            valid, dro_val, error_msg = self.conversational.validate_max_spindle_rpm(self.dro_list['spindle_rpm_dro'])
            if not valid:
                self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
                return
            self.issue_mdi("m4")

    def on_spindle_stop_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        self.issue_mdi("m5")

    def on_cw_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.window.set_focus(None)
        if self.door_switch == 1:
            self.error_handler.write("Cannot start spindle with door open", ALARM_LEVEL_LOW)
        else:
            # TODO: this should not use the conversational rpm validation
            # it should have its own that uses INI file settings and selected chuck/5c config setting
            valid, dro_val, error_msg = self.conversational.validate_max_spindle_rpm(self.dro_list['spindle_rpm_dro'])
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        self.issue_mdi("m3")



    # ---------------------------------------------------------------------
    # File tab
    #----------------------------------------------------------------------

    # edit selected or else currently loaded gcode file
    def on_edit_gcode_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        if not self.moving():
            path = self.hd_file_chooser.selected_path or self.usb_file_chooser.selected_path

            # if nothing selected edit the currently loaded gcode file if there is one
            if path == '':
                path = self.current_gcode_file_path
            # path can be blank if no file loaded
            if path != '':
                try:
                    # run the editor
                    pid = subprocess.Popen([self.gcode_edit_program_to_run, path]).pid
                except OSError:
                    self.error_handler.write("OSError exception raised. Could not run edit program: %s" % self.gcode_edit_program_to_run, ALARM_LEVEL_LOW)

    # ---------------------------------------------------------------------
    # Alarms/settings tab
    #----------------------------------------------------------------------

    def on_netbios_name_activate(self, widget, data=None):
        nbentry_text = validate_netbios_name(widget.get_text())
        # if empty, fill in existing name
        if nbentry_text == '':
            nbentry_text = self.netbios_name
        # stuff the result back into the text entry
        widget.set_text(nbentry_text)
        # if changed, set the new NetBIOS name
        if nbentry_text != self.netbios_name:
            self.netbios_name = nbentry_text
            set_netbios_name(netbios_name_conf_file, nbentry_text)

    def on_settings_lockout_activate(self, widget, data=None):
        lockout_code = widget.get_text()
        # TODO - set up a self.settings_lockout_level

    def on_disable_keyboard_jogging_checkbutton_toggled(self, widget, data=None):
        self.keyboard_jogging_disabled = widget.get_active()
        self.redis.hset('machine_prefs', 'keyboard_jogging_disabled', self.keyboard_jogging_disabled)
        self.window.set_focus(None)

    def on_use_touchscreen_checkbutton_toggled(self, widget, data=None):
        self.touchscreen_enabled = widget.get_active()
        # change fileview objects flag so that the right click menu obeys the current touchscreen setting,
        # not the setting used when the objects were created.
        if not self.first_run:
            self.hd_file_chooser.touchscreen_enabled = self.touchscreen_enabled
            self.usb_file_chooser.touchscreen_enabled = self.touchscreen_enabled
        self.redis.hset('machine_prefs', 'touchscreen', self.touchscreen_enabled)
        self.window.set_focus(None)

    def on_use_atc_checkbutton_toggled(self, widget, data=None):
        if widget.get_active():
            self.redis.hset('machine_prefs', 'toolchange_type', 'atc')
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_ATC_8POSITION
            self.button_list['turret_fwd'].show()
            self.button_list['turret_fwd'].set_sensitive(True)

    def on_use_gang_tooling_checkbutton_toggled(self, widget, data=None):
        if widget.get_active():
            self.redis.hset('machine_prefs', 'toolchange_type', 'gang')
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_GANG
            self.button_list['turret_fwd'].hide()
            self.button_list['turret_fwd'].set_sensitive(False)

    def on_use_manual_toolchange_checkbutton_toggled(self, widget, data=None):
        if widget.get_active():
            self.redis.hset('machine_prefs', 'toolchange_type', 'manual')
            self.hal['tool-changer-type'] = TOOL_CHANGER_TYPE_MANUAL
            self.button_list['turret_fwd'].hide()
            self.button_list['turret_fwd'].set_sensitive(False)

    def on_use_5c_pulley_ratio_checkbutton_toggled(self, widget, data=None):
        # NB - for some reason, these checkbuttons emit the toggled signal when the UI starts.
        # next line is a kludge to make sure the spindle range doesn't get conked by UI startup
        if self.current_page != SETTINGS_PAGE: return
        if widget.get_active():
            self.hal["spindle-range"] = 1
            self.redis.hset('machine_prefs', 'spindle_range', 'high')

    def on_use_d1_4_chuck_pulley_checkbutton_toggled(self, widget, data=None):
        # NB - for some reason, these checkbuttons emit the toggled signal when the UI starts.
        # next line is a kludge to make sure the spindle range doesn't get conked by UI startup
        if self.current_page != SETTINGS_PAGE: return
        if widget.get_active():
            self.hal["spindle-range"] = 0
            self.redis.hset('machine_prefs', 'spindle_range', 'low')

    def on_switch_to_mill_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.switch_to_mill()



    def ini_float(self, ini_section, ini_var, def_val):
        val = def_val
        try:
            val =  float(self.inifile.find(ini_section, ini_var))
        except TypeError:
            pass
        return val

    def ini_flag(self, ini_section, ini_var, def_val):
        # return True for YES in INI file, def_val for anything else
        val = def_val
        val_str =  self.inifile.find(ini_section, ini_var)
        if val_str != None:
            if val_str.upper() == "YES":
                val = True
        return val

    def enable_home_switch(self, axis_index, enable_flag):
        print 'enable home axis %d, flag %d' % (axis_index, enable_flag)
        # if enable_flag is False sets home_offset, search velocity and latch velocity to 0.0
        # else sets all to INI file values
        # 0 for X, 1 for Y (duality), 2 for Z
        if axis_index == 0:
            axis_N = "AXIS_0"
        elif axis_index == 1:
            axis_N = "AXIS_1"
        elif axis_index == 2:
            axis_N = "AXIS_2"
        else:
            self.error_handler.write("invalid axis index '%d' for enable_home_switch()" % axis_index, ALARM_LEVEL_LOW)
            return
        home = self.ini_float(axis_N, "HOME", 0.0)
        if (enable_flag == True):
            home_offset = self.ini_float(axis_N, "HOME_OFFSET", 0.0)
            home_search_vel = self.ini_float(axis_N, "HOME_SEARCH_VEL", 0.0)
            home_latch_vel = self.ini_float(axis_N, "HOME_LATCH_VEL", 0.0)
        else:
            home_offset = 0.0
            home_search_vel = 0.0
            home_latch_vel = 0.0
        home_final_vel = self.ini_float(axis_N, "HOME_FINAL_VEL", -1)
        home_use_index = self.ini_flag(axis_N, "HOME_USE_INDEX", False)
        home_ignore_limits = self.ini_flag(axis_N, "HOME_IGNORE_LIMITS", False)
        home_home_is_shared = self.ini_flag(axis_N, "HOME_IS_SHARED", 0)
        home_sequence = self.ini_flag(axis_N, "HOME_SEQUENCE", 0)
        volatile_home = self.ini_flag(axis_N, "VOLATILE_HOME", 0)
        locking_indexer = self.ini_flag(axis_N, "LOCKING_INDEXER", 0)

        self.command.set_homing_params(axis_index, home, home_offset, home_final_vel, home_search_vel,
                                       home_latch_vel, home_use_index, home_ignore_limits,
                                       home_home_is_shared, home_sequence, volatile_home,
                                       locking_indexer)

    def on_disable_home_switches_checkbutton_toggled(self, widget, data=None):
        self.home_switches_enabled = not widget.get_active()
        self.hal["home-switch-enable"] = self.home_switches_enabled
        self.redis.hset('machine_prefs', 'home_switches_enabled', self.home_switches_enabled)
        self.window.set_focus(None)
        if self.home_switches_enabled:
            self.enable_home_switch(0, True)
            if self.is_rapidturn:
                self.enable_home_switch(1, True)
            self.enable_home_switch(2, True)
        else:
            self.enable_home_switch(0, False)
            if self.is_rapidturn:
                self.enable_home_switch(1, False)
            self.enable_home_switch(2, False)


    # ----------------------------------
    # Alarms page
    # ----------------------------------

    def on_clear_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.error_handler.clear_history()


    def on_update_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.software_update()


    # ---------------------------------------------------------------------
    # tool touch off tab
    #----------------------------------------------------------------------
    # on paging to the tool touch we should check the tool type for the currently loaded tool.  If the tool type is 0 (not yet configured)
    # then we display the spash screen.  Also, we need to offer a "reset tool" button that changes the type to 0 and displays the splash.

    # lathe tool highlighting on mouseover
    def on_tt_rear_tp_profile_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_profile_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_profile_highlight.png'))

    def on_tt_rear_tp_profile_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_profile_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_profile.png'))

    def on_tt_rear_tp_lh_turn_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_lh_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_turn_lh_highlight.png'))

    def on_tt_rear_tp_lh_turn_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_lh_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_turn_lh.png'))

    def on_tt_front_tp_profile_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_profile_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_profile_highlight.png'))

    def on_tt_front_tp_profile_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_profile_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_profile.png'))

    def on_tt_front_tp_lh_turn_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_lh_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_turn_highlight_lh.png'))

    def on_tt_front_tp_lh_turn_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_lh_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_turn_lh.png'))

    def on_tt_rear_tp_turn_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_turn_highlight.png'))

    def on_tt_rear_tp_turn_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_turn.png'))

    def on_tt_front_tp_turn_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_turn_highlight.png'))

    def on_tt_front_tp_turn_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_turn_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_turn.png'))

    def on_tt_rear_tp_thread_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_thread_highlight.png'))

    def on_tt_rear_tp_thread_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_thread.png'))

    def on_tt_front_tp_thread_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_thread_highlight.png'))

    def on_tt_front_tp_thread_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_thread.png'))

    def on_tt_rear_tp_part_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_part_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_part_highlight.png'))

    def on_tt_rear_tp_part_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_tp_part_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_part.png'))

    def on_tt_front_tp_part_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_part_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_part_highlight.png'))

    def on_tt_front_tp_part_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_tp_part_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_part.png'))

    def on_tt_rear_id_thread_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_id_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_id_thread_highlight.png'))

    def on_tt_rear_id_thread_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_id_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_id_thread.png'))

    def on_tt_center_drill_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_center_drill_image'].set_from_file(os.path.join(GLADE_DIR, 'center_drill_highlight.png'))

    def on_tt_center_drill_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_center_drill_image'].set_from_file(os.path.join(GLADE_DIR, 'center_drill.png'))

    def on_tt_rear_boring_bar_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_boring_bar_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_boring_bar_highlight.png'))

    def on_tt_rear_boring_bar_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_rear_boring_bar_image'].set_from_file(os.path.join(GLADE_DIR, 'rear_tp_boring_bar.png'))

    def on_tt_front_boring_bar_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_boring_bar_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_boring_bar_highlight.png'))

    def on_tt_front_boring_bar_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_boring_bar_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_boring_bar.png'))

    def on_tt_front_id_thread_eventbox_enter_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_id_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_id_thread_highlight.png'))

    def on_tt_front_id_thread_eventbox_leave_notify_event(self, widget, data=None):
        self.tool_image_list['tt_front_id_thread_image'].set_from_file(os.path.join(GLADE_DIR, 'front_tp_id_thread.png'))

    # lathe tool select callbacks
    # check to ensure tool number is OK before allowing user past the splash screen
    def on_tt_rear_tp_profile_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_profile.png")
        self.write_tool_type(self.current_tool, "rtp_profile")

    def on_tt_front_tp_profile_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_profile.png")
        self.write_tool_type(self.current_tool, "ftp_profile")

    def on_tt_rear_tp_lh_turn_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_lh_turn.png")
        self.write_tool_type(self.current_tool, "rtp_lh_turn")

    def on_tt_front_tp_lh_turn_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_lh_turn.png")
        self.write_tool_type(self.current_tool, "ftp_lh_turn")

    def on_tt_rear_tp_turn_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_turn.png")
        self.write_tool_type(self.current_tool, "rtp_turn")

    def on_tt_front_tp_turn_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_turn.png")
        self.write_tool_type(self.current_tool, "ftp_turn")

    def on_tt_rear_tp_thread_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_thread.png")
        self.write_tool_type(self.current_tool, "rtp_thread")

    def on_tt_front_tp_thread_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_thread.png")
        self.write_tool_type(self.current_tool, "ftp_thread")

    def on_tt_rear_tp_part_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_part.png")
        self.write_tool_type(self.current_tool, "rtp_part")

    def on_tt_front_tp_part_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_part.png")
        self.write_tool_type(self.current_tool, "ftp_part")

    def on_tt_center_drill_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("center_drill_chuck.png")
        self.write_tool_type(self.current_tool, "center_drill_chuck")

    def on_tt_rear_boring_bar_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_boring_bar.png")
        self.write_tool_type(self.current_tool, "rtp_boring_bar")

    def on_tt_front_boring_bar_eventbox_button_press_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_boring_bar.png")
        self.write_tool_type(self.current_tool, "ftp_boring_bar")

    def on_tt_rear_id_thread_eventbox_button_release_event(self, widget, data=None):
        self.clear_tool_select_images("rtp_id_thread.png")
        self.write_tool_type(self.current_tool, "rtp_id_thread")

    def on_tt_front_id_thread_eventbox_button_release_event(self, widget, data=None):
        self.clear_tool_select_images("ftp_id_thread.png")
        self.write_tool_type(self.current_tool, "ftp_id_thread")

    def on_notebook_switch_page(self, notebook, page, page_num):
	TormachUIBase.on_notebook_switch_page(self,notebook, page, page_num)
        if notebook.get_current_page() == CONVERSATIONAL_PAGE:
            self.save_title_dro()
        if page_num != ALARMS_PAGE:
            # we're leaving the alarm page, so clear interp_alarm flag as we've already seen the alarm
            # and this is the only mechanism for "clearing" these alarms.
            # the estop alarm is ONLY cleared on RESET button press.
            self.interp_alarm = False
            # don't want to erase the alarms log every time.  Should come up with a good way to distinguish between
            # old and current error messages.  Color code?
            self.error_handler.clear()

        if page_num == ALARMS_PAGE:
            pass

        if page_num == SETTINGS_PAGE:
            # refresh active codes display
            self.gcodes_display.highlight_active_codes(self.active_gcodes())

        if page_num == TOOL_PAGE:
            tool_type = self.get_tool_type(self.current_tool)
            # set this proactively, because the turret_fwd button gets focus on page switch, preventing
            # the dro from updating.
            self.dro_list['tool_dro'].set_text(self.dro_short_format % self.current_tool)
            nose_radius = ( 0.5 * self.status.tool_table[self.current_tool].diameter)
            tip_orient = self.status.tool_table[self.current_tool].orientation
            tool_type += ".png"
            self.clear_tool_select_images(tool_type, True)
            # prevent highlighting of tool dro on page switch
            self.dro_list['tool_dro'].select_region(0, 0)
            # clear touch off LEDs if tool has no offset
            self.set_touch_x_z_leds(self.current_tool)
            self.refresh_tool_liststore()
            self.window.set_focus(None)

        if page_num == CONVERSATIONAL_PAGE:
            self.load_title_dro()            
            if self.g21:
                self.conv_rough_sfm_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >roughing    mpm:</span>')
                self.conv_finish_sfm_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >finishing    mpm:</span>')

                self.conv_rough_fpr_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >roughing    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >mm/rev</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" > :</span>')
                self.conv_finish_fpr_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >finishing    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >mm/rev</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" > :</span>')

                self.tap_pitch_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >Pitch    (mm)    :</span>')
                self.tap_tpu_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >Threads/mm    :</span>')
                self.conv_thread_calc_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >threads/mm    :</span>')
                self.conv_thread_pitch_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >pitch    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >(mm)</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" > :</span>')

                self.thread_chart_combobox.set_model(self.thread_chart_g21_liststore)

            else:
                self.conv_rough_sfm_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >roughing    sfm:</span>')
                self.conv_finish_sfm_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >finishing    sfm:</span>')

                self.conv_rough_fpr_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >roughing    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >in/rev</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >   :</span>')
                self.conv_finish_fpr_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >finishing    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >in/rev</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >   :</span>')

                self.tap_pitch_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >Pitch     (inches)    :</span>')
                self.tap_tpu_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >Threads/inch    :</span>')
                self.conv_thread_calc_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >threads/inch    :</span>')
                self.conv_thread_pitch_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >pitch    </span><span weight="light" font_desc="Bebas 10" font_stretch="ultracondensed" foreground="white" >(inch)</span><span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" > :</span>')

                self.thread_chart_combobox.set_model(self.thread_chart_g20_liststore)

            self.window.set_focus(None)

            #self.conv_page_number = OD_TURN


    def on_offsets_notebook_switch_page(self, notebook, page, page_num):
        offsets_table_page_num = 1
        if page_num == offsets_table_page_num:
            self.refresh_tool_liststore()

    def _conversational_notebook_switch_page(self, conv_page_num ):
        self.save_title_dro()
        self.load_title_dro(conv_page_num)
        if conv_page_num == OD_TURN:
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)

        if conv_page_num == ID_TURN:
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)

        if conv_page_num == FACE:
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)

        if conv_page_num == CHAMFER:
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)

        if conv_page_num == GROOVE_PART:  # Groove/Part button should also set these
            if self.conv_groove_part == 'groove':
                self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(True)
                self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
                self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
                self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
                self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)
            else:   # self.conv_groove_part == 'part':
                self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
                self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(False)
                self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(True)
                self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(True)
                self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(True)

        if conv_page_num == DRILL_TAP:  # Settings for Drill, Drill/Tap button should set Tap
            if self.conv_drill_tap == 'drill':
                self._update_drill_through_hole_hint_label(self.drill_dro_list['drill_tool_num_dro'].get_text(),
                                                           self.drill_dro_list['drill_z_end_dro'].get_text(),
                                                           self.drill_through_hole_hint_label)            
                
                self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
                self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
                self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(False)
                self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(False)
                self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(False)
            else:  # self.conv_drill_tap == 'tap':
                self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
                self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(False)
                self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(False)
                self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(False)
                self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(False)

        if conv_page_num == THREAD:
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(False)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(False)
            self.conv_dro_list['conv_max_spindle_rpm_dro'].set_sensitive(False)

    # startup actions for each conversational page, such as graying out DROs that don't apply
    def on_conversational_notebook_switch_page(self, notebook, page, conv_page_num ):
        self._conversational_notebook_switch_page(conv_page_num)
        

    def on_reset_tool_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.clear_tool_select_images('none.png')

        # get rid of call to write tool type saves two MDI and wait_complete - noticeably faster
        #self.write_tool_type(self.status.tool_in_spindle, 'none')
        wear_offest_register = self.status.tool_in_spindle
        self.redis.hset('tool_descriptions', str(wear_offest_register), "")
        geo_offset_register = wear_offest_register + 10000
        g10_command = "G10 L1 P%d X%f Z%f I%d R%f J%d Q%d" %(wear_offest_register, 0.0, 0.0, 0, NOSE_RADIUS_STANDARD, 0, 9)
        self.issue_mdi(g10_command)
        # wait_complete is needed to prevent next offset command from stepping on previous one
        self.command.wait_complete()
        # clear out geometery offsets
        g10_command = "G10 L1 P%d X%f Z%f I%d" %(geo_offset_register, 0.0, 0.0, 0)
        self.issue_mdi(g10_command)
        self.set_image('touch_z_image', 'touch_z_black_led.png')
        self.set_image('touch_x_image', 'touch_x_black_led.png')
        self.dro_list['touch_dia_dro'].set_text("")
        self.dro_list['touch_z_dro'].set_text("")
        # set values in offsets table.  liststore starts at entry 0
        row = wear_offest_register - 1
        self.tool_liststore[row][1] = None
        self.tool_liststore[row][2] = self.dro_long_format % 0
        self.tool_liststore[row][3] = self.dro_long_format % 0
        self.tool_liststore[row][4] = self.dro_long_format % 0
        self.tool_liststore[row][5] = self.dro_long_format % 0
        self.tool_liststore[row][6] = self.dro_long_format % NOSE_RADIUS_STANDARD
        self.tool_liststore[row][7] = 0
        self.tool_liststore[row][8] = BLACK
        self.tool_liststore[row][9] = BLACK


    def on_set_g30_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.set_image('set_g30_image', 'set_g30_green_led.png')
        # store current position in parameters 5161, 5163
        self.issue_mdi('g30.1')
        # force mode switch to synch interp and write out var file
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.gremlin.load()


    def on_goto_g30_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.issue_mdi('G30 Z #5422')


    def on_touch_z_dro_activate(self, widget, data=None):
        valid, dro_val, error_msg = self.conversational.validate_z_point(self.dro_list['touch_z_dro'])
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        self.dro_list['touch_z_dro'].set_text(self.dro_long_format % dro_val)
        self.set_image('touch_z_image', 'touch_z_highlight.png')
        # next line not working as it should.
        self.button_list['touch_z'].grab_focus()

    def on_touch_z_button_release_event(self, widget, data=None):
        self.notebook_tool_fixed.move(self.button_list['touch_z'], self.button_list['touch_z'].x, self.button_list['touch_z'].y)
        valid, dro_val, error_msg = self.conversational.validate_z_point(self.dro_list['touch_z_dro'])
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        # the G10 command wants values in mm if G21, but actual_postion and g5x_offsets are in machine units (in.)
        # so we take the sup value and turn it into machine units, then send the offset command in g20/21 units
        supplemental_offset = dro_val / self.mm_inch_scalar
        wear_offset_register = self.status.tool_in_spindle
        geo_offset_register = wear_offset_register + 10000
        z_offset = self.status.actual_position[2] - self.status.g5x_offset[2] - supplemental_offset
        z_offset = z_offset * self.mm_inch_scalar
        # clear wear offsets for this tool
        self.issue_g10('Z', wear_offset_register, 0.0)
        self.issue_g10('Z', geo_offset_register, z_offset)
        self.dro_list['touch_z_dro'].set_text(self.dro_long_format % dro_val)

    def on_touch_z_key_press_event(self, widget, event):
        # Return or Enter key valid.  couldn't find keysyms constant for Enter key.
        if event.keyval == gtk.keysyms.Return or event.keyval == 65421:
            self.on_touch_z_button_release_event(self, widget)
            self.window.set_focus(None)
        elif event.keyval == gtk.keysyms.Escape:
            self.set_image('touch_z_image', 'touch_z_black_led.png')
            self.window.set_focus(None)
        return True


    def on_touch_dia_dro_activate(self, widget, data=None):
        valid, dro_val, error_msg = self.conversational.validate_dia_val(self.dro_list['touch_dia_dro'])
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        self.dro_list['touch_dia_dro'].set_text(self.dro_long_format % dro_val)
        self.set_image('touch_x_image', 'touch_x_highlight.png')
        self.button_list['touch_x'].grab_focus()

    def on_touch_x_button_release_event(self, widget, data=None):
        self.notebook_tool_fixed.move(self.button_list['touch_x'], self.button_list['touch_x'].x, self.button_list['touch_x'].y)
        wear_offset_register = self.status.tool_in_spindle
        geo_offset_register = wear_offset_register + 10000
        valid, dro_val, error_msg = self.conversational.validate_dia_val(self.dro_list['touch_dia_dro'])
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        # convert supplemental offset to machine units
        supplemental_dia_offset = dro_val / self.mm_inch_scalar
        # tool offset should be machine pos - work offset, with adjustment for dia offset in radius meas.
        x_offset = (self.status.actual_position[0] - self.status.g5x_offset[0] - (supplemental_dia_offset / 2))
        # offset command needs to be in current g20/21 units, so convert back.
        x_offset = x_offset * self.mm_inch_scalar * 2
        # clear out wear offset register
        self.issue_g10('X', wear_offset_register, 0.0)
        # set geometery offsets
        self.issue_g10('X', geo_offset_register, x_offset)
        self.dro_list['touch_dia_dro'].set_text(self.dro_long_format % dro_val)
        # complain if negative for a front tool post tool
        tool_type = self.get_tool_type(wear_offset_register)
        if 'ftp' in tool_type:
            if not 'id' in tool_type:
                if dro_val > 0:
                    self.error_handler.write("Warning: Diameter touch off for front tool post tools should be expressed as negative values.\n", ALARM_LEVEL_LOW)

    def on_touch_x_key_press_event(self, widget, event):
        if event.keyval == gtk.keysyms.Return or event.keyval == 65421:
            self.on_touch_x_button_release_event(self, widget)
            self.window.set_focus(None)
        elif event.keyval == gtk.keysyms.Escape:
            self.set_image('touch_x_image', 'touch_x_black_led.png')
            self.window.set_focus(None)
        return True

    def on_touch_dia_dro_gets_focus(self, widget, data=None):
        if self.moving(): return
        self.set_image('touch_x_image', 'touch_x_black_led.png')
        widget.masked = True
        if self.touchscreen_enabled:
            keypad = numpad.numpad_popup(widget)
            self.window.set_focus(None)
            return True


    def on_touch_z_dro_gets_focus(self, widget, data=None):
        if self.moving(): return
        self.set_image('touch_z_image', 'touch_z_black_led.png')
        widget.masked = True
        if self.touchscreen_enabled:
            keypad = numpad.numpad_popup(widget)
            self.window.set_focus(None)
            return True

    def on_turret_fwd_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        valid, tool_number, err_msg = self.conversational.validate_tool_number(self.dro_list['tool_dro'])
        if not valid:
            self.error_handler.write(err_msg, ALARM_LEVEL_LOW)
            return
        if self.status.tool_in_spindle == tool_number:
            tool_number += 1
            if tool_number > 8:
                tool_number = 1
        self.issue_toolchange_command(tool_number)
        self.set_image('turret_fwd_image', 'turret-FWD.png')
        self.window.set_focus(None)

    def on_turret_fwd_key_press_event(self, widget, event):
        if event.keyval == gtk.keysyms.Return or event.keyval == 65421:
            valid, tool_number, err_msg = self.conversational.validate_tool_number(self.dro_list['tool_dro'])
            if not valid:
                self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
                return
            self.issue_toolchange_command(tool_number)
            self.set_image('turret_fwd_image', 'turret-FWD.png')
            self.window.set_focus(None)
        elif event.keyval == gtk.keysyms.Escape:
            self.set_image('turret_fwd_image', 'turret-FWD.png')
            self.window.set_focus(None)
        return True



    def on_tool_dro_activate(self, widget):
        # get DRO value
        valid, tool_number, err_msg = self.conversational.validate_tool_number(widget)
        if not valid:
            self.error_handler.write(err_msg, ALARM_LEVEL_LOW)
            return
        # if we're using a turret changer, prelight 'turret_FWD' button if the rquested tool is in the turret
        if self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_ATC_8POSITION:
            if tool_number < 9:
                # tool is 1 - 8, in turret, so prelight turret_fwd
                self.set_image('turret_fwd_image', 'turret-FWD-highlighted.png')
                self.button_list['turret_fwd'].grab_focus()
                return
            '''elif tool_number > 99 and tool_number < 10000:
                if math.floor(tool_number) < 9:
                    # calls with wear offsets - e.g. t0202
                    self.set_image('turret_fwd_image', 'turret-FWD-highlighted.png')
                    self.button_list['turret_fwd'].grab_focus()
                    return'''

        self.issue_toolchange_command(tool_number)
        self.window.set_focus(None)



    # wear offsets, nose radius, and tip orient table callbacks

    def on_tool_description_column_edited(self, cell, row, tool_description, model ):
        tool_number = int(row) + 1
        model[row][1] = tool_description
        self.redis.hset('tool_descriptions', str(tool_number), tool_description)

    def on_x_geo_offset_col_edited(self, cell, row, value, model):
        valid, value, error_msg = self.conversational.validate_wear_offset(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        geo_offset_register = int(row) + 1 + 10000
        self.issue_g10('X', geo_offset_register, (value))
        model[row][2] = self.dro_long_format % value


    def on_z_geo_offset_col_edited( self, cell, row, value, model ):
        valid, value, error_msg = self.conversational.validate_wear_offset(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        geo_offset_register = int(row) + 1 + 10000
        self.issue_g10('Z', geo_offset_register, (value))
        model[row][3] = self.dro_long_format % value

    def on_x_offset_col_edited(self, cell, row, value, model):
        valid, value, error_msg = self.conversational.validate_wear_offset(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        wear_offset_register = int(row) + 1
        tool_number = wear_offset_register
        tool_type = self.get_tool_type(tool_number)
        text_color = self.find_wear_offset_display_color(value, tool_type)
        # LCNC wants a neg value
        self.issue_g10('X', wear_offset_register, (value))
        model[row][4] = self.dro_long_format % value
        model[row][8] = text_color


    def on_z_offset_col_edited( self, cell, row, value, model ):
        valid, value, error_msg = self.conversational.validate_wear_offset(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        wear_offset_register = int(row) + 1
        self.issue_g10('Z', wear_offset_register, (value))
        model[row][5] = self.dro_long_format % value
        if value > 0.00009:
            text_color = ORANGE
        elif value < -0.00001:
            text_color = BLUE
        else:
            text_color = BLACK
        model[row][9] = text_color

    def on_nose_radius_col_edited( self, cell, row, value, model ):
        valid, value, error_msg = self.conversational.validate_nose_radius(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        wear_offset_register = int(row) + 1
        self.issue_g10('R', wear_offset_register, (value))
        model[row][6] = self.dro_long_format % value

    def on_tip_orient_col_edited( self, cell, row, value, model ):
        valid, value, error_msg = self.conversational.validate_tip_orientation(value)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        wear_offset_register = int(row) + 1
        # can't use issue_g10 here because lcnc wants an int for the Q word
        g10_command = "G10 L1 P%d Q%f" %(wear_offset_register, int(value))
        self.issue_mdi(g10_command)
        model[row][7] = int(value)



    def clear_tool_select_images(self, image, supress_error=False):
        # supress error message on notebook page switch
        if not supress_error:
            if self.status.tool_in_spindle == 0:
                self.error_handler.write("\nMust select valid tool before setting tool parameters \n", ALARM_LEVEL_LOW)
                return

        # if the tool has no type yet, display the splash screen controls and hide the tool table
        if image == 'none.png':
            splash = True
            image = 'chuck.png'
        else:
            splash = False

        # prevent/allow eventboxes from receiving events depending whether we are in splash screen or touch-off screen
        for name, eventbox in self.tool_touch_eventbox_list.iteritems():
            eventbox.set_sensitive(splash)
            # fix problem where center drill image grabbed touch_x button click events
            eventbox.set_visible(splash)

        if self.is_rapidturn:
            for name, eventbox in self.tool_touch_eventbox_list.iteritems():
                if ('front' in name) and not (('boring' in name) or ('id' in name)):
                    eventbox.set_sensitive(False)

        self.button_list['reset_tool'].set_sensitive(not splash)

        # change to appropriate background image
        self.set_image('tool_touch_chuck', image)

        # show or hide images and dros
        if splash:
            self.button_list['reset_tool'].hide()
            self.image_list['reset_tool_image'].hide()
            if not self.is_rapidturn: self.ftp_text.show()
            self.rtp_text.show()
            for name, image in self.tool_image_list.iteritems():
                if self.is_rapidturn:
                    if ('front' in name) and not (('boring' in name) or ('id' in name)):
                        image.hide()
                        continue
                image.show()
        else:
            self.button_list['reset_tool'].show()
            self.image_list['reset_tool_image'].show()
            for name, image in self.tool_image_list.iteritems():
                image.hide()





    # -------------------------------------------------------------------------------------------------
    # conversational tab callbacks
    # -------------------------------------------------------------------------------------------------

    # generate gcode helper

    def generate_gcode(self, tab_text=None):
        active_child = self.conv_notebook.get_nth_page(self.conv_notebook.get_current_page())
        active_tab_text = self.conv_notebook.get_tab_label_text(active_child) if tab_text is None else tab_text
        valid = False

        ini_file_name = sys.argv[2]
        inifile = linuxcnc.ini(ini_file_name)

        valid = False
        if active_tab_text == "OD Turn":
            (valid, gcode_output_list) = self.conversational.generate_od_turn_code(self.conv_dro_list, self.od_turn_dro_list)
            self.save_conv_paramters(self.od_turn_dro_list)

        if active_tab_text == "ID Turn":
            if self.conv_id_basic_ext == "basic":
                (valid, gcode_output_list) = self.conversational.generate_id_basic_code(self.conv_dro_list, self.id_basic_dro_list)
                self.save_conv_paramters(self.id_basic_dro_list)
            else:
                (valid, gcode_output_list) = self.conversational.generate_id_turn_code(self.conv_dro_list, self.id_turn_dro_list)
                self.save_conv_paramters(self.id_turn_dro_list)

        if active_tab_text == "Face":
            (valid, gcode_output_list) = self.conversational.generate_face_code(self.conv_dro_list, self.face_dro_list)
            self.save_conv_paramters(self.face_dro_list)

        if active_tab_text == "Chamfer":
            if self.conv_chamfer_radius == "chamfer":
                if self.conv_chamfer_od_id == 'od':
                    (valid, gcode_output_list) = self.conversational.generate_chamfer_code(self.conv_dro_list, self.corner_chamfer_od_dro_list, self.conv_chamfer_od_id)
                    self.save_conv_paramters(self.corner_chamfer_od_dro_list)
                else:  # self.conv_chamfer_od_id == 'id'
                    (valid, gcode_output_list) = self.conversational.generate_chamfer_code(self.conv_dro_list, self.corner_chamfer_id_dro_list, self.conv_chamfer_od_id)
                    self.save_conv_paramters(self.corner_chamfer_id_dro_list)
            else:  # self.conv_chamfer_radius == "radius"
                if self.conv_chamfer_od_id == 'od':
                    (valid, gcode_output_list) = self.conversational.generate_radius_code(self.conv_dro_list, self.corner_radius_od_dro_list, self.conv_chamfer_od_id)
                    self.save_conv_paramters(self.corner_radius_od_dro_list)
                else:  # self.conv_chamfer_od_id == 'id'
                    (valid, gcode_output_list) = self.conversational.generate_radius_code(self.conv_dro_list, self.corner_radius_id_dro_list, self.conv_chamfer_od_id)
                    self.save_conv_paramters(self.corner_radius_id_dro_list)

        if active_tab_text == "Groove/Part":
            if self.conv_groove_part == "groove":
                (valid, gcode_output_list) = self.conversational.generate_groove_code(self.conv_dro_list, self.groove_dro_list, inifile)
                self.save_conv_paramters(self.groove_dro_list)
            else:
                (valid, gcode_output_list) = self.conversational.generate_part_code(self.conv_dro_list, self.part_dro_list, inifile)
                self.save_conv_paramters(self.part_dro_list)

        if "Drill" in active_tab_text:
            if self.conv_drill_tap == "drill":
                (valid, gcode_output_list) = self.conversational.generate_drill_code(self.conv_dro_list, self.drill_dro_list)
                self.save_conv_paramters(self.drill_dro_list)
            else:
                pitch = self.tap_dro_list['tap_pitch_dro'].get_text()
                max_vel = self.inifile.find("AXIS_2", "MAX_VELOCITY")
                (valid, gcode_output_list) = self.conversational.generate_tap_code(self.conv_dro_list,
                                                                                   self.tap_dro_list,
                                                                                   pitch,
                                                                                   max_vel)
                self.save_conv_paramters(self.tap_dro_list)

        if active_tab_text == "Thread":
            pitch = self.thread_dro_list['thread_pitch_dro'].get_text()
            max_vel = self.inifile.find("AXIS_2", "MAX_VELOCITY")
            (valid, gcode_output_list) = self.conversational.generate_thread_code(self.conv_dro_list,
                                                                                  self.thread_dro_list,
                                                                                  self.conv_thread_ext_int,
                                                                                  inifile,
                                                                                  self.conv_thread_note,
                                                                                  pitch,
                                                                                  max_vel)
            self.save_conv_paramters(self.thread_dro_list)

        return valid, gcode_output_list




    # -------------------------------------------------------------------------------------------------
    # general setting DROs and "post to file/mdi" callbacks
    # -------------------------------------------------------------------------------------------------


    def on_post_to_file_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)

        # generate the code
        valid, gcode_output_list = self.generate_gcode()
        if valid:
            self.save_title_dro()                        
            self.post_to_file(self.conv_dro_list['conv_title_dro'].get_text(), gcode_output_list)


    def on_append_to_file_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)

        # generate the code:
        valid, gcode_output_list = self.generate_gcode()
        if not valid:
            return

        path = self.last_used_save_as_path + os.path.sep
        dialog = tormach_file_util.append_to_file_popup(path, self.touchscreen_enabled)
        dialog.run()

        response = dialog.response
        self.last_used_save_as_path = dialog.current_directory
        path = dialog.get_path()

        if response == gtk.RESPONSE_OK:
            # add the code to an exixting file.
            ngc_file = open(path, "r+")
            if ngc_file:
                # strip out old M30
                old_code = ngc_file.readlines()
                old_code = [i if ('M30' not in i.upper()) else '' for i in old_code]
                ngc_file.seek (0, 0)
                for ngc_line in old_code:
                    ngc_file.write(ngc_line)
                for ngc_line in gcode_output_list:
                    ngc_file.write(ngc_line + '\r\n')
                ngc_file.close()
                self.load_gcode_file(path)

        dialog.destroy()

    def on_conv_title_dro_gets_focus(self, widget, data=None):
        if self.touchscreen_enabled:
            keypad = numpad.numpad_popup(widget, True)
            self.window.set_focus(None)
            return True

    def on_conv_dro_gets_focus(self, widget, data=None):
        # really the button release event
        widget.prev_val = widget.get_text()
        widget.select_region(0, -1)
        if self.touchscreen_enabled:
            keypad = numpad.numpad_popup(widget)
            widget.select_region(0, 0)
            self.window.set_focus(None)
            
    def on_conv_dro_focus_in_event(self, widget, data=None):
        widget.prev_val = widget.get_text()            

    def on_entry_loses_focus(self, widget, data=None):
        # get rid of text highlight if you click out of a dro that has highlighted text
        widget.select_region(0, 0)
        return False


    # focus passing for conversational common DROs
    def on_conv_title_dro_activate(self, widget, data=None):
        self.conv_title = widget.get_text()
        self.conv_dro_list['conv_work_offset_dro'].grab_focus()

    def on_conv_work_offset_dro_activate(self, widget, data=None):
        (valid, work_offset, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(work_offset)
        current_page = self.conv_notebook.get_current_page()
        if current_page == DRILL_TAP:
            if self.conv_drill_tap == 'drill':
                self.conv_dro_list['conv_rough_fpr_dro'].grab_focus()
                return
            else:  # == 'tap'
                self.tap_dro_list['tap_tool_num_dro'].grab_focus()
                return

        elif current_page == THREAD:
            self.thread_dro_list['thread_tool_num_dro'].grab_focus()
            return

        elif current_page == GROOVE_PART:
            if self.conv_groove_part == 'part':
                self.part_dro_list['part_tool_num_dro'].grab_focus()
                return
            else:
                self.conv_dro_list['conv_rough_sfm_dro'].grab_focus()
                return
        else:
            self.conv_dro_list['conv_rough_sfm_dro'].grab_focus()

        return

    def on_conv_rough_sfm_dro_activate(self, widget, data=None):
        (valid, surface_speed, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % surface_speed)
        self.conv_dro_list['conv_finish_sfm_dro'].grab_focus()
        return

    def on_conv_finish_sfm_dro_activate(self, widget, data=None):
        (valid, surface_speed, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % surface_speed)
        self.conv_dro_list['conv_max_spindle_rpm_dro'].grab_focus()
        return

    def on_conv_max_spindle_rpm_dro_activate(self, widget, data=None):
        (valid, max_rpm, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % max_rpm)
        current_page = self.conv_notebook.get_current_page()
        if current_page == GROOVE_PART:
            if self.conv_groove_part == 'part':
                self.part_dro_list['part_tool_num_dro'].grab_focus()
                return
            else:
                self.conv_dro_list['conv_rough_fpr_dro'].grab_focus()
                return
        else:
            self.conv_dro_list['conv_rough_fpr_dro'].grab_focus()

        return

    def on_conv_rough_fpr_dro_activate(self, widget, data=None):
        (valid, feedrate, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % feedrate)
        current_page = self.conv_notebook.get_current_page()
        if current_page == DRILL_TAP:
            if self.conv_drill_tap == 'drill':
                self.drill_dro_list['drill_tool_num_dro'].grab_focus()
                return
        else:
            self.conv_dro_list['conv_finish_fpr_dro'].grab_focus()

        return

    def on_conv_finish_fpr_dro_activate(self, widget, data=None):
        (valid, feedrate, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % feedrate)
        active_child = self.conv_notebook.get_nth_page(self.conv_notebook.get_current_page())
        active_tab_text = self.conv_notebook.get_tab_label_text(active_child)

        if active_tab_text == "OD Turn":
            self.od_turn_dro_list['od_turn_tool_num_dro'].grab_focus()
            return

        if active_tab_text == "ID Turn":
            if self.conv_id_basic_ext == 'basic':
                self.id_basic_dro_list['id_basic_tool_num_dro'].grab_focus()
                return
            else:
                self.id_turn_dro_list['id_turn_tool_num_dro'].grab_focus()
                return

        if active_tab_text == "Face":
            self.face_dro_list['face_tool_num_dro'].grab_focus()
            return

        if active_tab_text == "Chamfer":
            if self.conv_chamfer_radius == 'chamfer' and self.conv_chamfer_od_id == 'od':
                self.corner_chamfer_od_dro_list['corner_chamfer_od_tool_num_dro'].grab_focus()
                return
            elif self.conv_chamfer_radius == 'chamfer' and self.conv_chamfer_od_id == 'id':
                self.corner_chamfer_id_dro_list['corner_chamfer_id_tool_num_dro'].grab_focus()
                return
            elif self.conv_chamfer_radius == 'radius' and self.conv_chamfer_od_id == 'od':
                self.corner_radius_od_dro_list['corner_radius_od_tool_num_dro'].grab_focus()
                return
            else:  # self.conv_chamfer_radius == 'radius' and self.conv_chamfer_od_id == 'id'
                self.corner_radius_id_dro_list['corner_radius_id_tool_num_dro'].grab_focus()
                return

        if active_tab_text == "Groove/Part":
            if self.conv_groove_part == 'groove':
                self.groove_dro_list['groove_tool_num_dro'].grab_focus()
                return

        self.conv_dro_list['conv_title_dro'].grab_focus()
        return
    
    def next_conv_dro(self, conv_list, conv_from):
        conv_list_len = len(conv_list)
        for n in range(conv_list_len):
            if conv_list[n] == conv_from:
                return conv_list[n+1] if (n+1) < conv_list_len else conv_list[0]
        return 'nowhere'
        #list_index = 0
        #go_here = 'nowhere'
        #for list_id in conv_list:
            #if list_id == conv_from:
                #if list_index >= conv_count - 1:  # wrap to the beginning of list
                    #go_here = conv_list[0]
                    #break
                #else:
                    #go_here =  conv_list[list_index + 1]
                    #break
            #list_index += 1
        #return go_here

    # -------------------------------------------------------------------------------------------------
    # OD Turn DROs
    # -------------------------------------------------------------------------------------------------

    def on_od_turn_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_tool_num_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        return

    def on_od_turn_stock_dia_dro_activate(self, widget, data=None):
        (valid, dia, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % dia)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_stock_dia_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        return

    def on_od_turn_final_dia_dro_activate(self, widget, data=None):
        (valid, dia, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % dia)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_final_dia_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_z_start_dro'].grab_focus()
        return

    def on_od_turn_z_start_dro_activate(self, widget, data=None):
        (valid, z_point, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % z_point)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_z_start_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_z_end_dro'].grab_focus()
        return

    def on_od_turn_z_end_dro_activate(self, widget, data=None):
        (valid, z_point, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % z_point)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_z_end_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_fillet_dro'].grab_focus()
        return

    def on_od_turn_rough_doc_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_rough_doc_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_stock_dia_dro'].grab_focus()
        return

    def on_od_turn_finish_doc_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_finish_doc_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_rough_doc_dro'].grab_focus()
        return

    def on_od_turn_tc_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_tc_dro')
        self.od_turn_dro_list[where_to].grab_focus()
        #self.od_turn_dro_list['od_turn_finish_doc_dro'].grab_focus()
        return

    def on_od_turn_fillet_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.od_focus_list, 'od_turn_fillet_dro')
        self.od_turn_dro_list[where_to].grab_focus()

        return

    # -------------------------------------------------------------------------------------------------
    # ID Turn DROs
    # -------------------------------------------------------------------------------------------------

    def toggle_id_basic_extended_button(self):
        self.save_title_dro()        
        if self.conv_id_basic_ext == 'basic':
            self.set_image('id_basic_extended_btn_image', 'id_extended_highlight.jpg')
            self.set_image('conv_id_turn_background', 'lathe_id_extended.svg')

            # turn ID Basic DROs OFF
            self.id_basic_dro_list['id_basic_tool_num_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_z_start_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_z_end_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_final_dia_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_pilot_dia_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_rough_doc_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_finish_doc_dro'].set_visible(False)
            self.id_basic_dro_list['id_basic_tc_dro'].set_visible(False)

            # turn ID Extended DROs ON
            self.id_turn_dro_list['id_turn_tool_num_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_z_start_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_z_end_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_final_dia_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_pilot_dia_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_fillet_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_pilot_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_rough_doc_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_finish_doc_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_tc_dro'].set_visible(True)
            self.id_turn_dro_list['id_turn_face_doc_dro'].set_visible(True)

            # turn ID Basic labels (text) OFF

            # turn ID Extended labels(text) ON
            self.id_turn_fillet_label.set_visible(True)
            self.id_turn_pilot_end_label.set_visible(True)
            self.id_turn_face_doc_label.set_visible(True)
            self.id_turn_tool_text1_label.set_visible(True)
            self.id_turn_roughing_label.set_visible(True)

            # these conv_DROs are also changed in on_conversational_notebook_switch_page
            #self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)

            self.conv_id_basic_ext = 'extended'

        else:
            self.set_image('id_basic_extended_btn_image', 'id_basic_highlight.jpg')
            self.set_image('conv_id_turn_background', 'lathe_id_basic.svg')

            # turn ID Extended DROs OFF
            self.id_turn_dro_list['id_turn_tool_num_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_z_start_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_z_end_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_final_dia_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_pilot_dia_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_fillet_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_pilot_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_rough_doc_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_finish_doc_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_tc_dro'].set_visible(False)
            self.id_turn_dro_list['id_turn_face_doc_dro'].set_visible(False)

            # turn ID Basic DROs ON
            self.id_basic_dro_list['id_basic_tool_num_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_z_start_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_z_end_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_final_dia_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_pilot_dia_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_rough_doc_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_finish_doc_dro'].set_visible(True)
            self.id_basic_dro_list['id_basic_tc_dro'].set_visible(True)

            # turn ID Extended labels(text) OFF
            self.id_turn_fillet_label.set_visible(False)
            self.id_turn_pilot_end_label.set_visible(False)
            self.id_turn_face_doc_label.set_visible(False)
            self.id_turn_tool_text1_label.set_visible(False)
            self.id_turn_roughing_label.set_visible(False)

            # turn ID Basic labels(text) ON
            self.conv_id_basic_ext = 'basic'
        self.load_title_dro()

    def on_id_basic_extended_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        # toggle button state between basic and extended
        self.toggle_id_basic_extended_button()


    # ID Basic DROs
    def on_id_basic_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_tool_num_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_tc_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_finish_doc_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_rough_doc_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_z_end_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_z_start_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_final_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_final_dia_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return

    def on_id_basic_pilot_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_basic_focus_list, 'id_basic_pilot_dia_dro')
        self.id_basic_dro_list[where_to].grab_focus()
        return



    # ID Extended DROs
    def on_id_turn_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_tool_num_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        return

    def on_id_turn_pilot_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_pilot_dia_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_pilot_dro'].grab_focus()
        return

    def on_id_turn_final_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_final_dia_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_pilot_dia_dro'].grab_focus()
        return

    def on_id_turn_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_z_start_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_z_end_dro'].grab_focus()
        return

    def on_id_turn_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_z_end_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_final_dia_dro'].grab_focus()
        return

    def on_id_turn_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_rough_doc_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_finish_doc_dro'].grab_focus()
        return

    def on_id_turn_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_finish_doc_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_tc_dro'].grab_focus()
        return

    def on_id_turn_tc_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_tc_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_face_doc_dro'].grab_focus()
        return

    def on_id_turn_fillet_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_fillet_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_rough_doc_dro'].grab_focus()
        return

    def on_id_turn_face_doc_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_face_doc_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_z_start_dro'].grab_focus()
        return

    def on_id_turn_pilot_dro_activate(self, widget, data=None):
        (valid, doc, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % doc)
        where_to = self.next_conv_dro(self.id_focus_list, 'id_turn_pilot_dro')
        self.id_turn_dro_list[where_to].grab_focus()
        #self.id_turn_dro_list['id_turn_fillet_dro'].grab_focus()
        #self.window.set_focus(None)
        return


    # -------------------------------------------------------------------------------------------------
    # Facing DROs
    # -------------------------------------------------------------------------------------------------

    def on_face_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_tool_num_dro')
        self.face_dro_list[where_to].grab_focus()
        return

    def on_face_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_z_end_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_stock_dia_dro'].grab_focus()
        return

    def on_face_stock_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_stock_dia_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_x_end_dro'].grab_focus()
        return

    def on_face_x_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_x_end_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_z_start_dro'].grab_focus()
        return

    def on_face_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_z_start_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_tc_dro'].grab_focus()
        return

    def on_face_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_tc_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_rough_doc_dro'].grab_focus()
        return

    def on_face_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_rough_doc_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_finish_doc_dro'].grab_focus()
        return

    def on_face_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.face_focus_list, 'face_finish_doc_dro')
        self.face_dro_list[where_to].grab_focus()
        #self.face_dro_list['face_z_end_dro'].grab_focus()
        #self.window.set_focus(None)
        return


    # -------------------------------------------------------------------------------------------------
    # Chamfer DROs
    # -------------------------------------------------------------------------------------------------

    def corner_clear(self):
        # turn Corner DROs OFF
        for droi in self.corner_chamfer_od_dro_list:
            self.corner_chamfer_od_dro_list[droi].set_visible(False)
        for droi in self.corner_chamfer_id_dro_list:
            self.corner_chamfer_id_dro_list[droi].set_visible(False)
        for droi in self.corner_radius_od_dro_list:
            self.corner_radius_od_dro_list[droi].set_visible(False)
        for droi in self.corner_radius_id_dro_list:
            self.corner_radius_id_dro_list[droi].set_visible(False)

        # turn Corner dynamic labels OFF
        self.corner_chamfer_od_angle_label.set_visible(False)
        self.corner_chamfer_id_angle_label.set_visible(False)
        self.corner_chamfer_od_rough_doc_label.set_visible(False)
        self.corner_chamfer_id_rough_doc_label.set_visible(False)
        self.corner_chamfer_od_finish_doc_label.set_visible(False)
        self.corner_chamfer_id_finish_doc_label.set_visible(False)
        self.corner_chamfer_od_tc_label.set_visible(False)
        self.corner_chamfer_id_tc_label.set_visible(False)

        self.corner_radius_od_rough_doc_label.set_visible(False)
        self.corner_radius_id_rough_doc_label.set_visible(False)
        self.corner_radius_od_finish_doc_label.set_visible(False)
        self.corner_radius_id_finish_doc_label.set_visible(False)
        self.corner_radius_od_tc_label.set_visible(False)
        self.corner_radius_id_tc_label.set_visible(False)

        self.corner_od_x_label.set_visible(False)
        self.corner_id_x_label.set_visible(False)

    def toggle_chamfer_radius_button(self):
        self.corner_clear()
        self.save_title_dro()

        # toggle button state between chamfer and corner radius
        if self.conv_chamfer_radius == 'chamfer':  # toggle to radius mode
            self.set_image('corner_chamfer_radius_btn_image', 'chamfer_radius_rad_highlight.jpg')

            if self.conv_chamfer_od_id == 'od':
                # Set tab graphics to Radius OD
                self.set_image('conv_corner_background', 'lathe_corner_radius_od.svg')

                # turn Radius OD labels ON
                self.corner_radius_od_rough_doc_label.set_visible(True)
                self.corner_radius_od_finish_doc_label.set_visible(True)
                self.corner_radius_od_tc_label.set_visible(True)

                self.corner_od_x_label.set_visible(True)

                # turn Radius OD DROs ON
                for droi in self.corner_radius_od_dro_list:
                    self.corner_radius_od_dro_list[droi].set_visible(True)

            else: # self.conv_chamfer_od_id == 'id'
                # Set tab graphics to Chamfer OD
                self.set_image('conv_corner_background', 'lathe_corner_radius_id.svg')

                # turn Radius ID labels ON
                self.corner_radius_id_rough_doc_label.set_visible(True)
                self.corner_radius_id_finish_doc_label.set_visible(True)
                self.corner_radius_id_tc_label.set_visible(True)

                self.corner_id_x_label.set_visible(True)

                # turn Corner ID DROs ON
                for droi in self.corner_radius_id_dro_list:
                    self.corner_radius_id_dro_list[droi].set_visible(True)

            self.conv_chamfer_radius = 'radius'

        else:  # toggle to chamfer mode
            self.set_image('corner_chamfer_radius_btn_image', 'chamfer_radius_cham_highlight.jpg')

            if self.conv_chamfer_od_id == 'od':
                # Set tab graphics to Chamfer OD
                self.set_image('conv_corner_background', 'lathe_corner_chamfer_od.svg')

                # turn Chamfer OD labels ON
                self.corner_chamfer_od_angle_label.set_visible(True)
                self.corner_chamfer_od_rough_doc_label.set_visible(True)
                self.corner_chamfer_od_finish_doc_label.set_visible(True)
                self.corner_chamfer_od_tc_label.set_visible(True)

                self.corner_od_x_label.set_visible(True)

                # turn Chamfer OD DROs ON
                for droi in self.corner_chamfer_od_dro_list:
                    self.corner_chamfer_od_dro_list[droi].set_visible(True)

            else: # self.conv_chamfer_od_id == 'id'
                # Set tab graphics to Chamfer ID
                self.set_image('conv_corner_background', 'lathe_corner_chamfer_id.svg')

                # turn Chamfer ID labels ON
                self.corner_chamfer_id_angle_label.set_visible(True)
                self.corner_chamfer_id_rough_doc_label.set_visible(True)
                self.corner_chamfer_id_finish_doc_label.set_visible(True)
                self.corner_chamfer_id_tc_label.set_visible(True)

                self.corner_id_x_label.set_visible(True)

                # turn Chamfer ID DROs ON
                for droi in self.corner_chamfer_id_dro_list:
                    self.corner_chamfer_id_dro_list[droi].set_visible(True)
            self.conv_chamfer_radius = 'chamfer'
        self.load_title_dro()
        
    def on_corner_chamfer_radius_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.toggle_chamfer_radius_button()
        return

    def toggle_id_od_button(self):
        self.corner_clear()
        self.save_title_dro()

        # toggle button state between ID and OD
        if self.conv_chamfer_od_id == 'od':  # toggle to ID mode
            self.set_image('corner_id_od_btn_image', 'id_od_button.jpg')

            if self.conv_chamfer_radius == 'chamfer':
                # Set tab graphics to Chamfer ID
                self.set_image('conv_corner_background', 'lathe_corner_chamfer_id.svg')

                # turn Chamfer ID labels ON
                self.corner_chamfer_id_angle_label.set_visible(True)
                self.corner_chamfer_id_rough_doc_label.set_visible(True)
                self.corner_chamfer_id_finish_doc_label.set_visible(True)
                self.corner_chamfer_id_tc_label.set_visible(True)

                self.corner_id_x_label.set_visible(True)

                # turn Chamfer ID DROs ON
                for droi in self.corner_chamfer_id_dro_list:
                    self.corner_chamfer_id_dro_list[droi].set_visible(True)

            else: # self.conv_chamfer_radius == 'radius'
                # Set tab graphics to Chamfer ID
                self.set_image('conv_corner_background', 'lathe_corner_radius_id.svg')

                # turn Radius ID labels ON
                self.corner_radius_id_rough_doc_label.set_visible(True)
                self.corner_radius_id_finish_doc_label.set_visible(True)
                self.corner_radius_id_tc_label.set_visible(True)

                self.corner_id_x_label.set_visible(True)

                # turn Radius ID DROs ON
                for droi in self.corner_radius_id_dro_list:
                    self.corner_radius_id_dro_list[droi].set_visible(True)

            self.conv_chamfer_od_id = 'id'

        else:  # self.conv_chamfer_od_id == 'id' # toggle to OD mode
            self.set_image('corner_id_od_btn_image', 'od_id_button.jpg')

            if self.conv_chamfer_radius == 'chamfer':
                # Set tab graphics to Chamfer OD
                self.set_image('conv_corner_background', 'lathe_corner_chamfer_od.svg')

                # turn Chamfer OD labels ON
                self.corner_chamfer_od_angle_label.set_visible(True)
                self.corner_chamfer_od_rough_doc_label.set_visible(True)
                self.corner_chamfer_od_finish_doc_label.set_visible(True)
                self.corner_chamfer_od_tc_label.set_visible(True)

                self.corner_od_x_label.set_visible(True)

                # turn Chamfer OD DROs ON
                for droi in self.corner_chamfer_od_dro_list:
                    self.corner_chamfer_od_dro_list[droi].set_visible(True)

            else: # self.conv_chamfer_radius == 'radius'
                # Set tab graphics to Corner Radius OD
                self.set_image('conv_corner_background', 'lathe_corner_radius_od.svg')

                # turn Radius OD labels ON
                self.corner_radius_od_rough_doc_label.set_visible(True)
                self.corner_radius_od_finish_doc_label.set_visible(True)
                self.corner_radius_od_tc_label.set_visible(True)

                self.corner_od_x_label.set_visible(True)

                # turn Radius OD DROs ON
                for droi in self.corner_radius_od_dro_list:
                    self.corner_radius_od_dro_list[droi].set_visible(True)

            self.conv_chamfer_od_id = 'od'
        self.load_title_dro()
        
    def on_corner_id_od_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.toggle_id_od_button()


    def on_corner_chamfer_od_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_tool_num_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_tool_num_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_tool_num_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_tool_num_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return


    def on_corner_chamfer_od_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to chamfer_z_width DRO
        try:
            z_end = float(self.corner_chamfer_od_dro_list['corner_chamfer_od_z_end_dro'].get_text())
        except:
            z_end = 0

        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_z_start_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to chamfer_z_width DRO
        try:
            z_end = float(self.corner_chamfer_id_dro_list['corner_chamfer_id_z_end_dro'].get_text())
        except:
            z_end = 0

        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_z_start_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to chamfer_od_z_width DRO
        try:
            z_start = float(self.corner_chamfer_od_dro_list['corner_chamfer_od_z_start_dro'].get_text())
        except:
            z_start = 0

        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_z_end_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to chamfer_id_z_width DRO
        try:
            z_start = float(self.corner_chamfer_id_dro_list['corner_chamfer_id_z_start_dro'].get_text())
        except:
            z_start = 0

        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_z_end_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_od_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_od_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_id_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_id_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_angle_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_dwell_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_angle_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_angle_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_dwell_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_angle_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_rough_doc_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_rough_doc_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_finish_doc_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_finish_doc_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_od_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_od_focus_list, 'corner_chamfer_od_tc_dro')
        self.corner_chamfer_od_dro_list[where_to].grab_focus()
        return

    def on_corner_chamfer_id_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.corner_chamfer_id_focus_list, 'corner_chamfer_id_tc_dro')
        self.corner_chamfer_id_dro_list[where_to].grab_focus()
        return

    # ---- Corner Radius DROs ----
    def on_corner_radius_od_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_tool_num_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_tool_num_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to corner_od_z_width DRO
        try:
            z_end = float(self.corner_od_dro_list['corner_od_z_end_dro'].get_text())
        except:
            z_end = 0

        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_z_start_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to corner_id_z_width DRO
        try:
            z_end = float(self.corner_radius_id_dro_list['corner_radius_id_z_end_dro'].get_text())
        except:
            z_end = 0

        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_z_start_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to corner_od_z_width DRO
        try:
            z_start = float(self.corner_radius_od_dro_list['corner_radius_od_z_start_dro'].get_text())
        except:
            z_start = 0

        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_z_end_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate the width and push to corner_id_z_width DRO
        try:
            z_start = float(self.corner_radius_id_dro_list['corner_radius_id_z_start_dro'].get_text())
        except:
            z_start = 0

        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_z_end_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_od_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_od_dro_list['corner_od_rough_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_od_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_id_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_id_dro_list['corner_id_rough_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_id_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_od_dro_list['corner_od_finish_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_rough_doc_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_id_dro_list['corner_id_finish_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_rough_doc_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_od_dro_list['corner_od_tc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_finish_doc_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_id_dro_list['corner_id_tc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_finish_doc_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_od_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_od_dro_list['corner_od_z_start_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_od_focus_list, 'corner_radius_od_tc_dro')
        self.corner_radius_od_dro_list[where_to].grab_focus()
        return

    def on_corner_radius_id_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.corner_id_dro_list['corner_id_z_start_dro'].grab_focus()
        where_to = self.next_conv_dro(self.corner_radius_id_focus_list, 'corner_radius_id_tc_dro')
        self.corner_radius_id_dro_list[where_to].grab_focus()
        return


    # -------------------------------------------------------------------------------------------------
    # Groove DROs
    # -------------------------------------------------------------------------------------------------

    def toogle_groove_part_dros(self):
        self.save_title_dro()
        is_grooving = self.conv_groove_part == 'part'
        is_parting = self.conv_groove_part == 'groove'
        
        # turn Groove DROs OFF
        self.groove_dro_list['groove_tool_num_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_z_start_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_stock_dia_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_final_dia_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_z_end_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_rough_doc_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_tw_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_finish_doc_dro'].set_visible(is_grooving)
        self.groove_dro_list['groove_tc_dro'].set_visible(is_grooving)
        
        # tool DROs
        self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(is_grooving)
        self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(is_grooving)
        
        # turn Groove labels(text) OFF
        self.groove_z_end_label.set_visible(is_grooving)
        self.groove_r_doc_label.set_visible(is_grooving)
        self.groove_f_doc_label.set_visible(is_grooving)
        self.groove_tw_label.set_visible(is_grooving)

        # turn Part DROs ON
        self.part_dro_list['part_tool_num_dro'].set_visible(is_parting)
        self.part_dro_list['part_z_start_dro'].set_visible(is_parting)
        self.part_dro_list['part_stock_dia_dro'].set_visible(is_parting)
        self.part_dro_list['part_final_dia_dro'].set_visible(is_parting)
        self.part_dro_list['part_tc_dro'].set_visible(is_parting)
        self.part_dro_list['part_peck_dro'].set_visible(is_parting)
        self.part_dro_list['part_retract_dro'].set_visible(is_parting)
        self.part_dro_list['part_ebw_dro'].set_visible(is_parting)
        self.part_dro_list['part_tw_dro'].set_visible(is_parting)

        # turn Part labels(text) ON
        self.part_peck_label.set_visible(is_parting)
        self.part_retract_label.set_visible(is_parting)
        self.part_ebw_label.set_visible(is_parting)
        self.part_tw_label.set_visible(is_parting)
                            
        # set the button and the image...
        btn_image = 'groove_part_part_highlight.jpg' if self.conv_groove_part == 'groove' else 'groove_part_grv_highlight.jpg'
        back_ground_image =  'lathe_part.svg' if self.conv_groove_part == 'groove' else 'lathe_groove.svg'
        self.set_image('groove_part_btn_image', btn_image)
        self.set_image('conv_groove_part_image', back_ground_image) 
        # toggle the state...
        self.conv_groove_part = 'groove' if self.conv_groove_part == 'part' else 'part'
        
        # if the 'parting' tool width is empty ...
        # copy over the tool width from the groove data
        if self.conv_groove_part == 'part':
            text = self.part_dro_list['part_tw_dro'].get_text()
            if len(text) == 0:
                self.part_dro_list['part_tw_dro'].set_text(self.groove_dro_list['groove_tw_dro'].get_text())
            self.part_dro_list[self.part_focus_list[0]].grab_focus()
        else:
            self.groove_dro_list[self.groove_focus_list[0]].grab_focus()
        self.load_title_dro()
            
            
    def on_groove_part_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.toogle_groove_part_dros()
        return

    def on_groove_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_tool_num_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_stock_dia_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_z_start_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_stock_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_final_dia_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_stock_dia_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_final_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_z_end_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_final_dia_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_rough_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_z_end_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_rough_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_tw_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_rough_doc_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_tw_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_finish_doc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_tw_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_finish_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_tc_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_finish_doc_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    def on_groove_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        #self.groove_dro_list['groove_z_start_dro'].grab_focus()
        where_to = self.next_conv_dro(self.groove_focus_list, 'groove_tc_dro')
        self.groove_dro_list[where_to].grab_focus()
        return

    # ---- Part DROs ---- #
    def on_part_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_tool_num_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_z_start_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_stock_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_stock_dia_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_final_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_final_dia_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_tc_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_tw_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_tw_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_css_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_css_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_fpr_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_fpr_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_peck_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_peck_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_retract_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_retract_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_finish_rpm_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_finish_rpm_dro')
        self.part_dro_list[where_to].grab_focus()
        return

    def on_part_ebw_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.part_focus_list, 'part_ebw_dro')
        self.part_dro_list[where_to].grab_focus()
        return


    # -------------------------------------------------------------------------------------------------
    # Drill Tap DROs
    # -------------------------------------------------------------------------------------------------

    def toggle_drill_tap_button(self):
        self.save_title_dro()                
        if self.conv_drill_tap == 'drill':
            self.set_image('drill_tap_btn_image', 'drill_tap_tap_highlight.jpg')
            self.set_image('conv_drill_tap_image', 'lathe_tap.svg')

            # turn Drill DROs OFF
            self.drill_dro_list['drill_tool_num_dro'].set_visible(False)
            self.drill_dro_list['drill_z_start_dro'].set_visible(False)
            self.drill_dro_list['drill_tc_dro'].set_visible(False)
            self.drill_dro_list['drill_peck_dro'].set_visible(False)
            self.drill_dro_list['drill_z_end_dro'].set_visible(False)
            self.drill_dro_list['drill_spindle_rpm_dro'].set_visible(False)
            self.drill_dro_list['drill_dwell_dro'].set_visible(False)

            # turn Tap DROs ON
            self.tap_dro_list['tap_tool_num_dro'].set_visible(True)
            self.tap_dro_list['tap_z_start_dro'].set_visible(True)
            self.tap_dro_list['tap_tc_dro'].set_visible(True)
            self.tap_dro_list['tap_z_end_dro'].set_visible(True)
            self.tap_dro_list['tap_spindle_rpm_dro'].set_visible(True)
            self.tap_dro_list['tap_tpu_dro'].set_visible(True)
            self.tap_dro_list['tap_pitch_dro'].set_visible(True)
            self.tap_dro_list['tap_peck_dro'].set_visible(True)

            # turn Drill labels(text) OFF
            self.drill_z_end_label.set_visible(False)
            self.drill_dwell_label.set_visible(False)

            # turn Tap labels(text) ON
            self.tap_z_end_label.set_visible(True)
            self.tap_tpu_label.set_visible(True)
            self.tap_pitch_label.set_visible(True)

            # these conv_DROs are also changed in on_conversational_notebook_switch_page
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(False)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(False)

            self.conv_drill_tap = 'tap'
            self.tap_dro_list[self.tap_focus_list[0]].grab_focus()
            self.drill_through_hole_hint_label.set_visible(False)
            
        else:
            self.set_image('drill_tap_btn_image', 'drill_tap_drill_highlight.jpg')
            self.set_image('conv_drill_tap_image', 'lathe_drill.svg')

            # turn Tap DROs OFF
            self.tap_dro_list['tap_tool_num_dro'].set_visible(False)
            self.tap_dro_list['tap_z_start_dro'].set_visible(False)
            self.tap_dro_list['tap_tc_dro'].set_visible(False)
            self.tap_dro_list['tap_z_end_dro'].set_visible(False)
            self.tap_dro_list['tap_spindle_rpm_dro'].set_visible(False)
            self.tap_dro_list['tap_tpu_dro'].set_visible(False)
            self.tap_dro_list['tap_pitch_dro'].set_visible(False)
            self.tap_dro_list['tap_peck_dro'].set_visible(False)

            # turn Drill DROs ON
            self.drill_dro_list['drill_tool_num_dro'].set_visible(True)
            self.drill_dro_list['drill_z_start_dro'].set_visible(True)
            self.drill_dro_list['drill_tc_dro'].set_visible(True)
            self.drill_dro_list['drill_peck_dro'].set_visible(True)
            self.drill_dro_list['drill_z_end_dro'].set_visible(True)
            self.drill_dro_list['drill_spindle_rpm_dro'].set_visible(True)
            self.drill_dro_list['drill_dwell_dro'].set_visible(True)

            # turn Tap labels(text) OFF
            self.tap_z_end_label.set_visible(False)
            self.tap_tpu_label.set_visible(False)
            self.tap_pitch_label.set_visible(False)

            # turn Drill labels(text) ON
            self.drill_z_end_label.set_visible(True)
            self.drill_dwell_label.set_visible(True)

            # these conv_DROs are also changed in on_conversational_notebook_switch_page
            self.conv_dro_list['conv_rough_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_rough_fpr_dro'].set_sensitive(True)
            self.conv_dro_list['conv_finish_sfm_dro'].set_sensitive(False)
            self.conv_dro_list['conv_finish_fpr_dro'].set_sensitive(False)

            self.conv_drill_tap = 'drill'
            self.drill_through_hole_hint_label.set_visible(True)            
            self.drill_dro_list[self.drill_focus_list[0]].grab_focus()
        self.load_title_dro()


    def on_drill_tap_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.toggle_drill_tap_button()
        return

    def on_drill_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        self._update_drill_through_hole_hint_label(self.drill_dro_list['drill_tool_num_dro'].get_text(),
                                                   self.drill_dro_list['drill_z_end_dro'].get_text(),
                                                   self.drill_through_hole_hint_label)                    
        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_tool_num_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_z_start_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_tc_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_peck_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_peck_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_spindle_rpm_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % value)
        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_spindle_rpm_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_dwell_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_dwell_format % value)
        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_dwell_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    def on_drill_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        self._update_drill_through_hole_hint_label(self.drill_dro_list['drill_tool_num_dro'].get_text(),
                                                   self.drill_dro_list['drill_z_end_dro'].get_text(),
                                                   self.drill_through_hole_hint_label)                    

        where_to = self.next_conv_dro(self.drill_focus_list, 'drill_z_end_dro')
        self.drill_dro_list[where_to].grab_focus()
        return

    # ---- Tap DROs ---- #
    def on_tap_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)
        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_tool_num_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_z_start_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)
        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_tc_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_spindle_rpm_dro_activate(self, widget, data=None):
        pitch = self.tap_dro_list['tap_pitch_dro'].get_text()
        max_vel = self.inifile.find("AXIS_2", "MAX_VELOCITY")
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21, pitch, max_vel)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % value)
        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_spindle_rpm_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_tpu_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_tpu_format % value)

        pitch = 1 / value
        self.tap_dro_list["tap_pitch_dro"].set_text(self.dro_long_format % pitch)

        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_tpu_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_pitch_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        tpu = 1 / value
        self.tap_dro_list["tap_tpu_dro"].set_text(self.dro_tpu_format % tpu)

        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_pitch_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_z_end_dro')
        self.tap_dro_list[where_to].grab_focus()
        return

    def on_tap_peck_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.tap_focus_list, 'tap_peck_dro')
        self.tap_dro_list[where_to].grab_focus()
        return


    # -------------------------------------------------------------------------------------------------
    # Threading DROs
    # -------------------------------------------------------------------------------------------------

    def toggle_thread_external_internal_button(self):
        self.save_title_dro()        
        if self.conv_thread_ext_int == 'external':
            self.set_image('thread_ext_int_image', 'thread_int_button.jpg')
            self.set_image('conv_thread_image', 'lathe_thread_internal.svg')
            self.thread_major_dia_label.set_label('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >X    End :</span>')
            self.thread_minor_dia_label.set_label('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >X    Start :</span>')
            self.conv_thread_ext_int = 'internal'
        else:
            self.set_image('thread_ext_int_image', 'thread_ext_button.jpg')
            self.set_image('conv_thread_image', 'lathe_thread_external.svg')
            self.thread_major_dia_label.set_label('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >X    Start :</span>')
            self.thread_minor_dia_label.set_label('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >X    End :</span>')
            self.conv_thread_ext_int = 'external'

        model = self.thread_chart_combobox.get_model()
        active_text = self.thread_chart_combobox.get_active()
        title = model[active_text][0]
        thread_str = model[active_text][1]

        if 'NPT' in title or 'npt' in title:
            (tpu, external_pitch_dia_s, internal_pitch_dia_s,
             thread_hieght_max_s, thread_hieght_min_s, effective_length_s,
             pipe_od, hole_id) = (
                 [x.strip() for x in thread_str.split(',') if x.strip()])

            external_pitch_dia = float(external_pitch_dia_s)
            internal_pitch_dia = float(internal_pitch_dia_s)
            mean_thread_hieght = (float(thread_hieght_max_s) + float(thread_hieght_min_s)) / 2.0
            effective_length   = float(effective_length_s)

            ext_major = external_pitch_dia + mean_thread_hieght
            ext_minor = external_pitch_dia - mean_thread_hieght
            int_major = internal_pitch_dia + mean_thread_hieght
            int_minor = internal_pitch_dia - mean_thread_hieght

            valid, z_start, error_msg  = self.conversational.validate_param(self.thread_dro_list['thread_z_start_dro'])
            valid, z_end, error_msg    = self.conversational.validate_param(self.thread_dro_list['thread_z_end_dro'])
            valid, tool_clr, error_msg = self.conversational.validate_param(self.thread_dro_list['thread_tc_dro'])
            valid, lead, error_msg     = self.conversational.validate_param(self.thread_dro_list['thread_lead_dro'])

            thread_range = math.fabs(z_end - z_start) + lead
            # Pipe threads have 1 degree 47 minute taper angle to the center
            # axis, or 1:16 on the diameter
            # angle = 1.7833 degrees, sin(1.7833 degrees) = 0.031062
            dx = thread_range * 0.031062

            if self.conv_thread_ext_int == 'external':
                taper = dx
                self.conv_thread_note = '(Note: for %s External Thread)\n(Pipe OD = %s)\n(    for use with OD Turn and Chamfer)\n(Effective Thread Length = %s)\n(    use as the distance between z_start and z_end in Chamfer)\n(    use an angle of 1.78 degrees)' % (title, pipe_od, effective_length_s)
                #tpu, external_pitch_dia_s, internal_pitch_dia_s, thread_hieght_max_s, thread_hieght_min_s, effective_length_s

            else:
                taper = dx * -1
                self.conv_thread_note = '(Note: for %s Internal Thread)\n(Hole ID = %s)\n(    for use with ID Turn and Chamfer)\n(Effective Thread Length = %s)\n(    use as the distance between z_start and z_end in Chamfer)\n(    use an angle of 1.78 degrees)' % (title, hole_id, effective_length_s)
        self.load_title_dro()
        
    def on_thread_ext_int_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        self.toggle_thread_external_internal_button()

    def on_thread_tool_num_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text('%d' % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_tool_num_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_z_start_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate thread length and push to length DRO
        try:
            z_end = float(self.thread_dro_list['thread_z_end_dro'].get_text())
        except:
            z_end = 0

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_z_start_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_z_end_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # calculate thread length and push to length DRO
        try:
            z_start = float(self.thread_dro_list['thread_z_start_dro'].get_text())
        except:
            z_start = 0

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_z_end_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_spindle_rpm_dro_activate(self, widget, data=None):
        pitch = self.thread_dro_list['thread_pitch_dro'].get_text()
        max_vel = self.inifile.find("AXIS_2", "MAX_VELOCITY")
        (valid, rpm, error_msg) = self.conversational.validate_param(widget, self.g21, pitch, max_vel)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_short_format % rpm)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_spindle_rpm_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_minor_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_minor_dia_dro')
        self.thread_dro_list[where_to].grab_focus()
        self.on_thread_doc_dro_activate(self.thread_dro_list["thread_doc_dro"])
        return

    def on_thread_major_dia_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_major_dia_dro')
        self.thread_dro_list[where_to].grab_focus()
        self.on_thread_doc_dro_activate(self.thread_dro_list["thread_doc_dro"])
        return

    def on_thread_tpu_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_tpu_format % value)

        pitch = 1 / value
        self.thread_dro_list['thread_pitch_dro'].set_text(self.dro_long_format % pitch)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_tpu_dro')
        self.thread_dro_list[where_to].grab_focus()
        self.on_thread_spindle_rpm_dro_activate(self.thread_dro_list["thread_spindle_rpm_dro"])
        return

    def on_thread_pitch_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        tpu = 1 / value
        self.thread_dro_list['thread_tpu_dro'].set_text(self.dro_tpu_format % tpu)

        # Normally the second arg would be 'thread_pitch_dro' but it is not in the focus list.
        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_doc_dro')
        self.thread_dro_list[where_to].grab_focus()

        self.on_thread_spindle_rpm_dro_activate(self.thread_dro_list["thread_spindle_rpm_dro"])
        return

    def on_thread_doc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        # TODO - wrap this in error handling.  This will break if bad values are entered in these DROs.  We should ideally highlight which DRO has the bad entry
        try:
            tool_clearance = float(self.thread_dro_list['thread_tc_dro'].get_text())
            doc = float(self.thread_dro_list['thread_doc_dro'].get_text())
            major_dia = float(self.thread_dro_list['thread_major_dia_dro'].get_text())
            minor_dia = float(self.thread_dro_list['thread_minor_dia_dro'].get_text())

            depth = math.fabs(tool_clearance) + math.fabs(doc)
            full_dia_depth = math.fabs(tool_clearance)
            cut_increment = math.fabs(doc)
            k_number = math.fabs(major_dia - minor_dia)
            end_depth = math.fabs(k_number) + math.fabs(tool_clearance)
            degression = 2

            if doc > k_number:
                self.error_handler.write("Depth of cut must be smaller than (major - minor dia)/2", ALARM_LEVEL_LOW)

            tpass = self.conversational.calc_num_threading_passes(depth, end_depth, full_dia_depth, cut_increment, degression)
            self.thread_dro_list['thread_pass_dro'].set_text(self.dro_short_format % tpass)
        except:
            self.error_handler.write("insufficient or inappropriate entries on threading conversational page to calculate number of passes", ALARM_LEVEL_LOW)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_doc_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_pass_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return

        req_passes = int(value)

        # TODO - wrap this in error handling.  This will break if bad values are entered in these DROs.  We should ideally highlight which DRO has the bad entry
        try:
            # TODO - better user entry validation here
            tool_clearance = float(self.thread_dro_list['thread_tc_dro'].get_text())
            major_dia = float(self.thread_dro_list['thread_major_dia_dro'].get_text())
            minor_dia = float(self.thread_dro_list['thread_minor_dia_dro'].get_text())

            depth = math.fabs(tool_clearance)
            full_dia_depth = math.fabs(tool_clearance)
            k_number = math.fabs(major_dia - minor_dia)
            end_depth = math.fabs(k_number) + math.fabs(tool_clearance)
            degression = 2


            imin = (end_depth - depth)/req_passes
            imax = end_depth - depth
            imid = (imin + imax)/2
            i = 0

            while (True):
                i += 1
                if (i == 100): break
                num_passes = self.conversational.calc_num_threading_passes(depth, end_depth, full_dia_depth, imid, degression)
                if num_passes == req_passes: break
                if num_passes > req_passes:
                    imin = imid
                    imid = (imin + imax) / 2
                elif num_passes < req_passes:
                    imax = imid
                    imid = (imin + imax) / 2

            self.thread_dro_list["thread_doc_dro"].set_text(self.dro_long_format % imid)
            widget.set_text(self.dro_short_format % num_passes)
        except:
            self.error_handler.write("insufficient or inappropriate entries on threading conversational page to calculate number of passes", ALARM_LEVEL_LOW)

        # Normally the second arg would be 'thread_pass_dro' but it is not in the focus list.
        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_doc_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_tc_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_tc_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_lead_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_lead_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    def on_thread_chart_changed(self, widget, data=None):
        # get active entry, populate DROs
        model = widget.get_model()
        active_text = widget.get_active()
        thread_str = model[active_text][1]
        if thread_str == ' ':
            # empty selected, do nothing
            note = 'notes:'
            self.thread_chart_notes_label.set_markup(
                '<span weight="light" font_desc="Bebas 8" font_stretch="ultracondensed" foreground="white">%s</span>' % note)
            return

        title = model[active_text][0]
        if 'NPT' in title or 'npt' in title:
            (tpu, external_pitch_dia_s, internal_pitch_dia_s,
             thread_hieght_max_s, thread_hieght_min_s, effective_length_s,
             pipe_od, hole_id) = (
                 [x.strip() for x in thread_str.split(',') if x.strip()])

            external_pitch_dia = float(external_pitch_dia_s)
            internal_pitch_dia = float(internal_pitch_dia_s)
            mean_thread_hieght = (float(thread_hieght_max_s) + float(thread_hieght_min_s)) / 2.0
            effective_length   = float(effective_length_s)

            ext_major = external_pitch_dia + mean_thread_hieght
            ext_minor = external_pitch_dia - mean_thread_hieght
            int_major = internal_pitch_dia + mean_thread_hieght
            int_minor = internal_pitch_dia - mean_thread_hieght

            valid, z_start, error_msg  = self.conversational.validate_param(self.thread_dro_list['thread_z_start_dro'])
            valid, z_end, error_msg    = self.conversational.validate_param(self.thread_dro_list['thread_z_end_dro'])
            valid, tool_clr, error_msg = self.conversational.validate_param(self.thread_dro_list['thread_tc_dro'])
            valid, lead, error_msg     = self.conversational.validate_param(self.thread_dro_list['thread_lead_dro'])

            thread_range = math.fabs(z_end - z_start) + lead
            # Pipe threads have 1 degree 47 minute taper angle to the center
            # axis, or 1:16 on the diameter
            # angle = 1.7833 degrees, sin(1.7833 degrees) = 0.031062
            dx = thread_range * 0.031062

            if self.conv_thread_ext_int == 'external':
                taper = dx
                self.conv_thread_note = '(Note: for %s External Thread)\n(Pipe OD = %s)\n(    for use with OD Turn and Chamfer)\n(Effective Thread Length = %s)\n(    use as the distance between z_start and z_end in Chamfer)\n(    use an angle of 1.78 degrees)' % (title, pipe_od, effective_length_s)
                #tpu, external_pitch_dia_s, internal_pitch_dia_s, thread_hieght_max_s, thread_hieght_min_s, effective_length_s

            else:
                taper = dx * -1
                self.conv_thread_note = '(Note: for %s Internal Thread)\n(Hole ID = %s)\n(    for use with ID Turn and Chamfer)\n(Effective Thread Length = %s)\n(    use as the distance between z_start and z_end in Chamfer)\n(    use an angle of 1.78 degrees)' % (title, hole_id, effective_length_s)

            pitch = 1.0 / float(tpu)
            sharp_thread_height = pitch / 1.15470  # pitch / (2 * tan(30 degrees))
            tip_truncation = (sharp_thread_height - mean_thread_hieght) / 2

            note = ('notes:    tip    truncation    =    %s,\neffective    thread    length    =    %s'
                    % (self.dro_long_format, self.dro_long_format)
                    % (tip_truncation, effective_length))
            self.thread_chart_notes_label.set_markup(
                '<span weight="light" font_desc="Bebas 8" font_stretch="ultracondensed" foreground="white">%s</span>' % note)

        else:
            note = 'notes:'
            self.thread_chart_notes_label.set_markup(
                '<span weight="light" font_desc="Bebas 8" font_stretch="ultracondensed" foreground="white">%s</span>' % note)

            # parse space delimited string in text
            tpu, ext_major, ext_minor, int_major, int_minor = [x.strip() for x in thread_str.split(',') if x.strip()]
            taper = 0.0

        # use external or internal diameters as required to set DROs
        if self.conv_thread_ext_int == 'external':
            major_diameter = float(ext_major)
            minor_diameter = float(ext_minor)
        else:
            major_diameter = float(int_major)
            minor_diameter = float(int_minor)

        #check for front or rear tool post
        valid, tool_number, error =  self.conversational.validate_param(self.thread_dro_list['thread_tool_num_dro'])
        if not valid:
            self.error_handler.write('Conversational Threading entry error - ' + error)
        tool_type = self.get_tool_type(tool_number)
        if 'ftp' in tool_type:
            major_diameter *= -1
            minor_diameter *= -1

        # set the DROs
        tpu = float(tpu)
        pitch = 1 / tpu
        self.thread_dro_list['thread_major_dia_dro'].set_text(self.dro_long_format % major_diameter)
        self.thread_dro_list['thread_minor_dia_dro'].set_text(self.dro_long_format % minor_diameter)
        self.thread_dro_list['thread_tpu_dro'].set_text(self.dro_tpu_format % tpu)
        self.thread_dro_list['thread_pitch_dro'].set_text(self.dro_long_format % pitch)
        self.thread_dro_list['thread_taper_dro'].set_text(self.dro_long_format % taper)

        # re-calc number of passes, re-check rpm
        self.on_thread_doc_dro_activate(self.thread_dro_list["thread_doc_dro"])
        self.on_thread_spindle_rpm_dro_activate(self.thread_dro_list["thread_spindle_rpm_dro"])
        return

    def on_thread_chart_popdown(self, widget, data=None):
        self.refresh_thread_data_liststores()

    def on_thread_taper_dro_activate(self, widget, data=None):
        (valid, value, error_msg) = self.conversational.validate_param(widget, self.g21)
        if not valid:
            self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            return
        widget.set_text(self.dro_long_format % value)

        where_to = self.next_conv_dro(self.thread_focus_list, 'thread_taper_dro')
        self.thread_dro_list[where_to].grab_focus()
        return

    # -------------------------------------------------------------------------------------------------
    # end of conversational tab callbacks
    # -------------------------------------------------------------------------------------------------

    # helper function for issuing MDI commands
    def issue_mdi(self, mdi_command):
        if self.moving(): return

        self.error_handler.write("issuing MDI command: " + mdi_command, ALARM_LEVEL_DEBUG)
        self.ensure_mode(linuxcnc.MODE_MDI)
        self.command.mdi(mdi_command)

    def issue_toolchange_command(self, tool_number):
        # build T MDI command
        if ((self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_ATC_8POSITION) and (self.hal['door-switch'] == 1) and (tool_number < 9)):
            self.error_handler.write("Must close enclosure door before indexing turret")
            return
        t_command = "T%d" % (tool_number)
        self.issue_mdi(t_command)
        # prevent unsettling flicker of tool number as old status.tool_no gets stuffed into dro on 50ms loop
        # before new tool takes hold
        # NB - not a good way of accomplishing this.
        #self.command.wait_complete()
        tool_type = self.get_tool_type(tool_number)
        tool_type += '.png'
        self.clear_tool_select_images(tool_type, True)
        self.set_touch_x_z_leds(tool_number)

    def issue_g10(self, axis, offset_register, value):
        # don't allow offset change without axes referenced
        if not (self.x_referenced and self.z_referenced):
            self.error_handler.write("Must reference X and Z axes before issuing command", ALARM_LEVEL_MEDIUM)
            return
        g10_command = "G10 L1 P%d %s%f" %(offset_register, axis, value)
        self.issue_mdi(g10_command)
        # prevent race condition when setting both geo and wear almost simultaneously
        self.command.wait_complete()
        # reload tool liststore if wear offsets were changed
        if offset_register < 10000:
            self.refresh_tool_liststore()
        else:
            # we're setting a geometery offset, so set touch x or z button leds appropriately
            if value == 0:
                if axis == 'X':
                    self.set_image('touch_x_image', 'touch_x_black_led.png')
                if axis == 'Z':
                    self.set_image('touch_z_image', 'touch_z_black_led.png')
            else:
                if axis == 'X':
                    self.set_image('touch_x_image', 'touch_x_green_led.png')
                if axis == 'Z':
                    self.set_image('touch_z_image', 'touch_z_green_led.png')

    def save_conv_paramters(self, dro_list):
        # loop through conv dro list and save values to redis
        for name, dro in self.conv_dro_list.iteritems():
            val = dro.get_text()
            self.redis.hset('conversational', name, val)
        for name, dro in dro_list.iteritems():
            val = dro.get_text()
            self.redis.hset('conversational', name, val)
        return

    def restore_conv_parameters(self):
        conv_dict = self.redis.hgetall('conversational')
        for dro_name , val in conv_dict.iteritems():
            try:
                if 'conv' in dro_name:
                    self.conv_dro_list[dro_name].set_text(val)
                if 'od_turn' in dro_name:
                    self.od_turn_dro_list[dro_name].set_text(val)
                if 'id_basic' in dro_name:
                    self.id_basic_dro_list[dro_name].set_text(val)
                if 'id_turn' in dro_name:
                    self.id_turn_dro_list[dro_name].set_text(val)
                if 'face' in dro_name:
                    if not 'id_turn' in dro_name:
                        self.face_dro_list[dro_name].set_text(val)
                if 'corner' in dro_name:
                    if 'chamfer' in dro_name:
                        if 'od' in dro_name:
                            self.corner_chamfer_od_dro_list[dro_name].set_text(val)
                        else:  # 'id'
                            self.corner_chamfer_id_dro_list[dro_name].set_text(val)
                    else:  # 'radius'
                        if 'od' in dro_name:
                            self.corner_radius_od_dro_list[dro_name].set_text(val)
                        else:  # 'id'
                            self.corner_radius_id_dro_list[dro_name].set_text(val)
                if 'groove' in dro_name:
                    self.groove_dro_list[dro_name].set_text(val)
                if 'part' in dro_name:
                    self.part_dro_list[dro_name].set_text(val)
                if 'drill' in dro_name:
                    self.drill_dro_list[dro_name].set_text(val)
                if 'tap_' in dro_name:
                    self.tap_dro_list[dro_name].set_text(val)
                if 'thread' in dro_name:
                    self.thread_dro_list[dro_name].set_text(val)
            except:
                pass
        return


    # debug only??
    def get_lcnc_mode_string(self, mode):
        tmp_str = 'unknown'
        if mode == linuxcnc.MODE_MANUAL:
            tmp_str = 'MODE_MANUAL'
        elif mode == linuxcnc.MODE_AUTO:
            tmp_str = 'MODE_AUTO'
        elif mode == linuxcnc.MODE_MDI:
            tmp_str = 'MODE_MDI'
        return tmp_str

    def get_lcnc_interp_string(self, state):
        tmp_str = 'unknown'
        if state == linuxcnc.INTERP_IDLE:
            tmp_str = 'INTERP_IDLE'
        elif state == linuxcnc.INTERP_READING:
            tmp_str = 'INTERP_READING'
        elif state == linuxcnc.INTERP_PAUSED:
            tmp_str = 'INTERP_PAUSED'
        elif state == linuxcnc.INTERP_WAITING:
            tmp_str = 'INTERP_WAITING'
        return tmp_str

    def get_lcnc_state_string(self, state):
        tmp_str = 'unknown'
        if state == linuxcnc.STATE_ESTOP:
            tmp_str = 'STATE_ESTOP'
        elif state == linuxcnc.STATE_ESTOP_RESET:
            tmp_str = 'STATE_ESTOP_RESET'
        elif state == linuxcnc.STATE_OFF:
            tmp_str = 'STATE_OFF'
        elif state == linuxcnc.STATE_ON:
            tmp_str = 'STATE_ON'
        return tmp_str

    def load_reset_image(self, image_color):
        # for 'white' load white image if not already loaded
        # for 'blink': if now 'red' load 'non-red' else load 'red'
        if image_color == 'white' and self.reset_image_color != 'white':
            # if not already green, make it green
            self.set_image('reset_image', 'Reset-White.jpg')
            self.reset_image_color = 'white'
        elif image_color == 'blink':
            # blink between red and non-red
            if self.reset_image_color == 'red':
                # load non-red
                self.set_image('reset_image', 'Reset-White.jpg')
                self.reset_image_color = 'non-red'
            else:
                # load red
                self.set_image('reset_image', 'Reset-Red.jpg')
                self.reset_image_color = 'red'


    # periodic updates to LEDs, DROs, more . . .
    def periodic_status(self):
        self.status.poll()
        # 50 ms functions is always called
        self.status_periodic_50ms()
        # increment loop counter and call slow period function as needed
        self.periodic_loopcount += 1
        if self.periodic_loopcount >= 10:
            self.status_periodic_500ms()
            self.periodic_loopcount = 0
        return True

    # called every 500 milliseconds to update various slower changing DROs and button images
    def status_periodic_500ms(self):

        # poll for errors
        error = self.error.poll()
        if error:
            error_kind, error_msg = error
            if error_kind == EMC_OPERATOR_DISPLAY_TYPE:
                # maybe revisit this after discussion - dispaly on gremlin message line?
                #self.set_message_line_text(error_msg)
                self.error_handler.write(error_msg, ALARM_LEVEL_LOW)
            else:
                # display on UI
                self.error_handler.write(error_msg)

        # current tool info, save to redis immediately on switch
        if not self.first_run and self.current_tool != self.status.tool_in_spindle:
            self.current_tool = self.status.tool_in_spindle
            self.redis.hset('machine_prefs', 'active_tool', self.current_tool)
            
        
        # metric/imperial switch
        self.g21 = self.status.gcodes[5] == 210
        if self.g21 != self.prev_g21:
            self.prev_g21 = self.g21
            self.switch_inch_metric_display()

        # axis ref'ed button LEDs
        self.x_referenced = self.status.homed[0]
        self.z_referenced = self.status.homed[2]

        if self.x_referenced != self.prev_x_referenced:
            self.prev_x_referenced = self.x_referenced
            if self.x_referenced:
                self.set_image('ref_x_image', 'Ref_X_Green.png')
            else:
                self.set_image('ref_x_image', 'Ref_X_Black.png')

        if self.y_referenced != self.prev_y_referenced:
            self.prev_y_referenced = self.y_referenced
            if self.y_referenced:
                self.set_image('ref_y_image', 'Ref_Y_Green.png')
            else:
                self.set_image('ref_y_image', 'Ref_Y_Black.png')

        if self.z_referenced != self.prev_z_referenced:
            self.prev_z_referenced = self.z_referenced
            if self.z_referenced:
                self.set_image('ref_z_image', 'Ref_Z_Green.png')
            else:
                self.set_image('ref_z_image', 'Ref_Z_Black.png')

        # Button LEDs
        # CSS
        self.css_active = self.status.gcodes[13] == 960
        if self.css_active != self.prev_css_active:
            self.prev_css_active = self.css_active
            self.refresh_css_label()
        # F/ REV button
        if self.f_per_rev_active != self.prev_f_per_rev_active:
            self.prev_f_per_rev_active = self.f_per_rev_active
            if self.f_per_rev_active:
                self.set_image('feed_per_rev_image', 'F-Rev-Yellow.jpg')
                self.f_per_min_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="#A19D9F" >/min</span>')
                self.f_per_rev_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >/rev</span>')
            else:
                self.set_image('feed_per_rev_image', 'F-Rev-Black.jpg')
                self.f_per_min_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >/min</span>')
                self.f_per_rev_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="#A19D9F" >/rev</span>')

        # NB: status.paused returns true any time single block is active
        # cycle start LED blinks on single block, m01 breaks, or manual tool changes.  only blink when a program is running.
        # returns True only if executing a g code program, False otherwise (including during MDI moves)
        program_running = self.program_running()

        if program_running:
            self.elapsed_time_label.set_text('%02d:%02d:%02d' % (self.hal['cycle-time-hours'], self.hal['cycle-time-minutes'], self.hal['cycle-time-seconds']))
            # only main screen available to user while running code
            # TODO - on error we should diplay alarms tab too.
            if not self.notebook_locked:
                hide_notebook_tabs(self.notebook)
                self.notebook_locked = True
            # use spindle_sync for f/rev status until interptime/motiontime disconnect is fixed
            # but only while running code.
            self.f_per_rev_active = self.status.spindle_sync != 0
            self.prev_f_word = abs(self.status.settings[1])
            # Blink CS button when the GUI is waiting for input on tool change, single block, feedhold, etc.
            if (self.single_block_active or self.feedhold_active or self.m01_break_active or (self.hal['tool-change'] and self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_MANUAL)):
                if (self.status.current_vel == 0) and (self.maxvel_override_adjustment.get_value() != 0):
                    self.load_cs_image('blink')
                else:
                    self.load_cs_image('green')
            else:
                self.load_cs_image('green')

        # many GUI updates only performed when we're not excecuting a g code program
        else:
            # unlock notebook
            if self.notebook_locked:
                show_notebook_tabs(self.notebook)
                self.notebook_locked = False

            # jog axis enabled update of LEDs
            self.refresh_jog_active_leds()

            # updates to images/labels on notebook pages other than main
            if self.current_page == FILE_PAGE:
                pass
            elif  self.current_page == TOOL_PAGE:
                # refresh the work offset liststore
                self.refresh_work_offset_liststore()
                # update tool DRO, but not if the turret fwd button has prelight
                if not (self.dro_list['tool_dro'].masked or self.button_list['turret_fwd'].has_focus()):
                    self.dro_list['tool_dro'].set_text(self.dro_short_format % self.status.tool_in_spindle)
                self.update_tool_orientation_image()
            # display active g code on settings screen
            elif self.current_page == SETTINGS_PAGE:
                self.gcodes_display.highlight_active_codes(self.active_gcodes())


            # f per rev val comes from status if we're not running a program
            self.f_per_rev_active = self.status.gcodes[7] == 950

            self.load_cs_image('dark')
            # not running a program - handle manual tool change request via MDI.  No need to confirm w/ CS button, just do it.
            if (self.hal['tool-change'] and self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_MANUAL):
                self.hal['manual-tool-changed'] = self.hal['tool-change']

        # the following updates occur regarless of program_running state:

        active_codes = " ".join(self.active_gcodes())

        self.active_gcodes_label.set_text(active_codes)

        # todo - do we really need to do this every .5 seconds, even when the info hasn't changed?
        if (self.hal['tool-change'] and self.hal['tool-changer-type'] == TOOL_CHANGER_TYPE_MANUAL):
            requested_tool = str(self.hal['tool-prep-number'])
            if self.program_running():
                self.set_message_line_text('Change     tools     to    T' + requested_tool)
                self.blink_tool_request_label(requested_tool)
        else:
            if self.current_tool < 10:
                current_tool_string = "0" + str(self.current_tool)
            else:
                current_tool_string = str(self.current_tool)
            # status.tool_offset returns the offset in whatever g20/21 units are current
            # since our hack stores the wear offset number in as the 'w' offset, we'll need to
            # translate this back if we happen to be in G21.
            wear_offset_num = int(round(self.mm_inch_scalar * self.status.tool_offset[8]))
            if wear_offset_num == 0:
                wear_offset_num = ""
            elif wear_offset_num < 10:
                wear_offset_num = "0" + str(wear_offset_num)
            else:
                wear_offset_num = str(wear_offset_num)
            self.tool_label.set_text("T:    " + current_tool_string + wear_offset_num)

        # update spindle direction
        self.spindle_direction = self.status.spindle_direction


        if self.spindle_direction != 0:
            # don't flag this if the spindle's not running
            # TODO - this alarm seems to occur often because of CSS at large diameters (e.g. moving to tool change position)
            # We should find a way to fix this...
            self.spindle_range_alarm = self.hal['spindle-range-alarm']
            # don't let these stack up.
            if self.spindle_range_alarm != self.prev_spindle_range_alarm:
                self.prev_spindle_range_alarm = self.spindle_range_alarm
                if self.spindle_range_alarm and not self.is_rapidturn:
                    self.error_handler.write("Commanded spindle speed outside of range for current belt position.  Speed will be clipped to acceptable max/min", ALARM_LEVEL_LOW)

        if self.spindle_direction != self.prev_spindle_direction:
            self.prev_spindle_direction = self.spindle_direction
            if self.spindle_direction == -1:
                # CCW
                self.set_image('rev_image', 'REV_Green.jpg')
                self.set_image('fwd_image', 'FWD_Black.jpg')
            elif self.spindle_direction == 0:
                # Off
                self.set_image('rev_image', 'REV_Black.jpg')
                self.set_image('fwd_image', 'FWD_Black.jpg')
            elif self.spindle_direction == 1:
                # CW
                self.set_image('rev_image', 'REV_Black.jpg')
                self.set_image('fwd_image', 'FWD_Green.jpg')

        # reset button
        if self.status.task_state == linuxcnc.STATE_ESTOP or \
           self.status.task_state == linuxcnc.STATE_ESTOP_RESET or \
           self.status.task_state == linuxcnc.STATE_OFF:
            # blink
            self.load_reset_image('blink')
            if self.estop_alarm == False:
                # only do this once per press press of reset
                # and don't alarm at startup
                # deref axes on estop
                self.command.unhome(0)
                self.command.unhome(2)
                self.error_handler.write(ESTOP_ERROR_MESSAGE)
            # set to true to prevent these messages from stacking up.
            # cleared in reset button handler
            self.estop_alarm = True
        else:
            # load white image
            self.load_reset_image('white')


        # DROs

        # spindle rpm and css DROs
        # if the machine is moving, use S value from HAL spindle-speed-out and F value from status.current_vel
        if self.moving():
            if not self.dros_locked:
                self.dros_locked = True
                # lock out DROs
                for name, dro in self.dro_list.iteritems():
                    dro.set_can_focus(False)
            self.rpm = int(60 * self.hal["spindle-speed-out"])
            # CS (in ft/min) = rpm * dia (in inches) * pi / 12
            surface_speed = abs((self.rpm * self.dia * 3.1416) / 12)
            # in metric, dia is in mm, but surface speed reported in meters
            if self.g21:
                # dia will be in mm, so 12/1000 gets us from the calc in inches to the calc in mm
                # seems clunky but saves an if statement
                surface_speed = surface_speed * 12/1000
            feed_per_min = self.status.current_vel * 60
            self.dro_list['spindle_rpm_dro'].set_text(self.dro_short_format % abs(self.rpm))
            self.dro_list['spindle_css_dro'].set_text(self.dro_short_format % abs(surface_speed))
            self.dro_list['feed_per_min_dro'].set_text(self.dro_medium_format % (self.mm_inch_scalar * feed_per_min))
            # calculate the F/rev DRO and stuff it
            if self.rpm != 0:
                self.dro_list['feed_per_rev_dro'].set_text(self.dro_long_format % (self.mm_inch_scalar * feed_per_min / self.rpm))
        # otherwise if we're not running, use the buggy status.settings values

        else:
            # if gcode file is loaded and has changed on disk since loading, reload it
            if self.current_gcode_file_path != '':
                self.check_for_gcode_program_reload()            
            if self.dros_locked:
                self.dros_locked = False
                # unlock DROs
                for name, dro in self.dro_list.iteritems():
                    dro.set_can_focus(True)
            if not (self.dro_list['spindle_rpm_dro'].masked or self.dro_list['spindle_css_dro'].masked):
                self.s_word  = self.status.settings[2]
                # if we're in CSS (G96), settings[2] contains the S word in Ft/min or M/min, so reassign surface speed to rpm
                if self.css_active:
                    surface_speed = self.s_word
                    if self.dia != 0.0:
                        self.rpm = 12 * surface_speed / (self.dia * 3.1416)
                        if self.g21:
                            # get rid of 12 inches/ft adjustment
                            self.rpm = self.rpm * 1000 / 12
                    else:
                        self.rpm = self.max_rpm
                    if abs(self.rpm) > self.max_rpm:
                        self.rpm = self.max_rpm
                else:
                    self.rpm = self.s_word
                    surface_speed = (self.rpm * self.dia * 3.1416) / 12
                    if self.g21:
                        surface_speed = surface_speed * 12/1000

                self.dro_list['spindle_rpm_dro'].set_text(self.dro_short_format % abs(self.rpm))
                self.dro_list['spindle_css_dro'].set_text(self.dro_short_format % abs(surface_speed))
            # feed per rev and per rpm
            if not (self.dro_list['feed_per_min_dro'].masked or self.dro_list['feed_per_rev_dro'].masked):
                self.f_word = abs(self.status.settings[1])
                # If we're in feed/rev, the F word has the f/rev value so we need to calc f/min
                if self.f_per_rev_active:
                    # until motion and interp are synched, we've got two possible sources of information here:
                    # uu_per_rev comes down from motion and displays the last feed/rev setting on a motion segment
                    # f_word comes from the interp and displays the value off the interp list, which if a program
                    # is stopped in the middle may display future instead of current information
                    if self.f_word != self.prev_f_word:
                        # someone's sent an MDI statement and changed the f_word, so DON'T use the uu_per_rev setting
                        # cycle start callback will set prev_f_word back equal to f_word, and therefore reset this to display the uu_per_rev val.
                        feed_per_rev = self.f_word
                    else:
                        # else no one's messed with our f word, so let's use the motion settings (which are in g20 units only!!!)
                        feed_per_rev = self.mm_inch_scalar * self.status.uu_per_rev
                    feed_per_min = feed_per_rev * self.rpm
                else:
                    feed_per_min = self.f_word
                    if self.rpm != 0:
                        feed_per_rev = self.f_word / self.rpm
                    else:
                        feed_per_rev = 0
                # stuff the DROs
                self.dro_list['feed_per_min_dro'].set_text(self.dro_medium_format % abs(feed_per_min))
                self.dro_list['feed_per_rev_dro'].set_text(self.dro_long_format % abs(feed_per_rev))

        # debug info - observe mode/status changes
        # this is far from perfect: the mode can change and return during time elapsed between these checks
        if self.status.task_mode != self.prev_lcnc_task_mode:
            # state changed, print to console
            self.error_handler.write('LinuxCNC status.task_mode change was %s is now %s' % (self.get_lcnc_mode_string(self.prev_lcnc_task_mode), self.get_lcnc_mode_string(self.status.task_mode)), ALARM_LEVEL_DEBUG)
            self.prev_lcnc_task_mode = self.status.task_mode
        if self.prev_lcnc_interp_state != self.status.interp_state:
            # interpreter state changed
            self.error_handler.write('LinuxCNC interp_state change was %s is now %s' % (self.get_lcnc_interp_string(self.prev_lcnc_interp_state), self.get_lcnc_interp_string(self.status.interp_state)), ALARM_LEVEL_DEBUG)
            self.prev_lcnc_interp_state = self.status.interp_state
            # kludge to rewind program after interp goes idle (usually when program is done at M30)
            if "IDLE" in self.get_lcnc_interp_string(self.status.interp_state):
                self.gcodelisting_mark_start_line(1)
                self.sourceview.scroll_to_mark(self.gcodelisting_start_mark, 0, True, 0, 0.5)             

        if self.prev_task_state != self.status.task_state:
            self.error_handler.write("status.task_state was %s is now %s" % (self.get_lcnc_state_string(self.prev_task_state), self.get_lcnc_state_string(self.status.task_state)), ALARM_LEVEL_DEBUG)
            self.prev_task_state = self.status.task_state

        if self.hal['mesa-watchdog-has-bit']:
            # problem! the Mesa card watchdog has bitten
            # high priority warning
            if not self.mesa_watchdog_has_bit_seen:
                # set state to ESTOP
                self.mesa_watchdog_has_bit_seen = True
                self.command.state(linuxcnc.STATE_ESTOP)
                self.error_handler.write("Mesa interface card watchdog has bitten. Must press RESET.", ALARM_LEVEL_MEDIUM)

        # log abnormal changes in cpu utilization
        usage = self.proc.cpu_percent()
        if usage > 15 and (usage > 2 * self.cpu_usage) or (usage < .5 * self.cpu_usage) or (usage > CPU_USAGE_THRESHOLD):
            self.error_handler.write("CPU usage was %.1f, is now %.1f" % (self.cpu_usage, usage), ALARM_LEVEL_DEBUG)
            self.cpu_usage = usage        

    # called every 50 milliseconds to update faster changing indicators
    # the '@profile' below is for use with line profiling http://pythonhosted.org/line_profiler/
    #@profile
    def status_periodic_50ms(self):
        # check button events from keyboard shortcuts
        self.check_keyboard_shortcut_fifo()

        self.current_page = self.notebook.get_current_page()
        p = self.status.actual_position
        # p[] register is always in machine units - commands to LCNC (MDI, offset commands) are in G20/21 units
        # diameter DRO
        if not self.dro_list['dia_dro'].masked:
            x = p[0] - self.status.g5x_offset[0] - self.status.tool_offset[0] - self.status.g92_offset[0]
            # p[0] always returns the X value in radius, regardless of G7/G8 state.
            self.dia = x * self.mm_inch_scalar * 2
            self.dro_list['dia_dro'].set_text(self.dro_long_format % self.dia)

        # z DRO
        if not self.dro_list['z_dro'].masked:
            z = p[2] - self.status.g5x_offset[2] - self.status.tool_offset[2] -  self.status.g92_offset[2]
            self.z = z * self.mm_inch_scalar
            self.dro_list['z_dro'].set_text(self.dro_long_format % self.z)

        # y DRO
        if self.is_rapidturn and not self.dro_list['y_dro'].masked:
            y = p[1] - self.status.g5x_offset[1] - self.status.tool_offset[1] -  self.status.g92_offset[1]
            y = y * self.mm_inch_scalar
            self.dro_list['y_dro'].set_text(self.dro_long_format % y)

        # dtg labels
        dtg = self.status.dtg
        x = dtg[0] * self.mm_inch_scalar * 2
        z = dtg[2] * self.mm_inch_scalar
        self.x_dtg_label.set_text(self.dro_long_format % x)
        self.z_dtg_label.set_text(self.dro_long_format % z)

        # the following updates are only performed if we ARE running a g code program
        self.update_gcode_display()

        # watch for LinuxCNC to change coolant request state
        # when it does copy it to tormach.coolant HAL pin
        if self.prev_coolant_iocontrol != self.hal['coolant-iocontrol']:
            if self.hal['coolant-iocontrol']:
                self.hal['coolant'] = 1
            else:
                self.hal['coolant'] = 0
            # current becomes previous
            self.prev_coolant_iocontrol = self.hal['coolant-iocontrol']

        # door switch status
        self.door_switch = self.hal['door-switch']
        if self.door_switch != self.prev_door_switch:
            self.prev_door_switch = self.door_switch
            if self.door_switch == 1:
                self.set_image('door_sw_led', 'LED-Yellow.jpg')
                if self.program_running(False):
                    # pause program
                    self.pause_program_for_door_sw_open()
                else:
                    # stop the spindle
                    if self.spindle_running():
                        self.issue_mdi('M5')
            else:
                self.set_image('door_sw_led', 'LED-Green.jpg')

        # limit switch status
        self.home_switch = self.hal['home-switch']
        if self.home_switch != self.prev_home_switch:
            self.prev_home_switch = self.home_switch
            if self.home_switch == 1:
                self.set_image('home_sw_led', 'LED-Yellow.jpg')
            else:
                self.set_image('home_sw_led', 'LED-Black.jpg')

        # Inputled0 status
        self.Inputled0 = self.hal['Inputled0']
        if self.Inputled0 != self.prev_Inputled0:
            self.prev_Inputled0 = self.Inputled0
            if self.Inputled0 == 1:
                self.set_image('Inputled0', 'LED-Green.jpg')
            else:
                self.set_image('Inputled0', 'LED-Black.jpg')

        # Inputled1 status
        self.Inputled1 = self.hal['Inputled1']
        if self.Inputled1 != self.prev_Inputled1:
            self.prev_Inputled1 = self.Inputled1
            if self.Inputled1 == 1:
                self.set_image('Inputled1', 'LED-Green.jpg')
            else:
                self.set_image('Inputled1', 'LED-Black.jpg')

        # Inputled2 status
        self.Inputled2 = self.hal['Inputled2']
        if self.Inputled2 != self.prev_Inputled2:
            self.prev_Inputled2 = self.Inputled2
            if self.Inputled2 == 1:
                self.set_image('Inputled2', 'LED-Green.jpg')
            else:
                self.set_image('Inputled2', 'LED-Black.jpg')

        # Inputled3 status
        self.Inputled3 = self.hal['Inputled3']
        if self.Inputled3 != self.prev_Inputled3:
            self.prev_Inputled3 = self.Inputled3
            if self.Inputled3 == 1:
                self.set_image('Inputled3', 'LED-Green.jpg')
            else:
                self.set_image('Inputled3', 'LED-Black.jpg')

        # Inputled4 status
        self.Inputled4 = self.hal['Inputled4']
        if self.Inputled4 != self.prev_Inputled4:
            self.prev_Inputled4 = self.Inputled4
            if self.Inputled4 == 1:
                self.set_image('Inputled4', 'LED-Green.jpg')
            else:
                self.set_image('Inputled4', 'LED-Black.jpg')

        # Inputled5 status
        self.Inputled5 = self.hal['Inputled5']
        if self.Inputled5 != self.prev_Inputled5:
            self.prev_Inputled5 = self.Inputled5
            if self.Inputled5 == 1:
                self.set_image('Inputled5', 'LED-Green.jpg')
            else:
                self.set_image('Inputled5', 'LED-Black.jpg')

        # Inputled6 status
        self.Inputled6 = self.hal['Inputled6']
        if self.Inputled6 != self.prev_Inputled6:
            self.prev_Inputled6 = self.Inputled6
            if self.Inputled6 == 1:
                self.set_image('Inputled6', 'LED-Green.jpg')
            else:
                self.set_image('Inputled6', 'LED-Black.jpg')

        # Inputled7 status
        self.Inputled7 = self.hal['Inputled7']
        if self.Inputled7 != self.prev_Inputled7:
            self.prev_Inputled7 = self.Inputled7
            if self.Inputled7 == 1:
                self.set_image('Inputled7', 'LED-Green.jpg')
            else:
                self.set_image('Inputled7', 'LED-Black.jpg')

        # Inputled8 status
        self.Inputled8 = self.hal['Inputled8']
        if self.Inputled8 != self.prev_Inputled8:
            self.prev_Inputled8 = self.Inputled8
            if self.Inputled8 == 1:
                self.set_image('Inputled8', 'LED-Green.jpg')
            else:
                self.set_image('Inputled8', 'LED-Black.jpg')

        # Inputled9 status
        self.Inputled9 = self.hal['Inputled9']
        if self.Inputled9 != self.prev_Inputled9:
            self.prev_Inputled9 = self.Inputled9
            if self.Inputled9 == 1:
                self.set_image('Inputled9', 'LED-Green.jpg')
            else:
                self.set_image('Inputled9', 'LED-Black.jpg')

        # Inputled10 status
        self.Inputled10 = self.hal['Inputled10']
        if self.Inputled10 != self.prev_Inputled10:
            self.prev_Inputled10 = self.Inputled10
            if self.Inputled10 == 1:
                self.set_image('Inputled10', 'LED-Green.jpg')
            else:
                self.set_image('Inputled10', 'LED-Black.jpg')

        # Inputled11 status
        self.Inputled11 = self.hal['Inputled11']
        if self.Inputled11 != self.prev_Inputled11:
            self.prev_Inputled11 = self.Inputled11
            if self.Inputled11 == 1:
                self.set_image('Inputled11', 'LED-Green.jpg')
            else:
                self.set_image('Inputled11', 'LED-Black.jpg')

        # Inputled12 status
        self.Inputled12 = self.hal['Inputled12']
        if self.Inputled12 != self.prev_Inputled12:
            self.prev_Inputled12 = self.Inputled12
            if self.Inputled12 == 1:
                self.set_image('Inputled12', 'LED-Green.jpg')
            else:
                self.set_image('Inputled12', 'LED-Black.jpg')

        # Inputled13 status
        self.Inputled13 = self.hal['Inputled13']
        if self.Inputled13 != self.prev_Inputled13:
            self.prev_Inputled13 = self.Inputled13
            if self.Inputled13 == 1:
                self.set_image('Inputled13', 'LED-Green.jpg')
            else:
                self.set_image('Inputled13', 'LED-Black.jpg')

        # Inputled14 status
        self.Inputled14 = self.hal['Inputled14']
        if self.Inputled14 != self.prev_Inputled14:
            self.prev_Inputled14 = self.Inputled14
            if self.Inputled14 == 1:
                self.set_image('Inputled14', 'LED-Green.jpg')
            else:
                self.set_image('Inputled14', 'LED-Black.jpg')

        # Inputled15 status
        self.Inputled15 = self.hal['Inputled15']
        if self.Inputled15 != self.prev_Inputled15:
            self.prev_Inputled15 = self.Inputled15
            if self.Inputled15 == 1:
                self.set_image('Inputled15', 'LED-Green.jpg')
            else:
                self.set_image('Inputled15', 'LED-Black.jpg')

        # Outputled0 status
        self.Outputled0 = self.hal['Outputled0']
        if self.Outputled0 != self.prev_Outputled0:
            self.prev_Outputled0 = self.Outputled0
            if self.Outputled0 == 1:
                self.set_image('Outputled0', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled0', 'LED-Black.jpg')

        # Outputled1 status
        self.Outputled1 = self.hal['Outputled1']
        if self.Outputled1 != self.prev_Outputled1:
            self.prev_Outputled1 = self.Outputled1
            if self.Outputled1 == 1:
                self.set_image('Outputled1', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled1', 'LED-Black.jpg')

        # Outputled2 status
        self.Outputled2 = self.hal['Outputled2']
        if self.Outputled2 != self.prev_Outputled2:
            self.prev_Outputled2 = self.Outputled2
            if self.Outputled2 == 1:
                self.set_image('Outputled2', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled2', 'LED-Black.jpg')

        # Outputled3 status
        self.Outputled3 = self.hal['Outputled3']
        if self.Outputled3 != self.prev_Outputled3:
            self.prev_Outputled3 = self.Outputled3
            if self.Outputled3 == 1:
                self.set_image('Outputled3', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled3', 'LED-Black.jpg')

        # Outputled4 status
        self.Outputled4 = self.hal['Outputled9']
        if self.Outputled4 != self.prev_Outputled4:
            self.prev_Outputled4 = self.Outputled4
            if self.Outputled4 == 1:
                self.set_image('Outputled4', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled4', 'LED-Black.jpg')

        # Outputled5 status
        self.Outputled5 = self.hal['Outputled5']
        if self.Outputled5 != self.prev_Outputled5:
            self.prev_Outputled5 = self.Outputled5
            if self.Outputled5 == 1:
                self.set_image('Outputled5', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled5', 'LED-Black.jpg')

        # Outputled6 status
        self.Outputled6 = self.hal['Outputled9']
        if self.Outputled6 != self.prev_Outputled6:
            self.prev_Outputled6 = self.Outputled6
            if self.Outputled6 == 1:
                self.set_image('Outputled6', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled6', 'LED-Black.jpg')

        # Outputled7 status
        self.Outputled7 = self.hal['Outputled7']
        if self.Outputled7 != self.prev_Outputled7:
            self.prev_Outputled7 = self.Outputled7
            if self.Outputled7 == 1:
                self.set_image('Outputled7', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled7', 'LED-Black.jpg')

        # Outputled8 status
        self.Outputled8 = self.hal['Outputled8']
        if self.Outputled8 != self.prev_Outputled8:
            self.prev_Outputled8 = self.Outputled8
            if self.Outputled8 == 1:
                self.set_image('Outputled8', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled8', 'LED-Black.jpg')

        # Outputled9 status
        self.Outputled9 = self.hal['Outputled9']
        if self.Outputled9 != self.prev_Outputled9:
            self.prev_Outputled9 = self.Outputled9
            if self.Outputled9 == 1:
                self.set_image('Outputled9', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled9', 'LED-Black.jpg')

        # Outputled10 status
        self.Outputled10 = self.hal['Outputled10']
        if self.Outputled10 != self.prev_Outputled10:
            self.prev_Outputled10 = self.Outputled10
            if self.Outputled10 == 1:
                self.set_image('Outputled10', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled10', 'LED-Black.jpg')

        # Outputled11 status
        self.Outputled11 = self.hal['Outputled14']
        if self.Outputled11 != self.prev_Outputled11:
            self.prev_Outputled11 = self.Outputled11
            if self.Outputled11 == 1:
                self.set_image('Outputled11', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled11', 'LED-Black.jpg')

        # Outputled12 status
        self.Outputled12 = self.hal['Outputled14']
        if self.Outputled12 != self.prev_Outputled12:
            self.prev_Outputled12 = self.Outputled12
            if self.Outputled12 == 1:
                self.set_image('Outputled12', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled12', 'LED-Black.jpg')

        # Outputled13 status
        self.Outputled13 = self.hal['Outputled13']
        if self.Outputled13 != self.prev_Outputled13:
            self.prev_Outputled13 = self.Outputled13
            if self.Outputled13 == 1:
                self.set_image('Outputled13', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled13', 'LED-Black.jpg')

        # Outputled14 status
        self.Outputled14 = self.hal['Outputled14']
        if self.Outputled14 != self.prev_Outputled14:
            self.prev_Outputled14 = self.Outputled14
            if self.Outputled14 == 1:
                self.set_image('Outputled14', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled14', 'LED-Black.jpg')

        # Outputled15 status
        self.Outputled15 = self.hal['Outputled15']
        if self.Outputled15 != self.prev_Outputled15:
            self.prev_Outputled15 = self.Outputled15
            if self.Outputled15 == 1:
                self.set_image('Outputled15', 'LED-Yellow.jpg')
            else:
                self.set_image('Outputled15', 'LED-Black.jpg')

        self.collet_closer_status = self.hal['collet-closer-status']
        if self.collet_closer_status != self.prev_collet_closer_status:
            self.prev_collet_closer_status = self.collet_closer_status
            if self.collet_closer_status == 1:
                self.set_image('collet_closer_image', 'collet-open.png')
            else:
                self.set_image('collet_closer_image', 'collet-closed.png')

        # watch for LinuxCNC to change spindle request state
        # when it does copy it to tormach.spindle HAL pin
        if self.prev_spindle_iocontrol != self.hal['spindle-iocontrol']:
            if self.hal['spindle-iocontrol']:
                if self.door_switch == 1:
                    # NB: this is a bit of a kludge.  It will show the error message to the user
                    # for MDI M3 with door open, but user will need to
                    # click Stop, Reset, or M5 before second M3 will be obeyed.
                    self.error_handler.write("Cannot start spindle with door open", ALARM_LEVEL_LOW)
                else:
                    self.hal['spindle'] = self.hal['spindle-iocontrol']
            else:
                self.hal['spindle'] = 0
            # current becomes previous
            self.prev_spindle_iocontrol = self.hal['spindle-iocontrol']

        # coolant status 
        if self.coolant_lockout <= 0:  #if not locked out
            self.coolant_status = self.hal['coolant']
            if self.coolant_status != self.prev_coolant_status:
                self.prev_coolant_status = self.coolant_status
                if self.coolant_status:
                    self.set_image('coolant_image', 'Coolant-Green.jpg')
                else:
                    self.set_image('coolant_image', 'Coolant-Black.jpg')
        else:
            #Turn off button always first pass during lockout period
            if self.coolant_lockout == COOLANT_LOCK_OUT_PERIOD:
                self.set_image('coolant_image', 'Coolant-Black.jpg')
            self.coolant_lockout -= 1 #decrement lockout count 
            self.hal['coolant'] = 0 #keep coolant off

        if not self.program_running(False):
            # step button from shuttlexpress
            # upon release advance or wrap to the next increment
            if self.hal['jog-step-button'] == 1:
                self.jog_step_button_was_pressed = 1
            elif self.hal['jog-step-button'] == 0 and self.jog_step_button_was_pressed:
                self.jog_step_button_was_pressed = 0
                self.jog_increment = self.jog_increment * 10
                # wrap if needed
                if self.jog_increment > 0.1:
                    self.jog_increment = 0.0001
                self.jog_mode_continuous = False
                self.clear_jog_LEDs()
                if self.jog_increment == 0.0001:
                    self.hal['jog-gui-step-index'] = 0
                    # TODO - I don't think that this hal pin is being used???
                    self.hal['jog-gui-ustep-size'] = self.jog_metric_scalar * self.jog_increment/1000.0
                    if self.g21:
                        self.set_image('jog_0001_image', 'Metric-001-Green.jpg')
                    else:
                        self.set_image('jog_0001_image', '0001-Pressed-Green.jpg')
                elif self.jog_increment == 0.001:
                    self.hal['jog-gui-step-index'] = 1
                    self.hal['jog-gui-ustep-size'] = self.jog_metric_scalar * self.jog_increment/1000.0
                    if self.g21:
                        self.set_image('jog_0010_image', 'Metric-100-Green.jpg')
                    else:
                        self.set_image('jog_0010_image', '0010-Pressed-Green.jpg')
                elif self.jog_increment == 0.01:
                    self.hal['jog-gui-step-index'] = 2
                    self.hal['jog-gui-ustep-size'] = self.jog_metric_scalar * self.jog_increment/1000.0
                    if self.g21:
                        self.set_image('jog_0100_image', 'Metric-1.00-Green.jpg')
                    else:
                        self.set_image('jog_0100_image', '0100-Pressed-Green.jpg')
                elif self.jog_increment == 0.1:
                    self.hal['jog-gui-step-index'] = 3
                    self.hal['jog-gui-ustep-size'] = self.jog_metric_scalar * self.jog_increment/1000.0
                    if self.g21:
                        self.set_image('jog_1000_image', 'Metric-10-Green.jpg')
                    else:
                        self.set_image('jog_1000_image', '1000-Pressed-Green.jpg')

        if self.hal['jog-axis-x-enabled'] or self.hal['jog-axis-y-enabled'] or self.hal['jog-axis-z-enabled'] or self.hal['jog-axis-a-enabled']:
            if (self.hal['jog-counts'] != self.prev_hal_jog_counts) or (self.hal['jog-ring-speed-signed'] != 0.0):
                self.ensure_mode(linuxcnc.MODE_MANUAL)

        # inner wheel step jogging
        jog_step_counts = self.hal['jog-counts']
        # if in a continuous jog do not do a step jog
        if jog_step_counts != self.prev_hal_jog_counts:
            count_delta = jog_step_counts - self.prev_hal_jog_counts

            # Choose jogging direction
            jog_dir = -1.0 if count_delta < 0 else 1.0

            self.jog_mode_continuous = False
            # do not apply jog rate override to shuttle jogging
            self.jog(self.jog_ring_axis, jog_dir, self.jog_speed, False)
            self.prev_hal_jog_counts = jog_step_counts

        # ring jogging
        ring_speed = self.hal['jog-ring-speed-signed']
        self.jog_ring_axis = self.hal['jog-ring-selected-axis']
        if not self.is_rapidturn:
            # in non-duality mode map Y and A to nothing
            if self.jog_ring_axis == 1 or self.jog_ring_axis == 3:
                self.jog_ring_axis = -1

        jog_ring_axis_changed = False
        if self.jog_ring_axis != self.prev_jog_ring_axis:
            self.stop_jog(self.prev_jog_ring_axis)
            self.prev_jog_ring_axis = self.jog_ring_axis
            jog_ring_axis_changed = True

        if self.jog_ring_axis >= 0 and (ring_speed != self.prev_jog_ring_speed or jog_ring_axis_changed):
            self.prev_jog_ring_speed = ring_speed

            jog_dir = -1.0 if ring_speed < 0.0 else 1.0

            # need new jog command
            # FIXME account for jog axis minimum speed here?
            if ring_speed == 0:
                self.stop_jog(self.jog_ring_axis)
            else:
                self.jog_mode_continuous = True
                # do not apply jog rate override to shuttle jogging
                self.jog(self.jog_ring_axis, jog_dir, abs(ring_speed), False)

        # always reset previous jog counts
        self.prev_hal_jog_counts = self.hal['jog-counts']

        if self.current_page == ALARMS_PAGE:
            self.refresh_encoder_leds()
            self.refresh_machine_ok_led()
            if self.usbio_enabled: self.refresh_usbio_leds()

        # limit switch overrides
        if self.status.limit[0] == 3:
            self.command.override_limits()
            # only unhome if axis is referenced
            if self.x_referenced and not self.status.axis[0]['homing']:
                self.command.unhome(0)

        # duality has Y axis
        if self.is_rapidturn:
            self.y_referenced = self.status.homed[1]
            if self.status.limit[1] == 3:
                self.command.override_limits()
                # only unhome if axis is referenced
                if self.y_referenced and not self.status.axis[1]['homing']:
                    self.command.unhome(1)

        if self.status.limit[2] == 3:
            self.command.override_limits()
            # only unhome if axis is referenced
            if self.z_referenced and not self.status.axis[2]['homing']:
                self.command.unhome(2)

    def save_persistent_data(self):
        self.redis.hset('machine_prefs', 'active_tool', self.current_tool)
        try:
            for item in self.mdi_history:
                self.redis.rpush('mdi_history', item)
                self.redis.hset('machine_prefs', 'mdi_history_index', self.mdi_history_index)
        except:
            self.error_handler.write("Failed to save MDI history.", ALARM_LEVEL_DEBUG)
            pass

        try:
            # delete old redis values to prevent this list from growing ad infinitum
            self.redis.delete('recent_file_history')
            for row in self.file_history_liststore:
                path = row[1]
                if path != CLEAR_CURRENT_PROGRAM:
                    self.redis.rpush('recent_file_history', path)
        except:
            self.error_handler.write("Failed to save recent file history.", ALARM_LEVEL_DEBUG)
            pass



    def on_exit_button_release_event(self, widget, data=None):
        do_button_release_motion(widget)
        conf_dialog = tormach_file_util.shutdown_confirmation_popup()
        self.window.set_sensitive(False)
        conf_dialog.run()
        response = conf_dialog.response
        if response == gtk.RESPONSE_CANCEL:
            self.window.set_sensitive(True)
            conf_dialog.destroy()
            return
        else:
            conf_dialog.destroy()
        self.quit()

    def quit(self):
        self.save_persistent_data()
        self._quit()
        gtk.main_quit()        
        
    def set_work_offset(self, axis, dro_text):
        if not (self.x_referenced and self.z_referenced):
            self.error_handler.write("Must reference X and Z axes before setting work offsets.", ALARM_LEVEL_MEDIUM)
            return
        dro_val = float(dro_text)
        # we're always displaying Dia, but LCNC wants radius
        if axis == "X":
            dro_val = dro_val/2
        current_work_offset = self.status.g5x_index
        offset_command = "G10 L20 P%d %s%.12f" % (current_work_offset, axis, dro_val)
        self.issue_mdi(offset_command)
        # be sure to switch modes to cause an interp synch, which
        # writes out the var file.
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.gremlin.clear_live_plotter()
        self.gremlin.load()


    # Axis style mode checks
    def program_running(self, do_poll=True):
        # are we running a gcode program?
        if do_poll:
            # need fresh status data, ask for it
            self.status.poll()
        return self.status.task_mode == linuxcnc.MODE_AUTO and self.status.interp_state != linuxcnc.INTERP_IDLE

    def spindle_running(self):
        if self.status.spindle_direction == 0:
            return False
        else:
            return True


    def ensure_mode(self, mode):
        # poll needed to prevent race condition on keyboard jogging where
        # ensure mode is called twice in quick succession and wait_complete()
        # results in runaway jog
        self.status.poll()
        if self.status.task_mode == mode:
            # if already in desired mode do nothing
            return True
        # set the desired mode
        self.error_handler.write("ensure_mode: changing LCNC mode to %s" % (self.get_lcnc_mode_string(mode)), ALARM_LEVEL_DEBUG)
        self.command.mode(mode)
        self.command.wait_complete()
        return True

    def active_gcodes(self):
        desired_codes = [8,6,5,1,4,7,13]
        active_codes = []
        for i in desired_codes:
            code = self.status.gcodes[i]
            if code % 10 == 0:
                active_codes.append("G%d" % (code/10))
            else:
                active_codes.append("G%(ones)d.%(tenths)d" % {'ones': code/10, 'tenths': code%10})
        return active_codes

    def refresh_tool_liststore(self):
        # open the tool table file and read it
        # alternately could use status.tool_table[] to retrieve these values
        # let's try this using the status.tool_table[], as the text file doesn't seem to get refreshed the way one would like
        # clear out old data
        self.tool_liststore.clear()
        for pocket in range(1, MAX_NUM_TOOLS):
            wear_offset_pocket = pocket
            geo_offset_pocket  = pocket + MAX_NUM_TOOLS
            tool_num = self.status.tool_table[wear_offset_pocket].id
            try:
                description = self.redis.hget('tool_descriptions', str(tool_num))
            except:
                description = "None"

            # some values stored as strings so we can format them for display
            tool_type = self.get_tool_type(tool_num)
            geo_x = self.dro_long_format % (2 * self.status.tool_table[geo_offset_pocket].xoffset * self.mm_inch_scalar)
            geo_z = self.dro_long_format % (self.status.tool_table[geo_offset_pocket].zoffset * self.mm_inch_scalar)
            wear_x = 2 * self.status.tool_table[wear_offset_pocket].xoffset * self.mm_inch_scalar
            wear_z = self.status.tool_table[wear_offset_pocket].zoffset * self.mm_inch_scalar
            tool_radius = self.dro_long_format % (self.status.tool_table[wear_offset_pocket].diameter * self.mm_inch_scalar/ 2) # Tool table stores the nose value as a diameter
            tool_orientation = self.status.tool_table[wear_offset_pocket].orientation

            dia_wear_text_color = self.find_wear_offset_display_color(wear_x, tool_type)
            if wear_z < -0.00009:
                z_wear_text_color = BLUE
            elif wear_z > 0.00009:
                z_wear_text_color = ORANGE
            else:
                z_wear_text_color = BLACK

            wear_x = self.dro_long_format % wear_x
            wear_z = self.dro_long_format % wear_z

            self.tool_liststore.append([tool_num, description, geo_x, geo_z, wear_x, wear_z, tool_radius, tool_orientation, dia_wear_text_color, z_wear_text_color])



    def refresh_work_offset_liststore(self):
        self.work_liststore.clear()
        self.work_liststore.append(["G54"] + list(self.status.g54_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G55"] + list(self.status.g55_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G56"] + list(self.status.g56_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G57"] + list(self.status.g57_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G58"] + list(self.status.g58_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G59"] + list(self.status.g59_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G59.1"] + list(self.status.g59_1_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G59.2"] + list(self.status.g59_2_offset[:4]) + ['GREY'] + ['WHITE'])
        self.work_liststore.append(["G59.3"] + list(self.status.g59_3_offset[:4]) + ['GREY'] + ['WHITE'])
        # highlight active work offset
        row = self.work_liststore[self.status.g5x_index - 1]
        row[5] = BLACK
        row[6] = ROW_HIGHLIGHT


    def find_wear_offset_display_color(self, wear, tool_type):
        # according to Smid, CNC Programming Handbook, pp 105 - 106,
        # a wear offset is defined as the difference between the programmed
        # value (e.g. 3.0000) and the measured value (e.g. 3.0040).  If the
        # part comes out large on a machine set up as a rear tool post lathe,
        # and you're using a rear tool post tool, you would enter a neg number
        # (e.g. -0.0040).  This is opposite for front tool post tools on a
        # rear tool post lathe: positive number makes the part smaller.
        # For boring bars, an rtp boring bar needs a positive number to make the
        # feature (the bored hole) bigger.
        if wear > 0.00009:
            if 'rtp' in tool_type:
                return ORANGE
            elif 'ftp' in tool_type:
                return BLUE
            else:
                return BLACK
        elif wear < -0.00009:
            if 'rtp' in tool_type:
                return BLUE
            elif 'ftp' in tool_type:
                return ORANGE
            else:
                return BLACK
        return BLACK


    def get_tool_type(self, tool):
        if tool == 0:
            return 'none'
        pocket = tool + MAX_NUM_TOOLS
        front_angle_val = self.status.tool_table[pocket].frontangle
        # if not yet set, return 'none'
        if front_angle_val == 0.0:
            return 'none'
        for tool_type_string, tool_type in self.tool_type_dic.iteritems():
            if tool_type == front_angle_val:
                return tool_type_string
        return 'none'

    def write_tool_type(self, tool, tool_type):
        # we only call this if a user selects an image form the tool tab,
        # and if they're selecting an image, the touch_x and touch_z leds should be balck
        self.set_touch_x_z_leds(tool)
        # get tool_type as string, convert to integer, store in the geo offset register frontangle
        # field of the tool table
        if tool == 0:
            return
        if 'rtp' in tool_type:
            self.rtp_text.show()
            self.ftp_text.hide()
        elif 'ftp' in tool_type:
            self.ftp_text.show()
            self.rtp_text.hide()
        else:
            self.rtp_text.hide()
            self.ftp_text.hide()
        coded_tool_type = self.tool_type_dic[tool_type]
        front_angle = self.front_angle_dic[tool_type]
        back_angle = self.back_angle_dic[tool_type]
        orientation = self.orientation_dic[tool_type]
        geo_offset_register = tool + 10000
        g10_command = "G10 L1 P%d I%f" % (geo_offset_register, coded_tool_type)
        self.issue_mdi(g10_command)
        self.command.wait_complete()
        g10_command = "G10 L1 P%d I%d J%d Q%d" % (tool, front_angle, back_angle, orientation)
        self.issue_mdi(g10_command)

    def load_cs_image(self, color):
        if color == 'green' and self.cycle_start_led_color != 'green':
            self.set_image('cycle_start_image', 'Cycle-Start-Green.jpg')
            self.cycle_start_led_color = 'green'
            return
        if color == 'dark' and self.cycle_start_led_color != 'dark':
            self.set_image('cycle_start_image', 'Cycle-Start-Black.jpg')
            self.cycle_start_led_color = 'dark'
            return
        if color == 'blink':
            if self.cycle_start_led_color == 'green':
                self.set_image('cycle_start_image', 'Cycle-Start-Black.jpg')
                self.cycle_start_led_color = 'dark'
                return
            else:
                self.set_image('cycle_start_image', 'Cycle-Start-Green.jpg')
                self.cycle_start_led_color = 'green'
                return

    def blink_tool_request_label(self, tool_number):
        if 'T' in self.tool_label.get_text():
            self.tool_label.set_text('')
        else:
            self.tool_label.set_text("T:    " + tool_number)


    def set_touch_x_z_leds(self, tool):
        tool += MAX_NUM_TOOLS
        self.status.tool_table[tool].xoffset
        if self.status.tool_table[tool].xoffset == 0:
            self.set_image('touch_x_image', 'touch_x_black_led.png')
            self.dro_list['touch_dia_dro'].set_text("")
        else:
            self.set_image('touch_x_image', 'touch_x_green_led.png')
        if self.status.tool_table[tool].zoffset == 0:
            self.set_image('touch_z_image', 'touch_z_black_led.png')
            self.dro_list['touch_z_dro'].set_text("")
        else:
            self.set_image('touch_z_image', 'touch_z_green_led.png')


    # ------------------------------------------------------------------------
    # helpers
    # ------------------------------------------------------------------------



    def stop_motion_safely(self):
        #Send abort message to motion to stop any movement
        self.command.abort()

        #self.command.wait_complete()
        if self.hal['coolant']:
            self.hal['coolant'] = False
        if self.feedhold_active:
            self.feedhold_active = False
            #TODO move this to status update
            self.set_image('feedhold_image', 'Feedhold-Black.jpg')
        if self.is_gcode_program_loaded: self.gcodelisting_mark_motion_line(0)


    def start_regression_test_vehicle(self):
        print "starting regression test vehicle"
        # pick a csv file or folder
        gcode_path = os.path.join(os.getenv('HOME')) + os.path.sep + 'gcode'
        with tormach_file_util.save_as_popup(gcode_path, self.touchscreen_enabled) as dialog:
            path = dialog.path
        dros = [self.conv_dro_list,
                self.od_turn_dro_list,
                self.id_basic_dro_list,
                self.id_turn_dro_list,
                self.face_dro_list,
                #self.chamfer_dro_list,
                #self.corner_dro_list,
                self.corner_chamfer_od_dro_list,
                self.corner_chamfer_id_dro_list,
                self.corner_radius_od_dro_list,
                self.corner_radius_id_dro_list,
                self.groove_dro_list,
                self.part_dro_list,
                self.drill_dro_list,
                self.tap_dro_list,
                self.thread_dro_list]

        test = regression_tests.test(self.status, dros, path, self.error_handler)


    # ----------------------------------------------------
    # load item named in self.image_list with file
    # used for most if not all gtk.Images
    # ----------------------------------------------------
    def set_image(self, image_list_name, file_name):
        # look for pixbuf with this filename's image in it
        try:
            self.image_list[image_list_name].set_from_pixbuf(self.pixbuf_dict[file_name])
        except KeyError:
            # not yet loaded, attempt to load it then add to pixbuf dictionary
            try:
                self.image_list[image_list_name].set_from_file(os.path.join(GLADE_DIR, file_name))
                self.pixbuf_dict[file_name] = self.image_list[image_list_name].get_pixbuf()
            except:
                print 'failed to load image %s with file %s' % (image_list_name, file_name)

    def refresh_gremlin_offsets(self):
#       if self.current_g5x_offset != self.status.g5x_offset:
#           self.current_g5x_offset = self.status.g5x_offset
#           self.redraw_gremlin()
#       elif self.current_g92_offset != self.status.g92_offset:
#           self.current_g92_offset = self.status.g92_offset
#           self.redraw_gremlin()
	self.redraw_gremlin()
    
    def redraw_gremlin(self):
        # redraw screen with new offset
        # be sure to switch modes to cause an interp synch, which
        # writes out the var file.
        self.ensure_mode(linuxcnc.MODE_MANUAL)
        self.ensure_mode(linuxcnc.MODE_MDI)
        self.gremlin.clear_live_plotter()
        self.gremlin.load()
        #Note:not catching warnings here since it's assumed the user has
        #already seen them            

    def refresh_jog_active_leds(self):
        self.x_jog_enabled = self.hal['jog-axis-x-enabled']
        if self.x_jog_enabled != self.prev_x_jog_enabled:
            self.prev_x_jog_enabled = self.x_jog_enabled
            # x jog enabled changed
            if self.x_jog_enabled:
                self.set_image('jog_x_active_led', 'LED-Green.jpg')
            else:
                self.set_image('jog_x_active_led', 'LED-Black.jpg')

        self.z_jog_enabled = self.hal['jog-axis-z-enabled']
        if self.z_jog_enabled != self.prev_z_jog_enabled:
            self.prev_z_jog_enabled = self.z_jog_enabled
            # z jog enabled changed
            if self.z_jog_enabled:
                self.set_image('jog_z_active_led', 'LED-Green.jpg')
            else:
                self.set_image('jog_z_active_led', 'LED-Black.jpg')


        if self.is_rapidturn:
            self.y_jog_enabled = self.hal['jog-axis-y-enabled']
            if self.y_jog_enabled != self.prev_y_jog_enabled:
                self.prev_y_jog_enabled = self.y_jog_enabled
                # z jog enabled changed
                if self.y_jog_enabled:
                    self.set_image('jog_y_active_led', 'LED-Green.jpg')
                else:
                    self.set_image('jog_y_active_led', 'LED-Black.jpg')


    def refresh_encoder_leds(self):
        if self.is_rapidturn:
            # do something appropriate for duality
            if self.hal['encoder-z']:
                self.set_image('encoder_z_led', 'LED-Green.jpg')
            else:
                self.set_image('encoder_z_led', 'LED-Black.jpg')
        else:
            # update encoder count dro
            # sign of count inverted to match sign of what is sent to LCNC from HAL
            self.dro_list['encoder_counts_dro'].set_text(str(1 * self.hal['encoder-count']))
            self.dro_list['encoder_counts_dro1'].set_text(str(1 * self.hal['encoder-count1']))
            self.dro_list['encoder_counts_dro2'].set_text(str(1 * self.hal['encoder-count2']))
            #self.encoder_counts_dro.set_text(self.hal['encoder-count'])
            # update encoder LEDs
            if self.hal['encoder-a']:
                self.set_image('encoder_a_led', 'LED-Green.jpg')
            else:
                self.set_image('encoder_a_led', 'LED-Black.jpg')
            if self.hal['encoder-b']:
                self.set_image('encoder_b_led', 'LED-Green.jpg')
            else:
                self.set_image('encoder_b_led', 'LED-Black.jpg')
            if self.hal['encoder-z']:
                self.set_image('encoder_z_led', 'LED-Green.jpg')
            else:
                self.set_image('encoder_z_led', 'LED-Black.jpg')


    def refresh_machine_ok_led(self):
        if self.hal['machine-ok']:
            self.set_image('machine_ok_led', 'LED-Green.jpg')
        else:
            self.set_image('machine_ok_led', 'LED-Black.jpg')


    def switch_inch_metric_display(self):
        # swap button art on jog step sizes
        self.clear_jog_LEDs()
        # store off in redis for startup in same mode next time.
        self.redis.hset('machine_prefs', 'g21', self.g21)
        self.refresh_css_label()
        if self.g21:
            self.dro_long_format = "%3.3f"
            self.dro_tpu_format = "%3.2f"
            self.mm_inch_scalar = 25.4
            self.jog_metric_scalar = 100
            self.hal['jog-is-metric'] = True
        else:
            self.dro_long_format = "%2.4f"
            self.dro_tpu_format = "%2.2f"
            self.mm_inch_scalar = 1
            self.jog_metric_scalar = 1
            self.hal['jog-is-metric'] = False

    def refresh_css_label(self):
        if self.g21:
            if self.css_active:
                self.set_image('css_image', 'CSS-Yellow.jpg')
                self.css_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >mpm</span>')
            else:
                self.set_image('css_image', 'CSS-Black.jpg')
                self.css_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="#A19D9F" >mpm</span>')
        else:
            if self.css_active:
                self.set_image('css_image', 'CSS-Yellow.jpg')
                self.css_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >sfm</span>')
            else:
                self.set_image('css_image', 'CSS-Black.jpg')
                self.css_label.set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="#A19D9F" >sfm</span>')



    def pause_program_for_door_sw_open(self):
        self.error_handler.write("Pausing program because door sw was opened", ALARM_LEVEL_DEBUG)
        self.program_paused_for_door_sw_open = True
        self.command.auto(linuxcnc.AUTO_PAUSE)
        self.feedhold_active = True
        self.set_image('feedhold_image', 'Feedhold-Green.jpg')
        if self.hal['coolant']:
            self.hal['coolant'] = False
        if self.hal['spindle']:
            self.hal['spindle'] = False


    def resume_program_from_door_sw_open(self):
        self.error_handler.write("Resuming program because door sw was closed", ALARM_LEVEL_DEBUG)
        self.program_paused_for_door_sw_open = False
        # iocontrol.coolant is checked in the 50ms periodic.  If it doesn't match the previous state,
        # we flip the hal.coolant bit accordingly.  This next line will force this to happen
        self.prev_coolant_iocontrol = not self.hal['coolant-iocontrol']
        self.prev_spindle_iocontrol = not self.hal['spindle-iocontrol']
        pass

    # rapidturn extras
    def rapidturn_init(self):
        # turn on the small Y axis DRO
        self.builder.get_object('y_dro').set_visible(True)
        self.builder.get_object('jog_y_active_led').set_visible(True)
        self.builder.get_object("y_dro_label").set_visible(True)

        # switch to mill button
        self.button_list['switch_to_mill'].show()
        self.button_list['switch_to_mill'].set_sensitive(True)
        self.builder.get_object("switch_to_mill_image").set_visible(True)

        # turn on the ref Y button
        self.button_list['ref_y'].show()
        self.button_list['ref_y'].set_sensitive(True)
        self.builder.get_object("ref_y_image").set_visible(True)
        
        # hide collet closer button
        self.button_list['collet_closer'].hide()
       
        # hide the turret choice on the settings screen (which hides the turret button too)
        self.builder.get_object('use_atc_checkbutton').hide()
        
        # change labels on lo/hi spindle checkboxes
        self.builder.get_object('use_5c_pulley_ratio_text').set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >high    speed</span>')
        self.builder.get_object('use_d1_4_pulley_ratio_text').set_markup('<span weight="light" font_desc="Bebas 12" font_stretch="ultracondensed" foreground="white" >low    speed</span>')

        # hide notes about negative side of workpiece in conv. screens
        list_of_notes_to_hide = ['tool_side_note_text', 'tool_side_note_text3', 'corner_tool_side_text', 'groove_part_tool_side_text', 
                                 'thread_tool_side_text', 'ftp_text']
        for note in list_of_notes_to_hide:
            self.builder.get_object(note).hide()
            
        # remove 'tap' text on drill/tap label and hide tap button
        self.builder.get_object('drill_tab_label').set_text('Drill')
        self.button_list['drill_tap'].hide()
        
        # door-sw led and encoder-related stuff on settings screen
        self.builder.get_object('door_sw_text').hide()
        self.builder.get_object('door_sw_led').hide()
        self.builder.get_object('encoder_counts_text').hide()
        self.builder.get_object('encoder_a_text').hide()
        self.builder.get_object('encoder_a_led').hide()
        self.builder.get_object('encoder_b_text').hide()
        self.builder.get_object('encoder_b_led').hide()
        
        # spindle range
        #FIXME these look like globals but are not!
        # get values from INI file?
        HIGH_SPEED_PULLEY_RATIO = (1.0/2.163)
        LOW_SPEED_PULLEY_RATIO = 1
        LOW_SPEED_PULLEY_MAX_RPM = 2000
        HIGH_SPEED_PULLEY_MAX_RPM = 4150

    def stop_jog(self, jog_axis):
        if self.status.task_mode == linuxcnc.MODE_MANUAL:
            self.command.jog(linuxcnc.JOG_STOP, jog_axis)

    def get_jog_increment(self, axis_ind):
        """Return a jog increment based on the specified axis index"""
        increment = self.jog_increment * ( self.jog_metric_scalar if axis_ind < 3 else 1.0)
        if axis_ind == 0: increment = increment/2
        return increment

    def jog(self, jog_axis, jog_direction, jog_speed, apply_pct_override=True):
        if self.is_front_toolpost_lathe and jog_axis == 0:
            jog_direction *= -1
        if not self.program_running(True):
            self.ensure_mode(linuxcnc.MODE_MANUAL)
            if self.jog_mode_continuous:
                speed = jog_direction * jog_speed
                if apply_pct_override:
                    speed = speed * self.jog_override_pct
                self.command.jog(linuxcnc.JOG_CONTINUOUS, jog_axis, speed)
            else:
                # next call is cribbed from Axis to prevent stacking up inc. commands on mulitple keystrokes
                if self.status.state != 1: return

                self.command.jog(linuxcnc.JOG_INCREMENT,  jog_axis, jog_direction * jog_speed * self.jog_override_pct, self.get_jog_increment(jog_axis)/self.get_axis_scale(jog_axis))



    # g code preview, set start line, edit g code


    def load_gcode_file(self, path):
        if self.program_running():
            return

        self.is_gcode_program_loaded = True
        # remember what was last loaded to watch for changes on disk and reload
        self.set_current_gcode_path(path)
        if not path:
            self.gcodelisting_buffer.set_text('')
            return
        # prevent changes to the combo box from causing file loads
        self.combobox_masked = True
        # remove filename from previous model position if it was previously in the model
        sort_file_history(self.file_history_liststore, path)
        # add filename, path to the model
        self.file_history_liststore.prepend([os.path.basename(path), path])
        # have to set active one, else the active file won't be displayed on the combobox
        self.loaded_gcode_filename_combobox.set_active(0)
        # unmask
        self.combobox_masked = False
        # read file directly into buffer
        tormach_file_util.open_text_file(self, path, self.gcodelisting_buffer)


        # can change this with right-click menu in source view
        # this is one based, the textbuffer is zero based
        self.gcode_start_line = 1;
        # must switch to mdi, then back to force clear of _setup.file_pointer, otherwise
        # we can't open a program if one is already open
        self.ensure_mode(linuxcnc.MODE_MDI)
        # load file into LinuxCNC
        self.ensure_mode(linuxcnc.MODE_AUTO)
        self.command.program_open(path)
        # note the time stamp
        self.gcode_file_mtime = os.stat(path).st_mtime
        # gremlin is unpredictable at the moment
        # wrap it for exceptions
        try:
            self.gremlin.clear_live_plotter()
        except Exception as e:
            print 'gremlin.clear_live_plotter() raised an exception'
            msg = "An exception of type {0} occured, these were the arguments:\n{1!r}"
            print msg.format(type(e).__name__, e.args)
            #traceback_txt = "".join(traceback.format_exception(*sys.exc_info()))
            #print traceback_txt

        try:
            # with no filename given, gremlin will ask LinuxCNC for the filename
            result, seq, warnings = self.gremlin.load()
            #Quick way to dump warnings to status window
            self.gremlin.report_gcode_warnings(warnings,os.path.basename(path))
        except Exception as e:
            print 'gremlin.load() raised an exception'
            msg = "An exception of type {0} occured, these were the arguments:\n{1!r}"
            print msg.format(type(e).__name__, e.args)
            #traceback_txt = "".join(traceback.format_exception(*sys.exc_info()))
            #print traceback_txt

        try:
            self.gremlin.set_view_y()
        except Exception as e:
            print 'gremlin.set_y_view() raised an exception'
            msg = "An exception of type {0} occured, these were the arguments:\n{1!r}"
            print msg.format(type(e).__name__, e.args)
            #traceback_txt = "".join(traceback.format_exception(*sys.exc_info()))
            #print traceback_txt
        # switch to gcode listing MDI main tab
        if not self.interp_alarm:
            self.notebook.set_current_page(MAIN_PAGE)
        self.window.set_focus(None)
        self.gcodelisting_mark_start_line(self.gcode_start_line)
        self.gcode_pattern_search.on_load_gcode()


    def update_tool_orientation_image(self):
        tip_orient = self.status.tool_table[self.current_tool].orientation
        
        # apologies for the following kludge: this code requires the orientation
        # images to follow a naming convention - either ftp_orientation_x.png or 
        # orientation_x.png where x is the numeric orientation.  ftp prefix is for
        # a front tool post lathe
        image_filename = 'orientation_' + str(tip_orient) + '.png'
        if self.is_front_toolpost_lathe:
            image_filename = 'ftp_' + image_filename
            
        try:
            self.set_image('tool_orientation_image',  image_filename)
        except:
            self.set_image('tool_orientation_image', 'orientation_0.png')

    def delete_event(self, widget, event):
        try:
            self.enqueue_button_press_release(self.button_list['exit'])
        except Exception as e:
            print 'enqueue button press failed'
            msg = "An exception of type {0} occured, these were the arguments:\n{1!r}"
            print msg.format(type(e).__name__, e.args)
        return True
    
    def ja_gremlin(self, width, height):
        return jaGremlin(self,width,height)
    
#--end 'lathe'-------------------------------------------------------------------
    
def hide_notebook_tabs(notebook):
    for i in [1, 2, 3, 4, 5]:
        page = notebook.get_nth_page(i)
        page.hide()

def show_notebook_tabs(notebook):
    for i in [1, 2, 3, 4, 5]:
        page = notebook.get_nth_page(i)
        page.show()


def is_dro_masked(dro_list):
    for name, dro in dro_list.iteritems():
        if dro.masked == True:
            return True
    return False




# this overrides 'posstrs()' to not display the DROs in Gremlin
class Tormach_Lathe_Gremlin(gremlin.Gremlin):
    def __init__(self, ui, width, height):
        gremlin.Gremlin.__init__(self, ui.inifile)
        self.status = ui.status
        self.ui_view = 'y'
        # necessary to support recent changes to gremlin in 2.6.  We might want to make this configurable
        # down the road, but for now it's set to a grid of .1 inches.  The grid doesn't display for me
        # but at least the UI will load without error again.
        self.grid_size = 0.5
        self.show_g53_origin = False

        #Crude way to store pointer to UI so we can access error handling
        #functions
        self.ui=ui
        self.connect("button_press_event", self.on_gremlin_double_click)
        self.set_size_request(width, height)
        
    def realize(self,widget):
        super(Tormach_Lathe_Gremlin, self).realize(widget)
        
    def set_grid_size(self, size):
        self.grid_size = size
        self._redraw()

    def get_grid_size(self):
        return self.grid_size

    def get_view(self):
        # gremlin as used as a gladevcp widegt has a propery for the view, which for a lathe
        # should be 'y'.  When it's not right the program extents won't be drawn.
        view_dict = {'x':0, 'y':1, 'z':2, 'p':3}
        return view_dict.get('y')

    def get_show_metric(self):
        if self.status.gcodes[5] == 200:
            return False
        else:
            return True

    def posstrs(self):
        l, h, p, d = gremlin.Gremlin.posstrs(self)
        return l, h, [''], ['']

    def report_gcode_error(self, result, seq, filename):
        import gcode
        error_str = gcode.strerror(result)
        error_str = "\n\nG-Code error in " + os.path.basename(filename) + "\n" + "Near line " + str(seq) + " of\n" + filename + "\n" + error_str + "\n"
        self.ui.error_handler.write(error_str)
        self.ui.interp_alarm = True

    def report_gcode_warnings(self, warnings, filename, suppress_after = 3):
        """ Show the warnings from a loaded G code file.
        Accepts a list of warnings produced by the load_preview function, the
        current filename, and an optional threshold to suppress warnings after.
        """

        num_warnings = len(warnings)
        if num_warnings == 0:
            return

        # Find the maximum number of warnings to print
        max_ind = min(max(suppress_after,0), num_warnings)
        # Find out how many we're suppressing if any
        num_suppressed = max(num_warnings-max_ind,0)

        # NOTE: copied from mill, to prevent status page switch on warning
        # warn, but still switch to main page if loading the file
        self.ui.notebook.set_current_page(MAIN_PAGE)


        #Iterate in reverse to print a coherent list to the status window, which shows most recent first
        if num_suppressed:
            self.ui.error_handler.write("Suppressed %d more warnings" % num_suppressed)
        else:
            self.ui.error_handler.write("*** End of warning list ***")

        for w in warnings[max_ind::-1]:
            # Add a space to show that the warning is part of the list
            self.ui.error_handler.write("  "+w)

        warning_header = "G-Code warnings in " + os.path.basename(filename)
        self.ui.error_handler.write(warning_header)

        self.ui.interp_alarm = True

    def load_display(self,filename = None):
        # must subclass to get rid of call to set_view, which (on work offset setting as opposed to
        # actual file load) causes a redraw and a loss of any zoom settigns.
        s = self.stat
        s.poll()
        if not filename and s.file:
            filename = s.file
        elif not filename and not s.file:
            return

        td = tempfile.mkdtemp()
        self._current_file = filename
        try:
            random = int(self.inifile.find("EMCIO", "RANDOM_TOOLCHANGER") or 0)
            canon = StatCanon(self.colors, self.get_geometry(),self.lathe_option, s, random)
            parameter = self.inifile.find("RS274NGC", "PARAMETER_FILE")
            temp_parameter = os.path.join(td, os.path.basename(parameter or "linuxcnc.var"))
            if parameter:
                shutil.copy(parameter, temp_parameter)
            canon.parameter_file = temp_parameter

            unitcode = "G%d" % (20 + (s.linear_units == 1))
            initcode = self.inifile.find("RS274NGC", "RS274NGC_STARTUP_CODE") or ""
            result, seq = self.load_preview(filename, canon, unitcode, initcode)
            if result > gcode.MIN_ERROR:
                self.report_gcode_error(result, seq, filename)

        finally:
            shutil.rmtree(td)
            
    def on_gremlin_double_click(self, widget, event, data=None):
        if event.type == gtk.gdk._2BUTTON_PRESS:
            self.clear_live_plotter()
            return
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            # it's a right click if event.button == 3
            menu = gtk.Menu()
            imperial = not self.ui.g21
            sml_text = "Grid 0.1 inch" if imperial else "Grid 5 mm"            
            med_text = "Grid 0.5 inch" if imperial else "Grid 10 mm"
            lrg_text = "Grid 1.0 inch" if imperial else "Grid 25 mm"
            
            set_grid_size_small = gtk.MenuItem(sml_text)
            set_grid_size_med = gtk.MenuItem(med_text)
            set_grid_size_large = gtk.MenuItem(lrg_text)
            set_grid_size_small.connect("activate", self.set_grid_size_small)
            set_grid_size_med.connect("activate", self.set_grid_size_med)
            set_grid_size_large.connect("activate", self.set_grid_size_large)
            menu.append(set_grid_size_small)
            menu.append(set_grid_size_med)
            menu.append(set_grid_size_large)
            menu.popup(None, None, None, event.button, event.time)
            set_grid_size_small.show()
            set_grid_size_med.show()
            set_grid_size_large.show()

    def set_current_ui_view(self):
        self.set_view_y()
        self._redraw()
    
    def set_grid_size_small(self, widget):
        size = (5/25.4) if self.ui.g21 else .1
        self.set_grid_size(size)

    def set_grid_size_med(self, widget):
        size = (10/25.4) if self.ui.g21 else .5
        self.set_grid_size(size)

    def set_grid_size_large(self, widget):
        size = (25/25.4) if self.ui.g21 else 1.0
        self.set_grid_size(size)

class jaGremlin(Tormach_Lathe_Gremlin):
    def __init__(self, ja, width, height):
        Tormach_Lathe_Gremlin.__init__(self, ja.conversational.ui, width, height)
        self.current_view = 'y'
        self.ui_view = 'y'        
        self.spooler = None
               
    def report_gcode_warnings(self, warnings, filename, suppress_after = 3):
        print 'jaGremlin.report_gcode_warnings: file: %s, %d warnings' % (filename,len(warnings))
    
    def load_gcode_list(self, gcode_list):
        if self.initialised:
            path = TormachUIBase.gcode_list_to_tmp_file(gcode_list)
            if path is not None:
                self.load(path)
                self.queue_draw()
            self.spooler = None
        else:
            self.spooler = gcode_list
       
    def realize(self,widget):
        super(jaGremlin, self).realize(widget)
        if self.spooler is not None:
            self.load_gcode_list(self.spooler)
            self.set_default_view()
            self.queue_draw()
        
    def set_default_view(self):
        self.current_view = self.ui_view = 'y'        
        if self.initialised:
            self.set_current_view()
            

if __name__ == "__main__":
    # unbuffer stdout so print() shows up in sync with other output
    # the pipe from the redirect in operator_login causes buffering
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
    UI = lathe()
    UI.window.show_all()
    screen_width = gtk.gdk.Screen().get_width()
    screen_height = gtk.gdk.Screen().get_height()
    if screen_width > 1024 and screen_height > 768:
        UI.window.set_decorated(True)
        UI.window.connect('delete_event', UI.delete_event)
    UI.window.resize(1024, 768)
    UI.notebook.set_current_page(0)
    UI.gremlin.set_view_y()
    UI.kill_splash_screen()
    gtk.main()

    sys.exit(UI.program_exit_code)

